
  Import Dom Prog.

  Definition InitCallState :=  Method * (OperandStack.t * LocalVar.t).
  Definition IntraNormalState := PC * (Heap.t * OperandStack.t * LocalVar.t).
  Definition IntraExceptionState := Heap.t * Location.
  Definition ReturnState := Heap.t * ReturnVal.


  Inductive NormalStep (p:Program) : Method -> IntraNormalState -> IntraNormalState  -> Prop :=
  | aconst_null : forall h m pc pc' s l,

    instructionAt m pc = Some Aconst_null ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,s,l)) (pc',(h,(Null::s),l))

  | arraylength : forall h m pc pc' s l loc length tp, 

    instructionAt m pc = Some Arraylength ->
    next m pc = Some pc' ->
    Heap.typeof  h loc = Some (Heap.LocationArray length tp) ->

   NormalStep p m  (pc,(h,(Ref loc::s),l)) (pc',(h,(Num (I length)::s),l))

 | checkcast1 : forall h m pc pc' s l val t,

    instructionAt m pc = Some (Checkcast t) ->
    next m pc = Some pc' ->
    assign_compatible p h val (ReferenceType t) ->

   NormalStep p m (pc,(h,(val::s),l))  (pc',(h,(val::s),l))

  | const : forall h m pc pc' s l t z,

    instructionAt m pc = Some (Const t z) ->
    next m pc = Some pc' ->
    (   (t=BYTE /\ -2^7 <= z < 2^7)
     \/ (t=SHORT /\ -2^15 <= z < 2^15)
     \/ (t=INT /\ -2^31 <= z < 2^31)   ) ->

   NormalStep p m (pc,(h,s,l)) (pc',(h,(Num (I (Int.const z))::s),l))

  | dup : forall h m pc pc' s l v,

    instructionAt m pc = Some Dup ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v::s),l))  (pc',(h,(v::v::s),l))

  | dup_x1 : forall h m pc pc' s l v1 v2,

    instructionAt m pc = Some Dup_x1 ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::s),l)) (pc',(h,(v1::v2::v1::s),l))

  | dup_x2 : forall h m pc pc' s l v1 v2 v3,

    instructionAt m pc = Some Dup_x2 ->
    next m pc = Some pc' ->

   NormalStep p  m (pc,(h,(v1::v2::v3::s),l))  (pc',(h,(v1::v2::v3::v1::s),l))

  | dup2 : forall h m pc pc' s l v1 v2,

    instructionAt m pc = Some Dup2 ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::s),l)) (pc',(h,(v1::v2::v1::v2::s),l))

  | dup2_x1 : forall h m pc pc' s l v1 v2 v3,

    instructionAt m pc = Some Dup2_x1 ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::v3::s),l)) (pc',(h,(v1::v2::v3::v1::v2::s),l))

  | dup2_x2 : forall h m pc pc' s l v1 v2 v3 v4,

    instructionAt m pc = Some Dup2_x2 ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::v3::v4::s),l)) (pc',(h,(v1::v2::v3::v4::v1::v2::s),l))

  | getfield : forall h m pc pc' s l loc f v cn,

    instructionAt m pc = Some (Getfield f) ->
    next m pc = Some pc' ->
    Heap.typeof h loc = Some (Heap.LocationObject cn) -> 
    defined_field p cn f ->
    Heap.get h (Heap.DynamicField loc f) = Some v ->    

   NormalStep p m (pc,(h,(Ref loc::s),l)) (pc',(h,(v::s),l))

 | getstatic : forall h m pc pc' s l  f v,

    instructionAt m pc = Some (Getstatic f) ->
    next m pc = Some pc' ->
    isStatic p f ->
    Heap.get h (Heap.StaticField f) = Some v ->

   NormalStep p m (pc,(h,s,l)) (pc',(h,(v::s),l))

  | goto : forall h m pc s l o,

    instructionAt m pc = Some (Goto o) ->

   NormalStep p m (pc,(h,s,l)) (OFFSET.jump pc o,(h,s,l))

  | i2b : forall h m pc pc' s l i,

    instructionAt m pc = Some I2b ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(Num (I i)::s),l)) (pc',(h,(Num (I (b2i (i2b i)))::s),l))

  | i2s : forall h m pc pc' s l i,

    instructionAt m pc = Some I2s ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(Num (I i)::s),l))  (pc',(h,(Num (I (s2i (i2s i)))::s),l))

  | ibinop : forall h m pc pc' s l op i1 i2,

    instructionAt m pc = Some (Ibinop op) ->
    next m pc = Some pc' ->
    (op = DivInt \/ op = RemInt -> ~ Int.toZ i2 = 0) -> 

   NormalStep p m (pc,(h,(Num (I i2)::Num (I i1)::s),l))
                          (pc',(h,(Num (I (SemBinopInt op i1 i2))::s),l))

  | if_acmp_step_jump : forall h m pc s l val2 val1 o cmp,
      instructionAt m pc = Some (If_acmp cmp o) ->
      SemCompRef cmp val1 val2 ->
  (******************************************************************)
     NormalStep p m (pc,(h,(val2:: val1::s),l))
                                   (OFFSET.jump pc o,(h,s,l))

  | if_acmp_step_continue : forall h m pc pc' s l val2 val1 o cmp,
      instructionAt m pc = Some (If_acmp cmp o) ->
      next m pc = Some pc' ->
      ~ SemCompRef cmp val1 val2 ->
  (******************************************************************)
    NormalStep p m (pc,(h,(val2::val1::s),l)) (pc',(h,s,l))

  | if_icmp_step_jump : forall h m pc s l cmp i2 i1 o,
      instructionAt m pc = Some (If_icmp cmp o) ->
      SemCompInt cmp (Int.toZ i1) (Int.toZ i2) ->
  (******************************************************************)
    NormalStep p m (pc,(h,(Num(I i2)::Num(I i1)::s),l))
                                  (OFFSET.jump pc o,(h,s,l))

  | if_icmpeq_step_continue : forall h m pc pc' s l cmp i2 i1 o,
      instructionAt m pc = Some (If_icmp cmp o) ->
      next m pc = Some pc' ->
      ~ SemCompInt cmp (Int.toZ i1) (Int.toZ i2) ->
  (******************************************************************)
    NormalStep p m (pc,(h,(Num(I i2)::Num(I i1)::s),l))
                                  (pc',(h,s,l))

  | ifeq_step_jump : forall h m pc s l cmp i o,
      instructionAt m pc = Some (If0 cmp o) ->
      SemCompInt cmp (Int.toZ i) 0 ->
  (******************************************************************)
    NormalStep p m (pc,(h,(Num(I i)::s),l))
                                  (OFFSET.jump pc o,(h,s,l))

  | ifeq_step_continue : forall h m pc pc' s l cmp i o,
      instructionAt m pc = Some (If0 cmp o) ->
      next m pc = Some pc' ->
      ~ SemCompInt cmp (Int.toZ i) 0 ->
  (******************************************************************)
    NormalStep p m (pc,(h,(Num(I i)::s),l)) (pc',(h,s,l))

  | ifnull_step_jump : forall h m pc s l loc o cmp,
      instructionAt m pc = Some (Ifnull cmp o) ->
      SemCompRef cmp loc Null ->
  (******************************************************************)
    NormalStep p m (pc,(h,(loc::s),l))
                                  (OFFSET.jump pc o,(h,s,l))

  | ifnull_step_continue : forall h m pc pc' s l o loc cmp,
    instructionAt m pc = Some (Ifnull cmp o) ->
    next m pc = Some pc' ->
    ~ SemCompRef cmp loc Null ->
  (******************************************************************)
    NormalStep p m (pc,(h,(loc::s),l)) (pc',(h,s,l))


  | iinc_step : forall h m pc s l pc' x z i,
    instructionAt m pc = Some (Iinc x z) ->

    next m pc = Some pc' ->
    -2^7 <= z < 2^7 ->
    METHOD.valid_var m x ->
    LocalVar.get l x = Some (Num (I i)) ->

   NormalStep p m (pc,(h,s,l))
                (pc',(h,s,(LocalVar.update l x (Num (I (Int.add i (Int.const z)))))))

  | ineg_step : forall h m pc s l pc' i,
    instructionAt m pc = Some Ineg ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(Num (I i)::s),l)) (pc',(h,(Num (I (Int.neg i))::s),l))

  | instanceof1 : forall h m pc pc' s l loc t,

    instructionAt m pc = Some (Instanceof t) ->
    next m pc = Some pc' ->
    assign_compatible p h (Ref loc) (ReferenceType t) ->

   NormalStep p m (pc,(h,(Ref loc::s),l)) (pc',(h,(Num (I (Int.const 1))::s),l))


  | instanceof2 : forall h m pc pc' s l t v,

    instructionAt m pc = Some (Instanceof t) ->
    next m pc = Some pc' ->
    isReference v ->
    (~ assign_compatible p h v (ReferenceType t) \/ v=Null) ->

   NormalStep p m (pc,(h,(v::s),l)) (pc',(h,(Num (I (Int.const 0))::s),l))


  | invokestatic_native : forall m pc h args s l pc' h_res s' M v_res implem mid,

    instructionAt m pc = Some (Invokestatic mid) ->
    findMethod p mid = Some M ->
    METHOD.isNative M = true ->
    length args = length (METHODSIGNATURE.parameters (snd mid)) ->
    native_implementation M = Implem implem ->
    implem (h,args) (h_res,Normal v_res) ->
    next m pc = Some pc' -> 
    s' = cons_option v_res s ->

    NormalStep p m (pc,(h,args++s,l)) (pc',(h_res,s',l))

  | lookupswitch1 : forall h m pc s l def listkey i i' o',

    instructionAt m pc = Some (Lookupswitch def listkey) ->

    List.In (pair i' o')listkey ->
    i' = Int.toZ i ->

   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc o',(h,s,l))

  | lookupswitch2 : forall h m pc s l def listkey i,

    instructionAt m pc = Some (Lookupswitch def listkey) ->
    (forall i' o', List.In (pair i' o')listkey ->  i' <> Int.toZ i) ->

   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc def,(h,s,l))

  | new : forall h m pc pc' s l c loc h',

    instructionAt m pc = Some (New c) ->
    next m pc = Some pc' ->
    Heap.new h p (Heap.LocationObject c) = Some (pair loc h') ->

   NormalStep p m (pc,(h,s,l))
                      (pc',(h',(Ref loc::s),l))

  | newarray : forall h m pc pc' s l t i loc h',

    instructionAt m pc = Some (Newarray t) ->
    next m pc = Some pc' ->
    0 <= Int.toZ i -> 
    Heap.new h p (Heap.LocationArray i t) = Some (pair loc h') ->

   NormalStep p m (pc,(h,(Num (I i)::s),l))
                            (pc',(h',(Ref loc::s),l))

  | nop : forall h m pc pc' s l,

    instructionAt m pc = Some Nop ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,s,l)) (pc',(h,s,l))

  | pop : forall h m pc pc' s l v,

    instructionAt m pc = Some Pop ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v::s),l)) (pc',(h,s,l))

  | pop2 : forall h m pc pc' s l v1 v2,

    instructionAt m pc = Some Pop2 ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::s),l)) (pc',(h,s,l))

  | putfield : forall h m pc pc' s l f loc cn v,

    instructionAt m pc = Some (Putfield f) ->
    next m pc = Some pc' ->
    Heap.typeof h loc = Some (Heap.LocationObject cn) -> 
    defined_field p cn f ->
    assign_compatible p h v (FIELDSIGNATURE.type (snd f)) ->

   NormalStep p m(pc,(h,(v::(Ref loc)::s),l))
                           (pc',(Heap.update h (Heap.DynamicField loc f) v,s,l))

  | putstatic : forall h m pc pc' s l f v,

    instructionAt m pc = Some (Putstatic f) ->
    next m pc = Some pc' ->
    isStatic p f ->
    assign_compatible p h v (FIELDSIGNATURE.type (snd f)) ->
    
   NormalStep p m (pc,(h,(v::s),l))
                            (pc',(Heap.update h (Heap.StaticField f) v,s,l))

  | swap : forall h m pc pc' s l v1 v2,

    instructionAt m pc = Some Swap ->
    next m pc = Some pc' ->

   NormalStep p m (pc,(h,(v1::v2::s),l)) (pc',(h,(v2::v1::s),l))

  | tableswitch1 : forall h m pc s l i def low high list_offset,

    instructionAt m pc = Some (Tableswitch def low high list_offset) ->
    Z_of_nat (length list_offset) = (high - low + 1)%Z ->
    Int.toZ i < low \/ high < Int.toZ i ->
   
   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc def,(h,s,l))

  | tableswitch2 : forall h m pc s l n o i def low high list_offset,

    instructionAt m pc = Some (Tableswitch def low high list_offset) ->
    Z_of_nat (length list_offset) = (high - low + 1)%Z ->
    low <= Int.toZ i <= high ->
    Z_of_nat n = (Int.toZ i) - low ->
    nth_error list_offset n = Some o ->
   
   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc o,(h,s,l))

  | vaload : forall h m pc pc' s l loc val i length t k,

    instructionAt m pc = Some (Vaload k) ->
    next m pc = Some pc' ->
    Heap.typeof h loc = Some (Heap.LocationArray length t) ->
    compat_ArrayKind_type k t ->
    0 <= Int.toZ i < Int.toZ length ->
    Heap.get h (Heap.ArrayElement loc i) = Some val ->
    compat_ArrayKind_value k val ->

    NormalStep p m 
          (pc,(h,((Num (I i))::(Ref loc)::s),l))
          (pc',(h,(conv_for_stack val::s),l))

  | vastore : forall h m pc pc' s l loc val i length t k,

    instructionAt m pc = Some (Vastore k) ->
    next m pc = Some pc' ->
    Heap.typeof h loc = Some (Heap.LocationArray length t) ->
    assign_compatible p h val t ->
    0 <= Int.toZ i < Int.toZ length ->
    compat_ArrayKind_type k t ->
    compat_ArrayKind_value k val ->

    NormalStep p m
         (pc,(h,(val::(Num (I i))::(Ref loc)::s),l))
         (pc',(Heap.update h (Heap.ArrayElement loc i) (conv_for_array val t),s,l))

  | vload : forall h m pc pc' s l x val k,

    instructionAt m pc = Some (Vload k x) ->
    next m pc = Some pc' ->
    METHOD.valid_var m x ->
    LocalVar.get l x = Some val ->
    compat_ValKind_value k val -> 

    NormalStep p m (pc,(h,s,l)) (pc',(h,(val::s),l))
 
  | vstore : forall h m pc pc' s l l' x v k,

    instructionAt m pc = Some (Vstore k x) ->
    next m pc = Some pc' ->
    METHOD.valid_var m x ->
    l' = LocalVar.update l x v ->
    compat_ValKind_value k v ->

   NormalStep p m  (pc,(h,(v::s),l)) (pc',(h,s,l'))
.


  Inductive JVMExceptionStep  (p:Program) : Method -> IntraNormalState -> ShortClassName -> Prop :=
  | arraylength_NullPointerException : forall h m pc s l,

    instructionAt m pc = Some Arraylength ->

    JVMExceptionStep p m (pc,(h,(Null::s),l))  NullPointerException

  | athrow_NullPointerException : forall h m pc s l,

    instructionAt m pc = Some Athrow ->

    JVMExceptionStep p m (pc,(h,(Null::s),l)) NullPointerException

  | checkcast_ClassCastException : forall h m pc s l loc t,

    instructionAt m pc = Some (Checkcast t) ->
    ~ assign_compatible p h (Ref loc) (ReferenceType t) ->

    JVMExceptionStep p m (pc,(h,(Ref loc::s),l)) ClassCastException

  | getfield_NullPointerException : forall h m pc s l f ,

    instructionAt m pc = Some (Getfield f) ->

    JVMExceptionStep p m (pc,(h,(Null::s),l)) NullPointerException

 | ibinop_ArithmeticException : forall h m pc s l op i1 i2,

    instructionAt m pc = Some (Ibinop op) ->
    op = DivInt \/ op = RemInt ->
    Int.toZ i2 = 0 ->

    JVMExceptionStep p m (pc,(h,(Num (I i2)::Num (I i1)::s),l)) ArithmeticException

  | invokevirtual_NullPointerException : forall h m pc s l mid args,

    instructionAt m pc = Some (Invokevirtual mid) ->
    length args = length (METHODSIGNATURE.parameters (snd mid)) ->

    JVMExceptionStep p m (pc,(h,(args++Null::s),l)) NullPointerException

  | newarray_NegativeArraySizeException : forall h m pc s l t i,

    instructionAt m pc = Some (Newarray t) ->
    ~ 0 <= Int.toZ i ->

    JVMExceptionStep p m (pc,(h,(Num (I i)::s),l)) NegativeArraySizeException

  | putfield_NullPointerException : forall h m pc s l f v,

    instructionAt m pc = Some (Putfield f) ->
   
    JVMExceptionStep p m (pc,(h,(v::Null::s),l)) NullPointerException

  | vaload_NullPointerException : forall h m pc s l i k,

    instructionAt m pc = Some (Vaload k) ->

    JVMExceptionStep p m (pc,(h,((Num (I i))::Null::s),l)) NullPointerException

  | vaload_ArrayIndexOutOfBoundsException : forall h m pc s l loc i length t k,

    instructionAt m pc = Some (Vaload k) ->
    Heap.typeof h loc = Some (Heap.LocationArray length t) ->
    compat_ArrayKind_type k t ->
    ~ 0 <= Int.toZ i < Int.toZ length ->

    JVMExceptionStep p m (pc,(h,((Num (I i))::(Ref loc)::s),l)) ArrayIndexOutOfBoundsException

  | vastore_NullPointerException : forall h m pc s l val i k,

    instructionAt m pc = Some (Vastore k) ->
    compat_ArrayKind_value k val ->

    JVMExceptionStep p m (pc,(h,(val::(Num (I i))::Null::s),l)) NullPointerException

  | vastore_ArrayIndexOutOfBoundsException : forall h m pc s l loc val i t length k,

    instructionAt m pc = Some (Vastore k) ->
    Heap.typeof h loc = Some (Heap.LocationArray length t) ->
    ~ 0 <= Int.toZ i < Int.toZ length ->
    compat_ArrayKind_type k t ->
    compat_ArrayKind_value k val ->

    JVMExceptionStep p m (pc,(h,(val::(Num (I i))::(Ref loc)::s),l)) ArrayIndexOutOfBoundsException

  | vastore_ArrayStoreException : forall h m pc s l loc val i t k length,

    instructionAt m pc = Some (Vastore k) ->
    Heap.typeof h loc = Some (Heap.LocationArray length t) ->
    ~ assign_compatible p h val t ->
    0 <= Int.toZ i < Int.toZ length ->
    compat_ArrayKind_type k t ->
    compat_ArrayKind_value k val ->

    JVMExceptionStep p m (pc,(h,(val::(Num (I i))::(Ref loc)::s),l)) ArrayStoreException
.






  Inductive ExceptionStep (p:Program) : Method -> IntraNormalState -> IntraExceptionState -> Prop :=
  | athrow : forall h m pc s l loc cn,

    instructionAt m pc = Some Athrow ->
    Heap.typeof h loc = Some (Heap.LocationObject cn) ->
    subclass_name p cn javaLangThrowable ->
    ExceptionStep p m (pc,(h,(Ref loc::s),l)) (h,loc)


  | invokestatic_native_exception : forall m pc h args s l h_res M loc implem mid cn,

    instructionAt m pc = Some (Invokestatic mid) ->
    findMethod p mid = Some M ->
    METHOD.isNative M = true ->
    length args = length (METHODSIGNATURE.parameters (snd mid)) ->
    native_implementation M = Implem implem ->
    implem (h,args) (h_res,Exception loc) ->
    Heap.typeof h_res loc = Some (Heap.LocationObject cn) ->

    ExceptionStep p m (pc,(h,args++s,l)) (h_res,loc)

  | jvm_exception : forall h m pc s l h' loc (e:ShortClassName),

    JVMExceptionStep p m (pc,(h,s,l)) e ->
    Heap.new h p (Heap.LocationObject (javaLang,e)) = Some (loc,h') ->

    ExceptionStep p m (pc,(h,s,l)) (h',loc).







  Inductive CallStep (p:Program) : Method -> IntraNormalState -> InitCallState -> Prop :=
  | invokestatic : forall h m pc s l mid M args bM,

    instructionAt m pc = Some (Invokestatic mid) ->
    findMethod p mid = Some M ->
    METHOD.isNative M = false ->
    length args = length (METHODSIGNATURE.parameters (snd mid)) ->
    METHOD.body M = Some bM ->
    
    CallStep p m (pc,(h,(args++s),l)) (M,(s,(stack2localvar (args++s) (length args))))

  | invokevirtual : forall h m pc s l mid cn M args loc cl bM,

    instructionAt m pc = Some (Invokevirtual (cn,mid)) ->
    lookup p cn mid (pair cl M) ->
    Heap.typeof h loc = Some (Heap.LocationObject cn) ->
    length args = length (METHODSIGNATURE.parameters mid) ->
    METHOD.body M = Some bM ->
 
    CallStep p m (pc,(h,(args++(Ref loc)::s),l)) (M,(s,stack2localvar (args++(Ref loc)::s)  (1+(length args)))).



  Inductive ReturnStep (p:Program) : Method -> IntraNormalState -> ReturnState -> Prop :=
  | void_return : forall h m pc s l,

    instructionAt m pc = Some Return ->
    METHODSIGNATURE.result (METHOD.signature m) = None ->

    ReturnStep p m  (pc,(h,s,l)) (h, Normal None)
  | vreturn : forall h m pc s l val t k,

    instructionAt m pc = Some (Vreturn k) ->
    METHODSIGNATURE.result (METHOD.signature m) = Some t ->
    assign_compatible p h val t ->
    compat_ValKind_value k val ->

    ReturnStep p m (pc,(h,(val::s),l)) (h,Normal (Some val))
.






  Inductive call_and_return : Method -> IntraNormalState -> InitCallState -> IntraNormalState -> ReturnState -> IntraNormalState -> Prop :=
  | call_and_return_void : forall m pc h s l m' l' bm' h'' s' pc',
      next m pc = Some pc' -> 
      METHOD.body m' = Some bm' ->
      call_and_return
                 m
                 (pc,(h,s,l))
                 (m',(s',l'))
                  (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'', Normal None) 
                 (pc',(h'',s',l))
  | call_and_return_value : forall m pc h s l m' l' bm' h'' v s' pc',
      next m pc = Some pc' -> 
      METHOD.body m' = Some bm' ->
      call_and_return
                 m
                 (pc,(h,s,l))
                 (m',(s',l') )
                 (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'', Normal (Some v)) 
                 (pc',(h'',v::s',l)).





  Inductive call_and_return_exception : Method -> IntraNormalState -> InitCallState -> IntraNormalState -> ReturnState -> IntraExceptionState -> Prop :=
  | call_and_return_exception_def : forall m pc h s l m' l' bm' h'' s' loc,
      METHOD.body m' = Some bm' ->
      call_and_return_exception
                 m
                 (pc,(h,s,l))
                 (m',(s',l'))
                 (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'',Exception loc) 
                 (h'',loc).

  Inductive exec_intra (p:Program) (m:Method) : IntraNormalState -> IntraNormalState -> Prop :=
  | exec_intra_normal : forall s1 s2,
     NormalStep p m s1 s2 ->
     exec_intra p m s1 s2
  | exec_exception : forall pc1 h1 h2 loc2 s1 l1 pc',
   ExceptionStep p m (pc1,(h1,s1,l1)) (h2,loc2) ->
   CaughtException p m (pc1,h2,loc2) pc' ->
   exec_intra p m (pc1,(h1,s1,l1)) (pc',(h2,Ref loc2::OperandStack.empty,l1)).

  Inductive exec_return (p:Program) (m:Method) : IntraNormalState -> ReturnState -> Prop :=
  | exec_return_normal : forall s h ov,
     ReturnStep p m s (h,Normal ov) ->
     exec_return p m s (h,Normal ov)
  | exec_return_exception : forall pc1 h1 h2 loc2 s1 l1,
     ExceptionStep p m (pc1,(h1,s1,l1)) (h2,loc2) ->
     UnCaughtException  p m (pc1,h2,loc2) ->
     exec_return p m (pc1,(h1,s1,l1)) (h2,Exception loc2).

  Inductive exec_call (p:Program) (m:Method) :
   IntraNormalState -> ReturnState -> Method  -> IntraNormalState -> IntraNormalState+ReturnState -> Prop :=
 | exec_call_normal : forall m2 pc1 pc1' h1 s1 l1 os l2 h2 bm2 ov,
     CallStep p m (pc1,(h1,s1,l1 )) (m2,(os,l2)) ->
     METHOD.body m2 = Some bm2 ->
     next m pc1 = Some pc1' ->
     exec_call p m
        (pc1,(h1,s1,l1))
        (h2,Normal ov)
        m2
        (BYTECODEMETHOD.firstAddress bm2,(h1,OperandStack.empty,l2))
        (inl _ (pc1',(h2,cons_option ov os,l1)))

 | exec_call_caught : forall m2 pc1 pc1' h1 s1 l1 os l2 h2 loc bm2,
     CallStep p m (pc1,(h1,s1,l1 )) (m2,(os,l2)) ->
     METHOD.body m2 = Some bm2 ->
     CaughtException p m (pc1, h2, loc) pc1' ->
     exec_call p m
        (pc1,(h1,s1,l1))
        (h2,Exception loc)
        m2
        (BYTECODEMETHOD.firstAddress bm2,(h1,OperandStack.empty,l2))
        (inl _(pc1',(h2,Ref loc::nil,l1)))
 | exec_call_uncaught : forall m2 pc1 h1 s1 l1 os l2 h2 loc bm2,
     CallStep p m (pc1,(h1,s1,l1 )) (m2,(os,l2)) ->
     METHOD.body m2 = Some bm2 ->
     UnCaughtException p m (pc1, h2, loc)  ->
     exec_call p m
       (pc1,(h1,s1,l1))
       (h2,Exception loc)
       m2
       (BYTECODEMETHOD.firstAddress bm2,(h1,OperandStack.empty,l2))
       (inr _ (h2,Exception loc)).

 Inductive IntraStep (p:Program) : 
    Method -> IntraNormalState -> IntraNormalState + ReturnState -> Prop :=
  | IntraStep_res :forall m s ret,
     exec_return p m s ret ->
     IntraStep p m s (inr _ ret)
  | IntraStep_intra_step:forall m s1 s2,
     exec_intra p m s1 s2 ->
     IntraStep p m s1 (inl _ s2) 
  | IntraStep_call :forall m m' s1 s' ret' r,
     exec_call p m s1 ret' m' s' r ->
     TransStep_l (IntraStep p m') s' (inr _ ret') ->
     IntraStep p m s1 r.
 
 Definition IntraStepStar p m s r := TransStep_l (IntraStep p m) s r.

 Definition IntraStepStar_intra p m s s' := IntraStepStar p m s (inl _ s').

 Definition BigStep  p m s ret := IntraStepStar p m s (inr _ ret).

 Inductive ReachableStep (P:Program) : 
      (Method*IntraNormalState)->(Method*IntraNormalState) ->Prop :=
   | ReachableIntra : forall M s s', 
       IntraStep P M s (inl _ s') ->
       ReachableStep P (M,s) (M,s')
   | Reachable_invS : forall M pc h os l M' os' l' bm',
       CallStep P M (pc,(h,os,l)) (M',(os',l')) ->
       METHOD.body M' = Some bm' ->
       ReachableStep P (M, (pc,(h,os,l)))
         (M', (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))).

 Definition Reachable P M s s' := 
   exists M',  ClosReflTrans (ReachableStep P) (M,s) (M',s').
