(* <Insert License Here>

    $Id: BigStep.v 68 2006-02-02 15:06:27Z davidpichardie $ *)

(** Big step semantics.

 @author Benjamin Gregoire *)

Require Export Program.
Require Export Domain.

Set Implicit Arguments.
Definition cons_option (A:Set) (o:option A) (l:list A) := 
   match o with 
   | None => l
   | Some e => e::l
   end.
Unset Implicit Arguments.

Section TRANS.
 Set Implicit Arguments.
 Variable A B :Type.
 Variable R : A -> A + B -> Prop.

 Inductive TransStep_l : A -> A + B -> Prop :=
   | TransStep_l_refl : forall a,
       TransStep_l a (inl _ a)
   | TransStep_l_one : forall a r,
        R a r -> TransStep_l a r
   | TransStep_l_trans : forall a1 a2 r,
       R a1 (inl _ a2) -> 
       TransStep_l a2 r ->
       TransStep_l a1 r.

 Variable R' : A -> A -> Prop.

 Inductive ClosReflTrans : A -> A -> Prop :=
   | ClosReflTrans_refl : forall a, ClosReflTrans a a
   | ClosReflTrans_trans : forall a a' a'',
       R' a a' ->
       ClosReflTrans a' a'' ->
       ClosReflTrans a a''.

 Unset Implicit Arguments.

End TRANS.

Open Scope type_scope.

Module Type BIGSTEP.

 Declare Module Dom : SEMANTIC_DOMAIN.

 Load "BigStepLoad.v".

  Parameter IntraStepStar_ind : 
   forall (p:Program) 
    (P : Method -> IntraNormalState -> IntraNormalState + ReturnState -> Prop),
      (forall m s, P m s (inl _ s)) ->
      (forall m s r, exec_return p m s r -> P m s (inr _ r)) ->
      (forall m s s' , exec_intra p m s s' -> 
         forall r, IntraStepStar p m s' r -> P m s' r ->
         P m s r) ->
      (forall m s s' ret m' r, 
         exec_call p m s ret m' s' (inr _ r) ->
         IntraStepStar p m' s' (inr _ ret) ->
         P m' s' (inr _ ret) ->
         P m s (inr _ r)) ->
      (forall m s s' ret m' s'' r, 
         exec_call p m s ret m' s' (inl _ s'') ->
         IntraStepStar p m' s' (inr _ ret) -> P m' s' (inr _ ret) ->
         IntraStepStar p m s'' r -> P m s'' r ->
         P m s r) ->
   forall m s r, IntraStepStar p m s r -> P m s r.

  Parameter IntraStepStar_intra_ind : 
   forall (p:Program) 
    (P : Method -> IntraNormalState -> IntraNormalState -> Prop),
      (forall m s, P m s s) ->
      (forall m s s', exec_intra p m s s' -> 
         forall s'', IntraStepStar_intra p m s' s'' -> P m s' s'' ->
         P m s s'') ->
      (forall m s s1 ret m' s2 s3, 
         exec_call p m s ret m' s1 (inl _ s2) ->
         BigStep p m' s1 ret -> 
         IntraStepStar_intra p m s2 s3 -> P m s2 s3 ->
         P m s s3) ->
   forall m s s', IntraStepStar_intra p m s s' -> P m s s'.
  
  Parameter BigStep_ind : 
   forall (p:Program) 
    (P : Method -> IntraNormalState -> ReturnState -> Prop),
      (forall m s r, exec_return p m s r -> P m s r) ->
      (forall m s s' , exec_intra p m s s' -> 
         forall r, BigStep p m s' r -> P m s' r ->
         P m s r) ->
      (forall m s s' ret m' r, 
         exec_call p m s ret m' s' (inr _ r) ->
         BigStep p m' s' ret ->
         P m' s' ret ->
         P m s r) ->
      (forall m s s' ret m' s'' r, 
         exec_call p m s ret m' s' (inl _ s'') ->
         BigStep p m' s' ret -> P m' s' ret ->
         BigStep p m s'' r -> P m s'' r ->
         P m s r) ->
   forall m s r, BigStep p m s r -> P m s r.

  Parameter ReachableStar_ind : 
   forall (p:Program) 
    (P : (Method * IntraNormalState) -> (Method * IntraNormalState) -> Prop),
      (forall m s, P (m,s) (m,s)) ->
      (forall m s s', exec_intra p m s s' -> 
         forall m' s'', ClosReflTrans (ReachableStep p) (m,s') (m',s'') -> 
         P (m,s') (m',s'') ->
         P (m,s) (m',s'')) ->
      (forall m s s1 ret m' s2, 
         exec_call p m s ret m' s1 (inl _ s2) ->
         BigStep p m' s1 ret -> 
         forall m' s3, ClosReflTrans (ReachableStep p) (m,s2) (m',s3) -> 
         P (m,s2) (m',s3) ->
         P (m,s) (m',s3)) ->
      (forall m pc h os l m' os' l' bm',
       CallStep p m (pc,(h,os,l)) (m',(os',l')) ->
       METHOD.body m' = Some bm' ->
       forall m'' s'', 
       ClosReflTrans (ReachableStep p) 
         (m', (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l')))
         (m'',s'') ->
       P (m', (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l')))
         (m'',s'') ->
       P (m, (pc,(h,os,l))) (m'',s'')) ->
   forall ms ms', 
      ClosReflTrans (ReachableStep p) ms ms' -> P ms ms'.

End BIGSTEP.

Module BIGSTEP_INV (BS:BIGSTEP).

  Import BS Dom Prog.
 
  Definition EINS_inv p m pc s pc' s' oins :=   
    exists h, exists os, exists l, 
     match oins with
    | None => False
       | Some ins =>
         match ins with
         | Vaload k =>
           exists loc, exists val, 
               exists i, exists length, exists t,
             s = (h, Num(I i) :: Ref loc :: os, l) /\
             s' = (h, conv_for_stack val :: os, l) /\
             next m pc = Some pc' /\
             Heap.typeof h loc = 
               Some (Heap.LocationArray length t) /\
             0 <= Int.toZ i < Int.toZ length /\
             Heap.get h (Heap.ArrayElement loc i) = Some val /\
             compat_ArrayKind_type k t /\
             compat_ArrayKind_value k val
         | Vastore k =>
           exists loc, exists val,
               exists i, exists length, exists t,
             s = (h, val :: Num(I i) :: Ref loc :: os, l) /\
             s' = (Heap.update h (Heap.ArrayElement loc i) (conv_for_array val t), os, l) /\
             next m pc = Some pc' /\
             Heap.typeof h loc = Some (Heap.LocationArray length t) /\
             assign_compatible p h val t /\
             0 <= Int.toZ i < Int.toZ length /\
             compat_ArrayKind_type k t /\
             compat_ArrayKind_value k val
         | Aconst_null =>
             s = (h, os, l) /\
             s' = (h, Null :: os, l) /\
             next m pc = Some pc'  
         | Vload k x =>
           exists val,
             s = (h, os, l) /\
             s' = (h, val :: os, l) /\
             next m pc = Some pc' /\
             METHOD.valid_var m x /\
             LocalVar.get l x = Some val /\
             compat_ValKind_value k val
         | Arraylength =>
           exists loc, exists length, exists tp,
             s = (h, Ref loc :: os, l) /\
             s' = (h, Num(I length) :: os, l) /\
             next m pc = Some pc' /\
             Heap.typeof  h loc = Some (Heap.LocationArray length tp) 
         | Vstore k x =>
           exists l', exists v,
             s = (h, v :: os, l) /\
             s' = (h, os, l') /\ 
             next m pc = Some pc' /\
             METHOD.valid_var m x /\
             l' = LocalVar.update l x v /\
             compat_ValKind_value k v
         | Checkcast t =>
           exists val,
             s = (h, val :: os, l)/\ 
             s' = (h, val :: os, l)/\
             next m pc = Some pc' /\
             assign_compatible p h val (ReferenceType t) 
         | Const t z =>
             s = (h, os, l) /\
             s' = (h, Num(I(Int.const z)) :: os, l) /\
             next m pc = Some pc' /\
             ( (t=BYTE /\ -2^7 <= z < 2^7) \/ (t=SHORT /\ -2^15 <= z < 2^15)
               \/ (t=INT /\ -2^31 <= z < 2^31) )
         | Dup =>
           exists v, 
             s = (h, v :: os, l) /\
             s' = (h, v :: v :: os, l) /\
             next m pc = Some pc' 
         | Dup_x1 =>
           exists v1, exists v2, 
             s = (h, v1 :: v2 :: os, l) /\
             s' = (h, v1 :: v2 :: v1 :: os, l) /\
             next m pc = Some pc' 
         | Dup_x2 =>
           exists v1, exists v2, exists v3,
             s = (h, v1 :: v2 :: v3 :: os, l) /\
             s' = (h, v1 :: v2 :: v3 :: v1 :: os, l) /\
             next m pc = Some pc' 
         | Dup2 => 
	   exists v1, exists v2,
	     s = (h, v1 :: v2 :: os, l) /\
	     s' = (h, v1 :: v2 :: v1 :: v2 :: os, l) /\ 
             next m pc = Some pc' 
         | Dup2_x1 =>
           exists v1, exists v2, exists v3,
             s = (h, v1 :: v2 :: v3 :: os, l) /\
             s' = (h, v1 :: v2 :: v3 :: v1 :: v2 :: os, l) /\
             next m pc = Some pc' 
         | Dup2_x2 =>
           exists v1, exists v2, exists v3, exists v4,
             s = (h, v1 :: v2 :: v3 :: v4 :: os, l) /\
             s' = (h, v1 :: v2 :: v3 :: v4 :: v1 :: v2 :: os, l) /\
             next m pc = Some pc' 
         | Getfield f =>
	   exists loc, exists v, 
	    s = (h, Ref loc :: os, l) /\
            s' = (h, v :: os, l) /\    
	    next m pc = Some pc' /\
            Heap.get h (Heap.DynamicField loc f) = Some v
         | Getstatic f =>
           exists v,
             s = (h, os, l) /\
             s' = (h, v :: os, l) /\
             next m pc = Some pc' /\
             isStatic p f
         | Goto o =>
             s = (h, os, l) /\
             s' = (h, os, l) /\
             pc' = OFFSET.jump pc o
         | I2b =>
           exists i,
             s = (h, Num(I i) :: os, l) /\
             s' = (h, Num(I(b2i(i2b i))) ::os, l) /\
             next m pc = Some pc' 
         | I2s =>
           exists i,
             s = (h, Num(I i) :: os, l) /\
             s' = (h, Num(I(s2i(i2s i))) ::os, l) /\
             next m pc = Some pc' 
         | Ibinop op =>
           exists i1, exists i2,
             s = (h, Num(I i2) :: Num(I i1) :: os, l) /\
             s' = (h, Num(I(SemBinopInt op i1 i2)) :: os, l) /\
             next m pc = Some pc' /\
             (op = DivInt \/ op = RemInt -> ~ Int.toZ i2 = 0) 
         | If_acmp cmp o =>
           exists val2, exists val1,
             s = (h, val2 :: val1 :: os, l) /\
             s' = (h, os, l) /\
             ((SemCompRef cmp val1 val2 /\ pc' = OFFSET.jump pc o) \/
              (~SemCompRef cmp val1 val2 /\ next m pc = Some pc'))
         | If_icmp cmp o =>
	   exists i1, exists i2,
             s = (h, Num(I i2) :: Num(I i1) :: os, l) /\
             s' = (h, os, l) /\
             ((SemCompInt cmp (Int.toZ i1) (Int.toZ i2) /\ pc' = OFFSET.jump pc o) \/
              (~SemCompInt cmp (Int.toZ i1) (Int.toZ i2) /\ next m pc = Some pc'))
         | If0 cmp o =>
           exists i,
             s = (h, Num(I i) :: os, l) /\
             s' = (h, os, l) /\
             ((SemCompInt cmp (Int.toZ i) 0 /\ pc' = OFFSET.jump pc o) \/
              (~SemCompInt cmp (Int.toZ i) 0 /\ next m pc = Some pc'))
         | Ifnull cmp o =>
           exists loc,
             s = (h, loc :: os, l) /\
             s' = (h, os, l) /\
             ((SemCompRef cmp loc Null /\ pc' = OFFSET.jump pc o) \/
              (~SemCompRef cmp loc Null /\ next m pc = Some pc'))
         | Iinc x z =>
           exists i,
             s = (h, os, l) /\
             s' = (h, os, LocalVar.update l x (Num(I(Int.add i (Int.const z))))) /\
             next m pc = Some pc' /\
	     -2^7 <= z < 2^7 /\
             METHOD.valid_var m x /\
             LocalVar.get l x = Some (Num (I i)) 
         | Ineg =>
           exists i,
             s = (h, Num(I i) :: os, l) /\
             s' = (h, Num (I(Int.neg i)) :: os, l) /\
             next m pc = Some pc' 
         | Instanceof t =>
	   next m pc = Some pc' /\
           ((exists loc,
              s = (h, Ref loc :: os, l) /\
              s' = (h, Num(I(Int.const 1)) :: os, l) /\
              assign_compatible p h (Ref loc) (ReferenceType t)) \/
            (exists v,
	      s = (h, v :: os, l) /\
              s' = (h, Num(I(Int.const 0)) :: os, l) /\
              (~assign_compatible p h v (ReferenceType t)\/v=Null)))
         | Invokestatic mid =>
           exists args, exists h_res, exists os', exists M, exists v_res, 
	       exists implem, exists mid,
             s = (h, args++os, l) /\
             s' = (h_res, os', l) /\
             findMethod p mid = Some M /\
	     METHOD.isNative M = true /\
	     length args = length (METHODSIGNATURE.parameters (snd mid)) /\
	     native_implementation M = Implem implem /\
	     implem (h,args) (h_res,Normal v_res) /\
	     next m pc = Some pc' /\
	     os' = cons_option v_res os 
         | Lookupswitch def listkey =>
           exists i, 
             s = (h, Num (I i) :: os, l) /\
             s' = (h, os, l) /\
             ((exists i', exists o',
                List.In (pair i' o') listkey /\
                i' = Int.toZ i /\
                pc' = OFFSET.jump pc o') \/
              ((forall i' o', List.In (pair i' o')listkey ->  i' <> Int.toZ i) /\
               pc' = OFFSET.jump pc def))
         | New c =>
           exists loc, exists h',
             s = (h, os, l) /\
             s' = (h', Ref loc :: os, l) /\
             next m pc = Some pc' /\
             Heap.new h p (Heap.LocationObject c) = Some (pair loc h') 
         | Newarray t =>
           exists i, exists loc, exists h',
             s = (h, Num(I i) :: os, l) /\
             s' = (h', Ref loc :: os, l) /\
             next m pc = Some pc' /\
             0 <= Int.toZ i /\
             Heap.new h p (Heap.LocationArray i t) = Some (pair loc h') 
         | Nop =>
	     s = (h, os, l) /\
             s' = (h, os, l) /\
             next m pc = Some pc' 
         | Pop => 
	   exists v,
             s = (h, v :: os, l) /\
             s' = (h, os, l) /\
             next m pc = Some pc'  
         | Pop2 => 
	   exists v1, exists v2,
             s = (h, v1 :: v2 :: os, l) /\
             s' = (h, os, l) /\
             next m pc = Some pc'  
         | Putfield f =>
	   exists loc, exists cn, exists v,
             s = (h, v :: Ref loc :: os, l) /\
             s' = (Heap.update h (Heap.DynamicField loc f) v, os, l) /\  
             next m pc = Some pc' /\
	     Heap.typeof h loc = Some (Heap.LocationObject cn) /\
	     defined_field p cn f /\
	     assign_compatible p h v (FIELDSIGNATURE.type (snd f)) 
         | Putstatic f =>
           exists v, 
	     s = (h, v :: os, l) /\
             s' = (Heap.update h (Heap.StaticField f) v, os, l) /\
	     next m pc = Some pc' /\
	     isStatic p f /\
	     assign_compatible p h v (FIELDSIGNATURE.type (snd f))
	 | Swap =>
	   exists v1, exists v2,
	     s = (h, v1 :: v2 :: os, l) /\
	     s' = (h, v2 :: v1 :: os, l) /\ 
             next m pc = Some pc'
         | Tableswitch def low high list_offset =>
	   exists i, 
	     s = (h, Num(I i) :: os, l) /\
	     s' = (h, os, l) /\
	     Z_of_nat (length list_offset) = (high - low + 1)%Z /\
	     ((Int.toZ i < low \/ high < Int.toZ i) \/
	      (low <= Int.toZ i <= high /\
               exists n, exists o,
                 Z_of_nat n = (Int.toZ i) - low /\
                 nth_error list_offset n = Some o))
         | _ => False
         end
       end.

  Definition EIJVMES_inv p s c oins := 
    exists h, exists os, exists l:LocalVar.t, 
    match oins with
    | None => False
    | Some ins =>
      match ins with
      | Vaload k =>
        exists i,
          ((s = (h, Num(I i) :: Null :: os, l) /\ 
            c = NullPointerException) \/
           (exists loc, exists t, exists length,
            s = (h, Num(I i) :: Ref loc :: os, l) /\
            c = ArrayIndexOutOfBoundsException /\
            Heap.typeof h loc = 
              Some (Heap.LocationArray length t) /\
            ~ 0 <= Int.toZ i < Int.toZ length))
      | Vastore k =>
        exists i, exists val,
	  ((s = (h, val :: Num(I i) :: Null :: os,l) /\
            c = NullPointerException) \/
           (exists loc, exists t, exists length,
            s = (h, val :: Num(I i) :: Ref loc :: os, l) /\
            Heap.typeof h loc = Some (Heap.LocationArray length t) /\
            compat_ArrayKind_type k t /\
            compat_ArrayKind_value k val /\
            ((~ 0 <= Int.toZ i < Int.toZ length /\
              c = ArrayIndexOutOfBoundsException) \/
             (0 <= Int.toZ i < Int.toZ length /\
              ~ assign_compatible p h val t /\
              c = ArrayStoreException))))
      | Arraylength =>
	  s = (h, Null :: os, l) /\ 
          c = NullPointerException
      | Athrow =>
          s = (h, Null :: os, l) /\ 
          c = NullPointerException
      | Checkcast t =>
	exists loc, 
	  s = (h, Ref loc :: os, l) /\
          c = ClassCastException /\ 
          ~ assign_compatible p h (Ref loc) (ReferenceType t)
      | Getfield f =>
          s = (h, Null :: os, l) /\
          c = NullPointerException
      | Ibinop op =>
	exists i1, exists i2, 
          s = (h, Num(I i2) :: Num(I i1) :: os, l) /\
          c = ArithmeticException /\ 
          (op = DivInt \/ op = RemInt) /\
          Int.toZ i2 = 0
      | Invokevirtual mid =>
	exists args,
	  s = (h, args++Null :: os, l) /\
	  c = NullPointerException /\
	  length args = length (METHODSIGNATURE.parameters (snd mid))
      | Newarray t =>
	exists i,
	  s = (h, Num(I i) :: os, l) /\
	  c = NegativeArraySizeException /\
	  ~ 0 <= Int.toZ i
      | Putfield f =>
	exists v,
	  s = (h, v :: Null :: os, l) /\
	  c = NullPointerException
      | _ => False
      end
    end.

  Lemma NormalStep_inv : forall p m pc s pc' s',
      NormalStep p m (pc,s) (pc',s') ->
      EINS_inv p m pc s pc' s' (instructionAt m pc).
  Proof.
     intros p m pc s pc' s' Hel;inversion Hel;clear Hel;subst;
     match goal with
      | [H:instructionAt m pc = _ |- _] => rewrite H;unfold EINS_inv;simpl
      | _=> idtac
      end;eauto 30.
   Qed.

  Lemma ElementaryIntraJVMExceptionStep_inv : forall p m pc s c,
      JVMExceptionStep p m (pc,s) c ->
      EIJVMES_inv p s c (instructionAt m pc).
  Proof.
     intros p m pc s c Hel;inversion Hel;clear Hel;subst;
     match goal with
      | [H:instructionAt m pc = _ |- _] => rewrite H;unfold EIJVMES_inv;simpl
      | _=> idtac
      end;eauto 30.
   Qed.

  Ltac destrhyp t := 
    match type of t with
    | @ex Method _ =>
      let x := fresh "m" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex OperandStack.t _ =>
      let x := fresh "os" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex LocalVar.t _ => 
      let x := fresh "l" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Heap.t _ => 
      let x := fresh "h" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex PC _ => 
      let x := fresh "pc" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Location _ => 
      let x := fresh "loc" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex value _ => 
      let x := fresh "val" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex refType _ => 
      let x := fresh "rt" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex primitiveType _ => 
      let x := fresh "pt" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Int.t _ => 
      let x := fresh "i" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Byte.t _ => 
      let x := fresh "b" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Short.t _ => 
      let x := fresh "sh" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex Z _ => 
      let x := fresh "z" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | @ex _ _ => 
      let x := fresh "x" in
      let H := fresh "H" in
      (destruct t as (x,H);destrhyp H)
    | (_ <= _ < _) => idtac
    | (_ <= _ <= _) => idtac
    | (_ < _ <= _) => idtac
    | (_ < _ < _) => idtac
    | ?A /\ ?B =>
      let H1 := fresh "H" in
      let H2 := fresh "H" in
      (destruct t as (H1,H2);destrhyp H1;destrhyp H2)
    | _ => idtac
    end.

 Ltac invertNormal H :=
   let H1 := fresh "H" in
   (assert (H1 := NormalStep_inv _ _ _ _ _ _ H);
    match type of H with
    | NormalStep _ ?m (?pc,_) _ =>
      match goal with
      | [HH: instructionAt m pc = _ |- _ ] =>
         rewrite HH in H1;
         unfold EINS_inv in H1;simpl in H1;destrhyp H1
      | _ => idtac
      end
    end) . 

 Ltac invertExcep H :=
   let H1 := fresh "H" in
   (assert (H1 := ElementaryIntraJVMExceptionStep_inv _ _ _ _ _ H);
    match type of H with
    | JVMExceptionStep _ ?m (?pc,_) _ =>
      match goal with
      | [HH: instructionAt m pc = _ |- _ ] =>
         rewrite HH in H1;
         unfold EIJVMES_inv in H1;simpl in H1;destrhyp H1
      | _ => idtac
      end
    end) . 
 
End BIGSTEP_INV.



