Add LoadPath "../Library".
Require Export List.
Require Export ZArith.
Require Export Program.
Require Export Relation_Operators.
Require Export EqBoolAux.
Require Export PosAux.

Section findList.
  Variable A B : Set.
  Variable getKey : A -> B.
  Variable Beq : B -> B -> bool.
  Variable Beq_spec : forall x y, if Beq x y then x = y else x <> y.

  Fixpoint find (b:B) (l:list A) {struct l} : option A :=
    match l with
      nil => None
    | a::q => if Beq (getKey a) b then Some a
              else find b q
    end.
  Lemma find_getKey : forall b l a,
    find b l = Some a -> b = getKey a.
  Proof.
    induction l; simpl.
    intros; discriminate.
    intro; generalize (Beq_spec (getKey a) b);destruct (Beq (getKey a) b); intros.
    injection H0;intros;subst;trivial.    auto.
  Qed.

  Fixpoint add (a:A) (l:list A) {struct l} : list A :=
    match l with
      | nil => a::nil
      | x::q => if Beq (getKey a) (getKey x) then x::q else x::(add a q)
    end.
  
  Fixpoint clean_aux (l acc:list A) {struct l} : list A :=
    match l with
      | nil => acc
      | a::q => clean_aux q (add a acc)
    end.

  Definition clean l := clean_aux l nil.

  Definition Cleaned' (l:list A) : Prop :=
    forall a1 a2, In a1 l -> In a2 l -> getKey a1 = getKey a2 -> a1 = a2.

  Inductive Cleaned : list A -> Prop :=
  | Cleaned_nil : Cleaned nil
  | Cleaned_cons : forall a l,
      (forall x, In x l -> getKey a <> getKey x) ->
      Cleaned l ->
      Cleaned (a::l).

  Lemma Cleaned_Cleaned' : forall l, Cleaned l -> Cleaned' l.
  Proof.
    induction 1; repeat intro.
    inversion H.
    destruct H1; destruct H2; subst; intuition.
    elim H with a2; auto.
    elim H with a1; auto.
  Qed.
    
  Lemma in_add : forall x a l, In x (add a l) -> x=a \/ In x l.
  Proof.
    induction l; simpl; intros.
    intuition.
    destruct (Beq (getKey a) (getKey a0)); simpl in H; intuition.
  Qed.

  Lemma add_Cleaned : forall a l, Cleaned l -> Cleaned (add a l).
  Proof.
    induction 1; simpl; try constructor.
    intuition.
    constructor.
    generalize (Beq_spec (getKey a) (getKey a0));
      destruct (Beq (getKey a) (getKey a0)); intros.
    constructor; auto.
    constructor; auto.
    intros.
    elim in_add with (1:=H2); intros; subst; auto.
  Qed.

  Lemma clean_aux_Cleaned : forall l acc,
    Cleaned acc -> Cleaned (clean_aux l acc).
  Proof.
    induction l; simpl; intros; auto.
    apply IHl; apply add_Cleaned; auto.
  Qed.

  Lemma clean_Cleaned : forall l, Cleaned (clean l).
  Proof.
    unfold clean; intros; apply clean_aux_Cleaned; constructor.
  Qed.

  Lemma in_clean_aux : forall a l acc, In a (clean_aux l acc) -> In a l \/ In a acc.
  Proof.
    induction l; simpl in *; intros; auto.
    elim IHl with (1:=H); intros; auto.
    elim in_add with (1:=H0); intuition.
  Qed.

  Lemma in_clean : forall a l, In a (clean l) -> In a l.
  Proof.
    unfold clean; intros.
    elim in_clean_aux with (1:=H); intuition.
    inversion_clear H0.
  Qed.

  Lemma in_add2 : forall a x l, In x l -> In x (add a l).
  Proof.
    induction l; simpl; intuition eauto.
    generalize (Beq_spec (getKey a) (getKey a0));
      destruct (Beq (getKey a) (getKey a0)); intros.
    subst; left; auto.
    subst; left; auto.
    generalize (Beq_spec (getKey a) (getKey a0));
      destruct (Beq (getKey a) (getKey a0)); intros.
    right; auto.
    right; auto.
  Qed.

  Lemma in_add3 : forall a l, exists a', In a' (add a l) /\ getKey a = getKey a'.
  Proof.
    induction l; simpl; intuition eauto.
    destruct IHl as [a' H].
    generalize (Beq_spec (getKey a) (getKey a0));
      destruct (Beq (getKey a) (getKey a0)); intros.
    exists a0; auto with datatypes.
    destruct H.
    exists a'; auto with datatypes.
  Qed.

  Lemma in_clean_aux2 : forall a l acc, 
    In a acc -> In a (clean_aux l acc).
  Proof.
    induction l; simpl; intuition.
    apply IHl.
    apply in_add2; auto.
  Qed.

  Lemma in_clean_aux3 : forall a l acc, 
    In a l -> exists a', In a' (clean_aux l acc) /\ getKey a = getKey a'.
  Proof.
    induction l; simpl; intuition.
    subst.
    elim in_add3 with a acc; intros a' [H1 H2].
    exists a'; split; auto.
    apply in_clean_aux2; auto.    
  Qed.

  Lemma in_cleaned_list_find_some : forall l a, Cleaned l ->
    In a l -> find (getKey a) l = Some a.
  Proof.
    induction l; simpl; intuition.
    subst.
    generalize (Beq_spec (getKey a0) (getKey a0)); 
      destruct (Beq (getKey a0) (getKey a0)); intuition eauto.
    generalize (Beq_spec (getKey a) (getKey a0)); 
      destruct (Beq (getKey a) (getKey a0)); intuition.
    inversion_clear H.    
    elim H2 with a0; auto.
    inversion_clear H; auto.
  Qed.
  
  Definition In' (a:A) (l:list A) :=
    exists a', In a' l /\ getKey a = getKey a'.

    Ltac Beq_case x y :=
    generalize (Beq_spec x y); 
      destruct (Beq x y); intuition eauto.

  Lemma find_add_none : forall x a l,
    find x (add a l) = None -> find x l = None.
  Proof.
    induction l; simpl; intros; auto.
    Beq_case (getKey a) (getKey a0).
    simpl in H.
    Beq_case (getKey a0) x.
  Qed.

  Lemma find_add_some : forall a l,
    find (getKey a) (add a l) <> None.
  Proof.
    induction l; simpl; intros.
    Beq_case (getKey a) (getKey a).
    discriminate.
    Beq_case (getKey a) (getKey a0); simpl in *.
    Beq_case (getKey a0) (getKey a).
    discriminate.
    Beq_case (getKey a0) (getKey a).
  Qed.

  Lemma find_some_add_new : forall a acc ,
    find (getKey a) acc = None -> find (getKey a) (add a acc) = Some a.
  Proof.
    induction acc; simpl; intros.
    Beq_case (getKey a) (getKey a).
    Beq_case (getKey a0) (getKey a); try discriminate.
    Beq_case (getKey a) (getKey a0).
    elim H0; auto.
    simpl.
    Beq_case (getKey a0) (getKey a).
  Qed.

  Lemma find_some_add_old : forall x a l,
    find x l <> None -> find x (add a l) = find x l.
  Proof.
    induction l; simpl; intros.
    Beq_case (getKey a) x.
    Beq_case (getKey a) (getKey a0).
    simpl.
    Beq_case (getKey a0) x.
  Qed.

  Lemma find_some_or : forall x a l,
    find a (add x l) = find a l
    \/ 
    find a (add x l) = Some x /\ find a l = None.
  Proof.
    induction l; simpl; intros.
    Beq_case (getKey x) a.
    Beq_case (getKey x) (getKey a0); simpl.
    Beq_case (getKey a0) a.
    Beq_case (getKey a0) a.
  Qed.

  Lemma find_clean_same_aux : forall a l acc,
    (find a acc = None /\ find a (clean_aux l acc) = find a l)
    \/
    (find a acc <> None /\ find a (clean_aux l acc) = find a acc).
  Proof.
    induction l; simpl; intros.
    destruct (find a acc); auto.
    right; split; auto; discriminate.
    elim IHl with (add a0 acc); intros [H H0].
    left; split.
    apply find_add_none with a0; auto.
    rewrite H0.
    Beq_case (getKey a0) a.
    subst; elim find_add_some with a0 acc; auto.
    rewrite H0.
    elim find_some_or with a0 a acc; intros HH.
    right; rewrite <- HH; auto.
    destruct HH; left.
    split; auto.
    Beq_case (getKey a0) a.
    elim H3; symmetry; eapply find_getKey; eauto.
  Qed.

  Lemma find_clean_same : forall a l, find a (clean l) = find a l.
  Proof.
    unfold clean; intros.
    elim find_clean_same_aux with a l (@nil A); intros [H1 H2]; auto.
    elim H1; reflexivity.
  Qed.

  Lemma in_clean_find_some : forall l a, 
    In a (clean l) -> find (getKey a) l = Some a.
  Proof.
    intros.
    rewrite <- find_clean_same.
    apply in_cleaned_list_find_some; auto.
    apply (clean_Cleaned l).
  Qed.

  Lemma find_some_in_list : forall l b a,
    find b l = Some a -> In a l.
  Proof.
    induction l; simpl; intros.
    discriminate H.
    destruct (Beq (getKey a) b); eauto.
    inversion H; auto.
  Qed.

  Lemma find_some_in_clean : forall l b a,
    find b l = Some a -> In a (clean l).
  Proof.
    intros.
    rewrite <- find_clean_same in H.
    apply find_some_in_list with b; auto.
  Qed.

End findList.

Module Make <: PROGRAM.

  Definition eq_dec (A:Set) := forall x y:A, x=y \/~x=y.

  Definition Var := N. 
  Lemma Var_eq_dec : eq_dec Var.
  Proof. exact (Aeq_dec _ Neq Neq_spec). Qed.
  Definition Var_toN : Var -> nat := nat_of_N.
  Definition N_toVar : nat -> Var := N_of_nat.
  Lemma Var_toN_bij1 : forall v, N_toVar (Var_toN v) = v.
  Proof. exact nat_of_N_bij2. Qed.
  Lemma Var_toN_bij2 : forall n, Var_toN (N_toVar n) = n.
  Proof. exact nat_of_N_bij1. Qed.

  Definition PC : Set := N.
  Definition PC_eq := Neq.
  Definition PC_eq_spec := Neq_spec.
  Lemma PC_eq_dec : eq_dec PC.
  Proof. exact Var_eq_dec. Qed.

  Definition PackageName : Set := positive.
  Definition ShortClassName : Set := positive.
  Definition ShortMethodName : Set := positive.
  Definition ShortFieldName : Set := positive.
  Open Local Scope type_scope.
  Definition ClassName := PackageName * ShortClassName.
  Definition InterfaceName := PackageName * ShortClassName.
  Definition MethodName := ClassName * ShortMethodName.
  Definition FieldName := ClassName * ShortFieldName.

 Definition eqClassName : ClassName -> ClassName -> bool := prod_eq _ Peq _ Peq.
  Lemma eqClassName_spec : forall x y, if eqClassName x y then x = y else x <> y.
  Proof. exact (prod_eq_spec _ Peq Peq_spec _ Peq Peq_spec). Qed.
  Lemma ClassName_eq_dec : eq_dec ClassName.
  Proof. exact (Aeq_dec _ eqClassName eqClassName_spec). Qed.

  Definition eqInterfaceName : InterfaceName ->InterfaceName -> bool := 
      prod_eq _ Peq _ Peq.
  Lemma eqInterfaceName_spec : forall x y, if eqInterfaceName x y then x = y else x <> y.
  Proof. exact (prod_eq_spec _ Peq Peq_spec _ Peq Peq_spec). Qed.
  Lemma InterfaceName_eq_dec : eq_dec InterfaceName.
  Proof. exact (Aeq_dec _ eqClassName eqClassName_spec). Qed.

  Definition eqMethodName : MethodName -> MethodName -> bool := 
    prod_eq _ eqClassName _ Peq.
  Lemma eqMethodName_spec : forall x y, if eqMethodName x y then x = y else x <> y.
  Proof. exact (prod_eq_spec _ eqClassName eqClassName_spec _ Peq Peq_spec). Qed.
  Lemma MethodName_eq_dec : eq_dec MethodName.
  Proof. exact (Aeq_dec _ eqMethodName eqMethodName_spec). Qed.

  Definition eqFieldName : FieldName -> FieldName -> bool :=
     prod_eq _ eqClassName _ Peq.
  Lemma eqFieldName_spec : forall x y, if eqFieldName x y then x = y else x <> y.
  Proof. exact (prod_eq_spec _ eqClassName eqClassName_spec _ Peq Peq_spec). Qed.
  Lemma FieldName_eq_dec : eq_dec FieldName.
  Proof. exact (Aeq_dec _ eqFieldName eqFieldName_spec). Qed.

  Open Scope positive_scope.
  (* IMPORTANT CONSTANT CONVENTIONS FOR PARSER !! *)
  Definition javaLang : PackageName := 1.
  Definition EmptyPackageName : PackageName := 2.
  Definition javaIo: PackageName := 3.
  
  Definition object : ShortClassName := 1.
  Definition exception : ShortClassName := 2.
  Definition RuntimeException : ShortClassName := 3.
  Definition NullPointerException : ShortClassName := 4.
  Definition ArrayIndexOutOfBoundsException : ShortClassName := 5.
  Definition ArrayStoreException : ShortClassName := 6.
  Definition NegativeArraySizeException : ShortClassName := 7.
  Definition ClassCastException : ShortClassName := 8.
  Definition ArithmeticException : ShortClassName := 9.
  Definition throwable : ShortClassName := 10.
  Definition cloneable: ShortClassName := 11.
  Definition serializable: ShortClassName := 12.
  
  Inductive Visibility : Set :=
    Package | Protected | Private | Public.

 Definition eqVisibility x y := 
    match x, y with 
    | Package, Package => true
    | Protected, Protected => true
    | Private, Private => true
    | Public, Public => true
    | _, _ => false
    end.
  Lemma eqVisibility_spec : forall x y, if eqVisibility x y then x = y else x <> y.
  Proof. destruct x;destruct y;simpl;trivial;intro; discriminate. Qed.
  Lemma Visibility_eq_dec : eq_dec Visibility.
  Proof. exact (Aeq_dec _ eqVisibility eqVisibility_spec). Qed.

  Inductive type : Set :=
      | ReferenceType (rt : refType)
      | PrimitiveType (pt: primitiveType)
  with refType :Set := 
      | ArrayType (typ:type) 
      | ClassType  (ct:ClassName)
      | InterfaceType (it:InterfaceName)
  with  primitiveType : Set := 
      | BOOLEAN  | BYTE | SHORT | INT.

  Scheme type_strong_rec := Induction for type Sort Set
    with refType_strong_rec := Induction for refType Sort Set.
 
  Scheme type_strong_ind := Induction for type Sort Prop
    with refType_strong_ind := Induction for refType Sort Prop.
  Definition eq_primitiveType x y :=
    match x, y with
    | BOOLEAN, BOOLEAN => true
    | BYTE, BYTE => true
    | SHORT, SHORT => true
    | INT, INT => true 
    | _, _ => false
    end.
  Lemma eq_primitiveType_spec : forall x y, if eq_primitiveType x y then x = y else x <> y.
  Proof.
   destruct x;destruct y;simpl;trivial; intro;discriminate.
  Qed.
  Lemma primitiveType_dec : eq_dec primitiveType.
  Proof.  exact (Aeq_dec _ eq_primitiveType eq_primitiveType_spec). Qed.

  Fixpoint eq_type (t1 t2:type) {struct t1} : bool := 
    match t1, t2 with 
    | ReferenceType rt1, ReferenceType rt2 => eq_reftype rt1 rt2
    | PrimitiveType pt1, PrimitiveType pt2 => eq_primitiveType pt1 pt2
    | _, _ => false
    end
  with eq_reftype (rt1 rt2: refType) {struct rt1} : bool := 
    match rt1, rt2 with
    | ArrayType t1, ArrayType t2 => eq_type t1 t2
    | ClassType cn1, ClassType cn2 => eqClassName cn1 cn2
    | InterfaceType in1, InterfaceType in2 => eqInterfaceName in1 in2
    |_, _ => false
    end.

  Lemma eq_type_spec : forall t1 t2, if eq_type t1 t2 then t1 = t2 else t1 <> t2.
  Proof.
   induction t1 using type_strong_ind with 
        (P0:=
          fun rt1 => forall rt2, if eq_reftype rt1 rt2 then rt1 = rt2 else rt1 <> rt2);intros.
   destruct t2;simpl;try (intro;discriminate;fail).
   assert (UU:=IHt1 rt0);destruct (eq_reftype rt rt0);subst;trivial.
   intro H;injection H;auto.
   destruct t2;simpl;try (intro;discriminate;fail).
   assert (UU := eq_primitiveType_spec pt pt0);destruct (eq_primitiveType pt pt0);subst;trivial.
   intro H;injection H;auto.
   destruct rt2;simpl;try (intro H;discriminate H).
   assert (UU:=IHt1 typ);destruct (eq_type t1 typ);subst;trivial.
   intro H;injection H;auto.
   destruct rt2;simpl;intros;try (intro;discriminate;fail).
   assert (UU := eqClassName_spec ct ct0);destruct (eqClassName ct ct0);subst;trivial.
   intro H;injection H;auto.
   destruct rt2;simpl;intros;try (intro;discriminate;fail).
   assert (UU := eqInterfaceName_spec it it0);destruct (eqInterfaceName it it0);subst;trivial.
   intro H;injection H;auto.
  Qed.
  Lemma type_dec : eq_dec type.
  Proof. exact (Aeq_dec _ eq_type eq_type_spec). Qed.

  Inductive CompInt : Set := EqInt | NeInt | LtInt | LeInt | GtInt | GeInt.
  Inductive CompRef : Set := EqRef | NeRef.

  Inductive BinopInt : Set := AddInt | AndInt | DivInt | MulInt | OrInt | RemInt 
                            | ShlInt | ShrInt | SubInt | UshrInt | XorInt.

  Module Type OFFSET_TYPE.
    Parameter t : Set.
    Parameter jump : PC -> t -> PC.
  End OFFSET_TYPE.
  Module OFFSET <: OFFSET_TYPE.
    Definition t := Z.
    Definition jump (pc:PC) (ofs:t) : PC := Zabs_N (Zplus (Z_of_N pc) ofs).
  End OFFSET.

  Module FIELDSIGNATURE.
    Record t :Set := {
      name : ShortFieldName;
      type : type
    }.
    Definition eq_t (x y : t) := 
         let (n1,t1) := x in
         let (n2,t2) := y in
         if Peq n1 n2 then eq_type t1 t2 else false.
    Lemma eq_t_spec : forall x y, if eq_t x y then x = y else x <> y.
    Proof. 
      intros (n1,t1) (n2,t2);simpl;generalize (Peq_spec n1 n2);
       destruct (Peq n1 n2);intros.
      generalize (eq_type_spec t1 t2);destruct (eq_type t1 t2);intros;subst;trivial.
      intro H;injection H;auto.
      intro H1;injection H1;auto.
    Qed.
    Lemma eq_dec : eq_dec t.
    Proof. exact (Aeq_dec _ eq_t eq_t_spec). Qed.
  End FIELDSIGNATURE.
  Definition ShortFieldSignature := FIELDSIGNATURE.t.
  Definition FieldSignature := (ClassName*ShortFieldSignature)%type.

  Module Type FIELDSIGNATURE_TYPE.
    Parameter name : ShortFieldSignature  -> ShortFieldName.
    Parameter type : ShortFieldSignature -> type.
    Parameter eq_dec : forall f1 f2:ShortFieldSignature,  f1=f2 \/~f1=f2.
  End FIELDSIGNATURE_TYPE.


  Module METHODSIGNATURE.
    Record t :Set := {
      name : ShortMethodName;
      parameters : list type;
      result : option type
    }.
     Definition eq_t (x y : t) :=
      let (n1,p1,r1) := x in
      let (n2,p2,r2) := y in
      if Peq n1 n2 then 
       if list_eq _ eq_type p1 p2 then opt_eq _ eq_type r1 r2 
       else false
      else false.
    Lemma eq_t_spec : forall x y, if eq_t x y then x = y else x <> y.
    Proof.
     intros (n1,p1,r1) (n2,p2,r2);simpl;generalize (Peq_spec n1 n2);
       destruct (Peq n1 n2);intros.
      generalize (list_eq_spec _ eq_type eq_type_spec p1 p2);
       destruct (list_eq _ eq_type p1 p2);intros.
      generalize (opt_eq_spec _ eq_type eq_type_spec r1 r2);
       destruct (opt_eq _ eq_type r1 r2);intros. subst;trivial.
      intro UU;injection UU;auto.
      intro UU;injection UU;auto.
      intro H1;injection H1;auto.
    Qed.
    Lemma eq_dec : eq_dec t.
    Proof. exact (Aeq_dec _ eq_t eq_t_spec). Qed.
  End METHODSIGNATURE.
  Definition ShortMethodSignature := METHODSIGNATURE.t.
  Definition MethodSignature := (ClassName*METHODSIGNATURE.t)%type.

  Module Type METHODSIGNATURE_TYPE.
    Parameter name : ShortMethodSignature -> ShortMethodName.
    Parameter parameters : ShortMethodSignature -> list type.
    Parameter result : ShortMethodSignature -> option type.
    Parameter eq_dec : forall mid1 mid2:ShortMethodSignature,  mid1=mid2 \/~mid1=mid2.
  End METHODSIGNATURE_TYPE.

  Inductive ArrayKind : Set :=
    | Aarray
    | Iarray
    | Barray
    | Sarray.
    
  Inductive ValKind : Set :=
    | Aval
    | Ival.

  Inductive Instruction : Set :=
   | Aconst_null
   | Arraylength 
   | Athrow
   | Checkcast (t:refType)
   | Const (t:primitiveType) (z:Z)
   | Dup
   | Dup_x1
   | Dup_x2
   | Dup2
   | Dup2_x1
   | Dup2_x2
   | Getfield (f:FieldSignature)
   | Getstatic  (f:FieldSignature) 
   | Goto (o:OFFSET.t)
   | I2b
   | I2s
   | Ibinop (op:BinopInt)
   | If_acmp (cmp:CompRef) (o:OFFSET.t)
   | If_icmp (cmp:CompInt) (o:OFFSET.t) 
   | If0 (cmp:CompInt) (o:OFFSET.t)
   | Ifnull (cmp:CompRef) (o:OFFSET.t)
   | Iinc (x:Var) (z:Z)
   | Ineg 
   | Instanceof (t:refType) 
   | Invokeinterface (m:MethodSignature)
   | Invokespecial (m:MethodSignature)
   | Invokestatic (m:MethodSignature)
   | Invokevirtual (m:MethodSignature)
   | Lookupswitch (def:OFFSET.t) (l:list (Z*OFFSET.t)) 
   | New (c:ClassName)
   | Newarray (t:type)
   | Nop
   | Pop
   | Pop2
   | Putfield (f:FieldSignature)
   | Putstatic (f:FieldSignature)
   | Return
   | Swap 
   | Tableswitch (def:OFFSET.t) (low high:Z) (l:list OFFSET.t)
   | Vaload (k:ArrayKind) 
   | Vastore (k:ArrayKind)
   | Vload (k:ValKind) (x:Var)
   | Vreturn (k:ValKind)
   | Vstore (k:ValKind) (x:Var).

  Module FIELD.
    Inductive value : Set :=
    | Int (v:Z)
    | NULL 
    | UNDEF.
  Definition eq_value x y := 
     match x, y with
     | Int z1, Int z2 => Zeq_bool z1 z2
     | NULL, NULL => true
     | UNDEF, UNDEF => true
     | _, _ => false
     end.
    Lemma eq_value_spec : forall x y, if eq_value x y then x = y else x <> y.
    Proof.
      destruct x;destruct y;intros;simpl;trivial;try (intro;discriminate;fail).
      generalize (Zeq_spec v v0);destruct (Zeq_bool v v0);intros. subst;trivial.
      intro H1;injection H1;auto.
    Qed.

    Record t : Set := {
      signature : ShortFieldSignature;
      isFinal : bool;
      isStatic : bool;
      visibility : Visibility;
      initValue : value
    }.
    Definition eq_t (x y:t) := 
       let (s1,f1,st1,v1,val1) := x in
       let (s2,f2,st2,v2,val2) := y in
       if FIELDSIGNATURE.eq_t s1 s2 then
        if bool_eq f1 f2 then 
         if bool_eq st1 st2 then
          if eqVisibility v1 v2 then eq_value val1 val2
          else false
        else false
      else false
     else false.
    Lemma eq_t_spec : forall x y, if eq_t x y then x = y else x <> y.
    Proof with try (intro UU;injection UU;auto;fail).     
     intros (s1,f1,st1,v1,val1)(s2,f2,st2,v2,val2);simpl.
     generalize (FIELDSIGNATURE.eq_t_spec s1 s2);destruct (FIELDSIGNATURE.eq_t s1 s2);intros...
     generalize (bool_eq_spec f1 f2);destruct (bool_eq f1 f2);intros...
     generalize (bool_eq_spec st1 st2);destruct (bool_eq st1 st2);intros...
     generalize (eqVisibility_spec v1 v2);destruct (eqVisibility v1 v2);intros...
     generalize (eq_value_spec val1 val2);destruct (eq_value val1 val2);intros...
     subst;trivial. 
    Qed.
    Lemma eq_dec : eq_dec t.
    Proof. exact (Aeq_dec _ eq_t eq_t_spec). Qed.
   
  End FIELD.

  Definition Field := FIELD.t.

  Module Type FIELD_TYPE.
    Parameter signature : Field -> ShortFieldSignature.    
    Parameter isFinal : Field -> bool.
    Parameter isStatic : Field -> bool.
    Parameter visibility : Field -> Visibility.
    Inductive value : Set :=
    | Int (v:Z)
    | NULL 
    | UNDEF.
    Parameter initValue : Field ->  value.
  End FIELD_TYPE.

  Module EXCEPTIONHANDLER.
    Record t : Set := {
      catchType : option ClassName;
      begin : PC;
      end_e : PC;
      handler : PC
    }.
    (* TODO : check correctness of isPCinRange *)
    Definition isPCinRange : t -> PC -> bool :=
      fun ex i => 
        match Ncompare (begin ex) i with
          Gt => false
        | _ => 
             match Ncompare i (end_e ex) with
               Gt => false
             | _ => true
             end
        end.            
   Definition eq_t (x y:t) :=
     let (c1,b1,e1,h1) := x in
     let (c2,b2,e2,h2) := y in
     if opt_eq _ eqClassName c1 c2 then
      if Neq b1 b2 then 
       if Neq e1 e2 then Neq h1 h2 
       else false
      else false
     else false.
    Lemma eq_t_spec : forall x y, if eq_t x y then x = y else x <> y.    
    Proof with  try (intro UU;injection UU;auto;fail).     
      intros (c1,b1,e1,h1)(c2,b2,e2,h2);simpl.
      generalize (opt_eq_spec _ eqClassName eqClassName_spec c1 c2);
        destruct (opt_eq ClassName eqClassName c1 c2);intros ...
      generalize (Neq_spec b1 b2);destruct (Neq b1 b2);intros...
      generalize (Neq_spec e1 e2);destruct (Neq e1 e2);intros ...
      generalize (Neq_spec h1 h2);destruct (Neq h1 h2);intros ...
      subst;trivial.
     Qed.
    End EXCEPTIONHANDLER.
  Definition ExceptionHandler := EXCEPTIONHANDLER.t.

  Module Type EXCEPTIONHANDLER_TYPE.
    Parameter catchType : ExceptionHandler -> option ClassName.
    Parameter isPCinRange : ExceptionHandler -> PC -> bool.
    Parameter handler : ExceptionHandler -> PC.
  End EXCEPTIONHANDLER_TYPE.
   
  Module BYTECODEMETHOD.

    Record t : Set := {
      instr : list (PC*Instruction);
      exceptionHandlers : list ExceptionHandler;
      max_locals : nat;
      max_operand_stack_size : nat
    }.

    Definition  firstAddress (bm:t) : PC :=
      match (instr bm) with
        nil => 0%N
      | (pp,_)::_ => pp
      end.
    Fixpoint instructionAt_rec (pp0:PC) (l:list (PC*Instruction)) {struct l} : option Instruction :=
       match l with
         nil => None
       | (pp,i)::q =>
         if Neq pp0 pp then Some i
         else instructionAt_rec pp0 q
       end.
    Definition instructionAt (bm:t) (pp:PC) : option Instruction :=
     instructionAt_rec pp (instr bm).

    Fixpoint nextAddress_rec (pp0:PC) (l:list (PC*Instruction)) {struct l} : option PC :=
       match l with
         nil => None
       | (pp1,_)::q =>
         if Neq pp0 pp1 then 
            match q with
              nil => None
            | (pp2,_)::_ => Some pp2
            end
         else nextAddress_rec pp0 q
       end.

    Definition nextAddress (bm:t) (pp:PC) : option PC :=
      nextAddress_rec pp (instr bm).

    Definition DefinedInstruction (bm:t) (pc:PC) : Prop :=
      exists i, instructionAt bm pc = Some i.
   
  End BYTECODEMETHOD.
  Definition BytecodeMethod := BYTECODEMETHOD.t.


  Module Type BYTECODEMETHOD_TYPE.
    Parameter firstAddress : BytecodeMethod -> PC.
    Parameter nextAddress : BytecodeMethod -> PC -> option PC.
    Parameter instructionAt : BytecodeMethod -> PC -> option Instruction.
    Parameter exceptionHandlers : BytecodeMethod -> list ExceptionHandler.
    Parameter max_locals : BytecodeMethod -> nat.
    Parameter max_operand_stack_size : BytecodeMethod -> nat.

    Definition DefinedInstruction (bm:BytecodeMethod) (pc:PC) : Prop :=
      exists i, instructionAt bm pc = Some i.

  End BYTECODEMETHOD_TYPE.

  Module METHOD.
    Record t : Set := {
      signature : ShortMethodSignature;
      body : option BytecodeMethod;
      isFinal : bool;
      isStatic : bool;
      isNative : bool;
      visibility : Visibility      
    }.
    Definition isAbstract (m : t) : bool :=
      match body m with
        None => true
      | Some _ => false
    end.
    Definition valid_var (m:t) (x:Var) : Prop :=
      forall bm, body m = Some bm ->
         (Var_toN x) <= (BYTECODEMETHOD.max_locals bm).
    Definition valid_stack_size (m:t) (length:nat) : Prop :=
      forall bm, body m = Some bm ->
         length <= (BYTECODEMETHOD.max_operand_stack_size bm).
  End METHOD.

  Definition Method := METHOD.t.

  Module Type METHOD_TYPE.
    Parameter signature : Method -> ShortMethodSignature.
    Parameter body : Method -> option BytecodeMethod.
    Parameter isFinal : Method -> bool.
    Parameter isStatic : Method -> bool.
    Parameter isNative : Method -> bool.
    Parameter visibility : Method -> Visibility.
    Definition isAbstract (m : Method) : bool :=
      match body m with
        None => true
      | Some _ => false
    end.

    Definition valid_var (m:Method) (x:Var) : Prop :=
      forall bm, body m = Some bm ->
         (Var_toN x) <= (BYTECODEMETHOD.max_locals bm).

    Definition valid_stack_size (m:Method) (length:nat) : Prop :=
      forall bm, body m = Some bm ->
         length <= (BYTECODEMETHOD.max_operand_stack_size bm).
  End METHOD_TYPE.

  Module CLASS.
    Record t : Set := {
      name : ClassName;
      superClass : option ClassName;
      superInterfaces : list InterfaceName;
      fields : list Field;
      methods : list Method;
      isFinal : bool;
      isPublic : bool;
      isAbstract : bool
    }.
    Definition field : t -> ShortFieldName -> option Field :=
      fun cl s => 
      find Field ShortFieldName 
           (fun f => FIELDSIGNATURE.name (FIELD.signature f))
            Peq
           s (fields cl).
    Lemma field_shortname_prop : forall cl s f,
      field cl s = Some f -> s =(FIELDSIGNATURE.name (FIELD.signature f)).
    Proof.
      unfold field; intros cl s f.
      apply (find_getKey Field ShortFieldName 
           (fun f => FIELDSIGNATURE.name (FIELD.signature f))).
      exact Peq_spec.
    Qed.

  Definition definedFields (c:CLASS.t) : list FIELD.t := 
    clean  Field ShortFieldName 
           (fun f => FIELDSIGNATURE.name (FIELD.signature f))
            Peq
           (CLASS.fields c).

  Lemma in_definedFields_field_some : forall (c:CLASS.t) (f:FIELD.t),
    In f (definedFields c) ->
    CLASS.field c (FIELDSIGNATURE.name (FIELD.signature f)) = Some f.
  Proof.
    unfold CLASS.field, definedFields.
    intros c f H.
    apply (in_clean_find_some _ _ _ _ Peq_spec _ _ H).
  Qed.

  Lemma field_some_in_definedFields : forall (c:CLASS.t) (f:FIELD.t) (sfs:ShortFieldName),
    CLASS.field c sfs = Some f-> In f (definedFields c).
  Proof.
    unfold definedFields, CLASS.field; intros.
    apply (find_some_in_clean _ _ _ _ Peq_spec _ _ _ H).
  Qed.


    Definition method : t -> ShortMethodSignature -> option Method :=
      fun cl mid =>
       find Method ShortMethodSignature METHOD.signature
            METHODSIGNATURE.eq_t
            mid (methods cl) .
    Lemma method_signature_prop : forall cl mid m,
      method cl mid = Some m -> mid = METHOD.signature m.
    Proof.
      unfold method; intros p cl mid.
      apply find_getKey.
      exact METHODSIGNATURE.eq_t_spec.
    Qed.
    Definition defined_Method (cl:t) (m:Method) :=
      method cl (METHOD.signature m) = Some m.
  End CLASS.

  Definition Class := CLASS.t.


  Module Type CLASS_TYPE.
    Parameter name : Class -> ClassName.
    Parameter superClass : Class -> option ClassName.
    Parameter superInterfaces : Class -> list InterfaceName.
    Parameter field : Class -> ShortFieldName -> option Field.
    Parameter definedFields : Class -> list Field.
    Parameter in_definedFields_field_some : forall c f,
      In f (definedFields c) ->
      field c (FIELDSIGNATURE.name (FIELD.signature f)) = Some f.
    Parameter field_some_in_definedFields : forall c f sfn,
      field c sfn = Some f -> In f (definedFields c).
    Parameter method : Class -> ShortMethodSignature -> option Method.
    Parameter method_signature_prop : forall cl mid m,
      method cl mid = Some m -> mid = METHOD.signature m.
    Parameter isFinal : Class -> bool.
    Parameter isPublic : Class -> bool.
    Parameter isAbstract : Class -> bool.
    Definition defined_Method (cl:Class) (m:Method) :=
      method cl (METHOD.signature m) = Some m.
  End CLASS_TYPE.

  Module INTERFACE.
    Record t : Set := {
      name : InterfaceName;
      superInterfaces : list InterfaceName;
      fields : list Field;
      methods : list Method;
      isFinal : bool;
      isPublic : bool;
      isAbstract : bool
    }.

    Definition field : t -> ShortFieldName -> option Field :=
      fun cl s => 
      find Field ShortFieldName 
           (fun f => FIELDSIGNATURE.name (FIELD.signature f))
            Peq
           s (fields cl).
    Lemma field_shortname_prop : forall cl s f,
      field cl s = Some f -> s = FIELDSIGNATURE.name (FIELD.signature f).
    Proof.
      unfold field; intros cl s f.
      apply (find_getKey Field ShortFieldName 
           (fun f => FIELDSIGNATURE.name (FIELD.signature f))).
      exact Peq_spec.
    Qed.

    Definition method : t -> ShortMethodSignature -> option Method :=
      fun cl mid =>
       find Method ShortMethodSignature METHOD.signature
            METHODSIGNATURE.eq_t
            mid (methods cl) .
    Lemma method_signature_prop : forall cl mid m,
      method cl mid = Some m -> mid = METHOD.signature m.
    Proof.
      unfold method; intros p cl mid.
      apply find_getKey.
      exact METHODSIGNATURE.eq_t_spec.
    Qed.
  End INTERFACE.
 
  Definition Interface := INTERFACE.t.

  Module Type INTERFACE_TYPE.
    Parameter name : Interface -> InterfaceName. 
    Parameter superInterfaces : Interface ->  list InterfaceName.
    Parameter field : Interface -> ShortFieldName -> option Field.
    Parameter method : Interface -> ShortMethodSignature -> option Method.
    Parameter isFinal : Interface -> bool.
    Parameter isPublic : Interface -> bool.
    Parameter isAbstract : Interface -> bool.
  End INTERFACE_TYPE.

  Module PROG.
    Record t : Set := {
      classes : list Class;
      interfaces : list Interface
    }.
    Definition class : t -> ClassName -> option Class :=
      fun p cn =>
       find Class ClassName CLASS.name
          eqClassName
          cn (classes p).
    Definition defined_Class (p:t) (cl:Class) :=
      class p (CLASS.name cl) = Some cl.
    Lemma name_class_invariant1 : forall p cn cl,
      class p cn = Some cl -> cn = CLASS.name cl.
    Proof.
      unfold class; intros p cn cl.
      apply (find_getKey Class ClassName CLASS.name).
      exact eqClassName_spec.
    Qed.

    Lemma In_classes_defined_Class : forall p cl,
      distinct CLASS.name (classes p) ->
      In cl (classes p) -> defined_Class p cl.
    Proof.
      unfold defined_Class, class; intros p. 
      generalize (classes p).
      induction l; simpl; intros.
      elim H0.
      destruct H0; subst.
      generalize (eqClassName_spec (CLASS.name cl) (CLASS.name cl));
        destruct (eqClassName (CLASS.name cl) (CLASS.name cl));intros.
      reflexivity.
      elim H0; reflexivity.
      generalize (eqClassName_spec (CLASS.name a) (CLASS.name cl));
       destruct (eqClassName (CLASS.name a) (CLASS.name cl));intros.
      rewrite (H a cl); auto with datatypes.
      apply IHl; auto.
      repeat intro; apply H; auto with datatypes.
    Qed.
    Lemma defined_Class_In_classes : forall p cl,
      defined_Class p cl -> In cl (classes p).
    Proof.
      unfold defined_Class, class; intros p. 
      generalize (classes p).
      induction l; simpl; intro.
      intros; discriminate.
      generalize (eqClassName_spec (CLASS.name a) (CLASS.name cl));
        destruct (eqClassName (CLASS.name a) (CLASS.name cl)); intros.
      inversion H0; auto.
      right; auto.
    Qed.

    Definition interface : t -> InterfaceName -> option Interface :=
      fun p iname =>
       find Interface InterfaceName INTERFACE.name
          eqInterfaceName
          iname (interfaces p).
    Definition defined_Interface (p:t) (i:Interface) :=
      interface p (INTERFACE.name i) = Some i.
    Lemma name_interface_invariant1 : forall p iname i,
      interface p iname = Some i -> iname = INTERFACE.name i.
    Proof.
      unfold interface; intros p iname i.
      apply (find_getKey Interface InterfaceName INTERFACE.name).
      exact eqInterfaceName_spec.
    Qed.
  End PROG.

  Definition Program := PROG.t.

  Module Type PROG_TYPE.
    Parameter class : Program -> ClassName -> option Class.
    Definition defined_Class (p:Program) (cl:Class) :=
      class p (CLASS.name cl) = Some cl.
    Parameter name_class_invariant1 : forall p cn cl,
      class p cn = Some cl -> cn = CLASS.name cl.
    Parameter interface : Program -> InterfaceName -> option Interface.
    Definition defined_Interface (p:Program) (i:Interface) :=
      interface p (INTERFACE.name i) = Some i.
    Parameter name_interface_invariant1 : forall p cn cl,
      interface p cn = Some cl -> cn = INTERFACE.name cl.
  End PROG_TYPE.

  Inductive isStatic (p:Program) (fs: FieldSignature) : Prop :=
    isStatic_field : forall (cn:ClassName) (cl:Class) (f:Field),
     PROG.class p (fst fs) = Some cl ->
     CLASS.field cl (FIELDSIGNATURE.name (snd fs)) = Some f ->
     FIELD.isStatic f = true ->
     isStatic p fs.

  Definition javaLangObject : ClassName := (javaLang,object).
  Definition javaLangThrowable : InterfaceName := (javaLang,throwable).
  Definition javaLangCloneable : InterfaceName := (javaLang,cloneable).
  Definition javaIoSerializable : InterfaceName := (javaIo,serializable).

  Inductive direct_subclass (p:Program) (c:Class) (s:Class) : Prop :=
    | direct_subclass1 : 
        PROG.defined_Class p c -> 
        PROG.defined_Class p s ->
        CLASS.superClass c = Some (CLASS.name s) -> 
        direct_subclass p c s.

  (** [subclass] is the reflexive transitive closure of the [direct_subclass] relation 
    (defined over the classes of the program) *)
  Definition subclass (p:Program) : Class -> Class -> Prop :=
    clos_refl_trans Class (direct_subclass p).

  Inductive subclass_name (p:Program) : ClassName -> ClassName -> Prop :=
    | subclass_name1 : forall c1 c2, 
       subclass p c1 c2 -> 
       subclass_name p (CLASS.name c1) (CLASS.name c2).

  Inductive direct_subclass_name (p:Program) : ClassName -> ClassName -> Prop :=
    | direct_subclass_name1 : forall c s,
       direct_subclass p c s ->
       direct_subclass_name p (CLASS.name c) (CLASS.name s).

  (** Similar definitions for interfaces *)
  Inductive direct_subinterface (p:Program) (c:Interface) (s:Interface) : Prop :=
    | direct_subinterface1 : 
      PROG.defined_Interface p c -> 
      PROG.defined_Interface p s ->
      In (INTERFACE.name s) (INTERFACE.superInterfaces c) -> 
      direct_subinterface p c s.

  (** [subinterface] is the reflexive transitive closure of the [direct_subinterface] 
      relation (defined over the interfaces of the program) *)
  Definition subinterface (p:Program) : Interface -> Interface -> Prop :=
    clos_refl_trans Interface (direct_subinterface p).

  Inductive subinterface_name (p:Program) : InterfaceName -> InterfaceName -> Prop :=
    | subinterface_name1 : forall c1 c2, 
       subinterface p c1 c2 -> 
       subinterface_name p (INTERFACE.name c1) (INTERFACE.name c2).

  Inductive direct_subinterface_name (p:Program) : InterfaceName -> InterfaceName -> Prop :=
    | direct_subinterface_name1 : forall c s,
       direct_subinterface p c s ->
       direct_subinterface_name p (INTERFACE.name c) (INTERFACE.name s).
 
  Inductive class_declares_field (p:Program) (cn:ClassName) (fd:ShortFieldSignature) : Field -> Prop :=
    | class_decl_field : forall cl f, 
      PROG.class p cn = Some cl -> 
      CLASS.field cl (FIELDSIGNATURE.name fd) = Some f -> 
      class_declares_field p cn fd f.

  Inductive interface_declares_field (p:Program) (cn:InterfaceName) (fd:ShortFieldSignature) : Field -> Prop :=
    | interface_decl_field : forall cl f, 
      PROG.interface p cn = Some cl -> 
      INTERFACE.field cl (FIELDSIGNATURE.name fd) = Some f -> 
      interface_declares_field p cn fd f.

  (** [defined_field p c fd] holds if the class [c] declares or inherits a field 
      of signature [fd] *)
 Inductive is_defined_field (p:Program) : ClassName -> FieldSignature -> Field -> Prop :=
    | def_class_field : forall cn fd cn' f,
        subclass_name p cn cn' -> 
        class_declares_field p cn' fd f -> 
        is_defined_field p cn (cn',fd) f
    | def_interface_field : forall cn fd cl i1 i' f, 
        PROG.class p cn = Some cl -> 
        In i1 (CLASS.superInterfaces cl) ->
        subinterface_name p i1 i' -> 
        interface_declares_field p i' fd f -> 
        is_defined_field p cn (i',fd) f.

  Definition defined_field (p:Program) (cn:ClassName) (fs:FieldSignature) : Prop :=
    exists f, is_defined_field p cn fs f.

   Definition findMethod (p:Program) (msig: MethodSignature) : option Method :=
    let (cn,smsig) := msig in
      match PROG.class p cn with
	| None => None
	| Some cl => CLASS.method cl smsig 
      end.

  Definition findField (p:Program) (fd: FieldSignature) : option Field :=
    let (cn,sfs) := fd in   
      match PROG.class p cn with
	| None => None
	| Some cl => CLASS.field cl (FIELDSIGNATURE.name sfs)
      end.


  Definition methodPackage (mname: MethodName) : PackageName :=  fst (fst mname).

  (* Relations [check_visibility,check_signature,overrides] are needed
  to define the method [lookup] algorithm **)

  Inductive check_visibility : Visibility -> PackageName -> PackageName ->  Prop :=
    | check_public :  forall p1 p2, check_visibility Public p1 p2
    | check_protected :forall p1 p2, check_visibility Protected p1 p2
    | check_package :forall p, check_visibility Package p p.


   Inductive lookup_here (p:Program) : ClassName ->  ShortMethodSignature -> Method -> Prop :=
    | lookup_here_c : forall cn msig meth, 
       findMethod p (cn,msig) = Some meth -> 
       lookup_here p cn msig meth.


  Inductive lookup (p:Program) : ClassName -> ShortMethodSignature -> (ClassName*Method) -> Prop :=
    | lookup_no_up : forall cn msig meth, lookup_here p cn msig meth -> lookup p cn msig (cn,meth)
    | lookup_up : forall cn  msig, (forall meth, ~ lookup_here p cn msig meth) -> 
      forall super res , direct_subclass_name p cn super -> lookup p super msig res -> lookup p cn msig res.

  (** Get the next pc *)
  Definition next (m:Method) (pc:PC) : option PC :=
    match METHOD.body m with
      None => None
    | Some body => BYTECODEMETHOD.nextAddress body pc
    end.

  (** Get the instruction at the given pc *)
  Definition instructionAt (m:Method) (pc:PC) : option Instruction :=
    match METHOD.body m with
      None => None
    | Some body => BYTECODEMETHOD.instructionAt body pc
    end.

  Inductive implements (p:Program) : ClassName -> InterfaceName -> Prop :=
    | implements_def : forall i cl i', 
           PROG.defined_Interface p i -> 
           PROG.defined_Class p cl ->
           In (INTERFACE.name i) (CLASS.superInterfaces cl) ->
           subinterface_name p (INTERFACE.name i) i' ->
           implements p (CLASS.name cl) i'.

  (** [compat_refType source target] holds if a reference value of type [source] can be 
    assigned to a reference variable of type [target]. See 
    #<a href=http://java.sun.com/docs/books/vmspec/2nd-edition/html/Concepts.doc.html##19674>assignment conversion rules</a># *)

  Inductive compat_refType (p:Program) : refType -> refType -> Prop :=
   | compat_refType_class_class : forall clS clT,
       subclass_name p clS clT ->
       compat_refType p (ClassType clS) (ClassType clT)
   | compat_refType_class_interface : forall clS clT,
       implements p clS clT ->
       compat_refType p (ClassType clS) (InterfaceType clT)
   | compat_refType_interface_class : forall itS,
       compat_refType p (InterfaceType itS) (ClassType javaLangObject)
   | compat_refType_interface_interface : forall itS itT,
       subinterface_name p itS itT ->
       compat_refType p (InterfaceType itS) (InterfaceType itT)
   | compat_refType_array_class : forall tpS,       
       compat_refType p (ArrayType tpS) (ClassType javaLangObject)
   | compat_refType_array_cloneable: forall atS,
       compat_refType p (ArrayType atS) (InterfaceType javaLangCloneable)
   | compat_refType_array_serializable: forall atS,
       compat_refType p (ArrayType atS) (InterfaceType javaIoSerializable)
   | compat_refType_array_array_primitive_type : forall t,       
       compat_refType p (ArrayType (PrimitiveType t)) (ArrayType (PrimitiveType t))
   | compat_refType_array_array_reference_type : forall tpS tpT,       
       compat_refType p tpS tpT ->
       compat_refType p (ArrayType (ReferenceType tpS)) (ArrayType (ReferenceType tpT)).


 
End Make.

Module P <: PROGRAM := Make.

(* 
 Local Variables: 
 coq-prog-name: "coqtop -emacs"
 End:
*)
