Require Import List.

Set Implicit Arguments.

Section FOLD.

Variable A : Type.
Variable f : A -> option A.

Fixpoint fold_opt (l:list A) : option (list A) :=
  match l with
  | nil  => None
  | a::l =>
    match f a with
    | None =>
      match fold_opt l with
      | None => None
      | Some l' => Some (a::l')
      end
    | Some a' => Some (a' :: l)  
    end
  end.

End FOLD.

Section SUBCLASS.

Variable name : Set.
Variable class : Type.
Variable get_name : class -> name.
Variable eq_name : name -> name -> bool.
Variable superclass : class -> option name.

Inductive tree : Set :=
  node : name -> list tree -> tree.

Fixpoint get_tree  (n:name) (lt: list tree) {struct lt} : 
                                    option (tree * list tree) := 
  match lt with
  | nil => None
  | t :: lt => 
    let (x,_) := t in
    if eq_name x n then Some (t, lt)
    else match get_tree n lt with
    | None => None
    | Some (nt,lt) => Some (nt, t::lt)
    end
  end. 

Definition init_tree n lt :=
  match get_tree n lt with
  | None => (node n nil, lt)
  | Some(tc, lt) => (tc, lt)
  end.
    
Fixpoint add_subclass (s:name) (c:tree) (t:tree) {struct t} : option tree :=
  match t with
  | node x l =>
    if eq_name x s then Some (node x (c::l))
    else match fold_opt (add_subclass s c) l with
    | None => None
    | Some l => Some (node x l)
    end
  end.
  
Section BUILD_TREE.

 Variable obj : name.

 Fixpoint build_tree (lt:list tree) (l:list class) {struct l} : 
                                                               list tree :=
   match l with   
   | nil => lt
   | c::l => 
     let (tc,lt) := init_tree (get_name c) lt in
     let super := match superclass c with Some s => s | None => obj end in
     let lt := 
       match fold_opt (add_subclass super tc) lt with
       | Some lt => lt
       | None => (node super (cons tc nil)) :: lt
       end in
     build_tree lt l 
   end.
  
End BUILD_TREE.


Variable map : Type -> Type.
Variable empty : forall C:Set, map C.
Variable update : forall C:Set, map C -> name -> C -> map C.
Variable get : forall C:Set, map C -> name -> option C.

Fixpoint build_supers (super:list name) (t:tree) (m:map(list name)){struct t} :
                            map (list name) := 
  match t with
  | node a lt =>
    update (fold_right (build_supers (a::super)) m lt) a super
  end.

Fixpoint test_In (a:name) (l:list name) {struct l} : bool :=
  match l with
    nil => false
  | x::q => if eq_name a x then true else test_In a q
  end.

Fixpoint get_obj (l:list class) : option (name * list class) := 
 match l with
 | nil => None
 | c::l =>
    match superclass c with
    | None => Some (get_name c, l)
    | Some _ =>
      match get_obj l with
      | None => None
      | Some (obj, l) => Some (obj, c::l)
      end
    end
 end.

Definition subclass_test (l:list class) : option (name -> name -> bool) :=
  match get_obj l with
  | None => None
  | Some (obj, l) =>
    match build_tree obj nil l with
    | t::nil =>
      let m := build_supers nil t (empty (list name)) in
      Some (fun x y =>
              match get m x with
              | None => false
              | Some supers => if eq_name x y then true else test_In y supers
              end)
    | _ => None
    end 
  end.

End SUBCLASS.


Module A.

Section __A__.
Variable A : Set.
Variable test : A -> A -> bool.

Inductive tree (A:Set) : Set :=
  node : A -> list (tree A) -> tree A.

Section fold_opt.
Variable B:Type.
Variable f:B->option B.

Fixpoint fold_opt (l:list B) {struct l} : option (list B) :=
 match l with
   nil => None
 | a::q =>
    match f a with
     | Some a' => Some (a'::q)
     | None => 
        match fold_opt q with
          None => None
        | Some q' => Some (a::q')
        end
    end
 end.

End fold_opt.

Fixpoint add (n s:A) (t:tree A) {struct t} : option (tree A) :=
  match t with
    node x l =>
      if test x s then Some (node x (node n nil::l))
      else
        match fold_opt (add n s) l with
          None => None
        | Some l' => Some (node x  l')
        end
  end.

Variable class : Type.
Variable name : class -> A.
Variable superclass : class -> option A.

Fixpoint sort1_aux (sup:A) (n l:list (A*class)) {struct l} : option (list (A*class)) :=
  match l with
    | nil => None
    | (supc,c)::q =>
      if test (name c) sup 
        then Some ((supc,c)::n++q)
        else match sort1_aux sup n q with
               | None => None
               | Some l' => Some ((supc,c)::l')
             end
  end.

Definition sort1 (sup:A) (n:list (A*class)) (l:list (list (A*class))) : list (list (A*class)) :=
  match @fold_opt (list (A*class)) (sort1_aux sup n) l with
    | None => n::l
    | Some l' => l'
  end.

Definition test_sup_list (a:A) (l:list (A*class)) : bool :=
  match l with
    | nil => false
    | (a0,_)::_ => test a a0
  end.

Fixpoint sort2 (a:A) (c:class) (l:list (list (A*class))) {struct l} : 
  (list (A*class) * list (list (A*class))) :=
  match l with
    | nil => (nil,nil)
    | cons b q => 
      let (lc,ln) := sort2 a c q in
        if test_sup_list (name c) b
          then (b++lc,ln)
          else (lc,b::ln)
  end.

Fixpoint sort3 (l:list class) : list (list (A*class)) :=
  match l with
    | nil => nil
    | c::q => 
      match superclass c with
        | None => sort3 q
        | Some a => 
          let ll := sort3 q in
          let (lc,ll') := sort2 a c ll in
            sort1 a ((a,c)::lc) ll'
      end
  end.

Definition sort (o:class) (l:list class) : list class :=
  o::(flat_map (map (@snd A class)) (sort3 l)).


Fixpoint build_tree_rec (t:tree A) (l:list class) {struct l} : option (tree A) :=
 match l with
  nil => Some t
 | c::q => 
     match superclass c with
      Some s =>
         match (add (name c) s t) with
          None => None
        | Some t' => build_tree_rec t' q
        end
     | _ => None
     end
 end.

Definition build_tree_aux (l:list class) : option (tree A) :=
 match l with 
 | nil => None
 | c::l => 
    match superclass c with
     None => build_tree_rec (node (name c) nil) l
    | _ => None
    end
 end.

Fixpoint find_object (l:list class) : option class :=
  match l with
    | nil => None
    | a::q => match superclass a with
                | None => Some a
                | _ => find_object q
              end
  end.

Definition build_tree l := 
  match find_object l with
    | None => None
    | Some o => build_tree_aux (sort o l)
  end.

Section find.
Variable a : A.

Fixpoint find_in_tree (t:tree A) {struct t} : option (tree A) :=
  match t with
    node x l =>
    if (test x a) 
      then Some t 
      else fold_right 
        (fun t ot => 
          match ot with
            | None => find_in_tree t
            | Some t' => Some t'
          end)
          None
          l
  end.

End find.

Definition flat_map' :=
  fun (A B : Type) (f : A -> list B) =>
    fix flat_map (l : list A) {struct l} : list B :=
    match l with
      | nil => nil (A:=B)
      | x :: t => f x ++ flat_map t
    end.

Fixpoint tree_to_list (t:tree A) : list A :=
  match t with
    node x l => x::(flat_map' tree_to_list l)
  end.

Fixpoint test_In (a:A) (l:list A) {struct l} : bool :=
  match l with
    nil => false
  | x::q => if test a x then true else test_In a q
  end.

Fixpoint find (C:Type) (l:list (A*C)) (a:A) {struct l} : option C :=
  match l with
    nil => None
  | (x,c)::q => if test a x then Some c else find q a
  end.

Variable B:Type -> Type.
Variable empty : forall C:Type, B C.
Variable update : forall C:Type, B C -> A -> C -> B C.
Variable get : forall C:Type, B C -> A -> option C.
Variable M Ms F Fs: Type.
Variable findM : A*Ms -> option M.
Variable findF : A*Fs -> option F.

Fixpoint lookup (ms:Ms)(l:list A) : option (A*M) :=
  match l with
    | nil => None
    | c::q =>
      match findM (c,ms) with
        | None => lookup ms q
        | Some m => Some (c,m)
      end
  end.

Fixpoint static_lookup (ms:Ms)(l:list A) : list (A*M) :=
  match l with
    | nil => nil
    | c::q =>
      match findM (c,ms) with
        | None => static_lookup ms q
        | Some m => (c,m)::(static_lookup ms q)
      end
  end.

Fixpoint lookupF (fs:Fs) (l:list A) : option (A*F) :=
  match l with
    | nil => None
    | c::q =>
      match findF (c,fs) with
        | None => lookupF fs q
        | Some f => Some (c,f)
      end
  end.

Fixpoint build_supers (l:list A) (t:tree A) (b:B (list A)) {struct t}  : B (list A) := 
  match t with
    node a ltt =>
      update (fold_right (build_supers (a::l)) b ltt)  a l
  end.

Definition get_tools (l:list class) : 
  option ((A -> (list A)) *
          (A -> A -> bool) *
          (A -> Ms -> option (A*M)) *
          (A -> Fs -> option (A*F)) *
          (A -> Ms -> list (A*M))) :=
  match build_tree l with
    None => None
  | Some t =>
      let m := build_supers nil t (empty (list A)) in
        Some (
          fun x => 
            match get m x with
              | None => nil
              | Some l => l
            end,
          fun x y =>
            match get m x with
              | None => false
              | Some supers => test_In y supers
            end,
          fun c ms =>
            match get m c with
              | None => None
              | Some supers => lookup ms (c::supers)
            end,
          fun c fs =>
            match get m c with
              | None => None
              | Some supers => lookupF fs (c::supers)
            end,
          fun c ms =>
            match find_in_tree c t with
              | None => nil
              | Some sub => 
                match get m c with
                  | None => static_lookup ms (tree_to_list sub)
                  | Some supers => 
                    match lookup ms supers with
                      | Some m => m :: static_lookup ms (tree_to_list sub)
                      | None => static_lookup ms (tree_to_list sub)
                    end
                end
            end
        )
  end.

Definition subclass_test (l:list class) : option (A -> A -> bool) :=
  match get_tools l with
    | None => None
    | Some (_,f,_,_,_) => Some f
  end.

End __A__.

End A.

