(*  <Insert License Here>

    $Id: Wf.v 69 2006-03-06 20:16:11Z davidpichardie $ *)

(** Wellformedness properties of programs are stated here.
  These are properties that should be enforced by the parser!
  These properties are decidable and the appropriate checks could be implemented.

 @author Frederic Besson *)


Require Import List.
Require Import ZArith.
Require Import Program.



Module WF (Prog:PROGRAM).

  Import Prog.

  (* Interface methods do not have a bytecode body *)
  Definition interface_method_prop (i :Interface) := 
    forall (sig:ShortMethodSignature) (meth:Method),
      INTERFACE.method i sig = Some meth -> METHOD.body meth = None.

  (* The inheritance relation is acyclic (it forms a class hierarchy) 
    BEWARE: subclass is the REFLEXIVE transitive closure of direct_subclass *)
  
   Definition class_hierarchy1 (p:Program) := 
     forall (cl cl1:Class), direct_subclass p cl cl1 -> cl <> cl1.
   Definition class_hierarchy2 (p:Program)  :=  
     forall (cl cl1 cl':Class), 
       direct_subclass p cl cl1 -> subclass p cl1 cl' -> cl' <> cl.

  (* TODO: Final classes/methods are not extended/overriden - fb *)

	(** For one signature there is only one method *)
	Definition method_uniq_prop  := 
	  forall (m1 m2: Method) ,
	      METHOD.signature m1 = METHOD.signature m2 -> m1 = m2.
	  
  (* TODO: state conditions over interfaces - fb *)
 
  (* TODO: Symbolic names are resolved - fb
   Definition is_resolved_class (p:Program) (cn:ClassName) := PROG.class p cn <> None.

   Definition is_visible (p:Program) (scope:ClassName) (cl:ClassName) :=
     match CLASS.visibility cl with
       | CLASS.Public => True
       | CLASS.Protected => subclass p scope cl
       | CLASS.Package => 

    [cn'] can be resolved within scope [cn] 
   Definition is_resolved_class_within (p:Program) (cn:ClassName) (cn':ClassName) :=
     is_resolved_class p cn /\ is_resolved_class p cn' /\ is_visible_from 
  *)

   Record well_formed_program (P:Program) : Set := {
     interface_method_cond2 : forall cn cl, 
                          PROG.interface P cn = Some cl -> interface_method_prop cl;
     class_hierarchy_cond1 : class_hierarchy1 P;
     class_hierarchy_cond2 : class_hierarchy2 P       
   }.

End WF.

   (* TODO: implement something like
       Definition check_wf (P:Program) : option (well_formed_program P).
    This sould be possible modulo decidable equalities... - fb
  *)




