Set Implicit Arguments.

Inductive optionT (A:Type) : Type := 
   NoneT : optionT A
 | SomeT : A -> optionT A.
  
Implicit Arguments NoneT [A].

Section LISTTYPE.

 Variable A:Type.

 Inductive listT : Type :=
  | NilT : listT
  | ConsT : A -> listT -> listT.
 
 Fixpoint inT (a:A) (l:listT) {struct l} : Prop :=
    match l with
      | NilT => False
      | ConsT a' l' => a' = a \/ inT a l'
    end.

 Fixpoint rev_appT (l1 l2:listT) {struct l1} : listT :=
   match l1 with
   | NilT => l2
   | ConsT a l1 => rev_appT l1 (ConsT a l2)
   end.

 Lemma inT_rev_appT_r : forall a l1 l2, inT a l2 -> inT a (rev_appT l1 l2).
 Proof. 
  induction l1;simpl;intros;auto.
  apply IHl1;right;trivial.
 Qed.

 Lemma inT_rev_appT_l : forall a l1 l2, inT a l1 -> inT a (rev_appT l1 l2).
 Proof. 
  induction l1;simpl;intros l2 H;destruct H;auto.
  apply inT_rev_appT_r;left;trivial.
 Qed.

 Lemma inT_rev_appT : forall a l1 l2, inT a (rev_appT l1 l2) -> inT a l1 \/ inT a l2.
 Proof.
  induction l1;simpl;intros;auto.
  destruct (IHl1  _ H);auto.
  destruct H0;auto.
 Qed.

End LISTTYPE.

Implicit Arguments NilT [A].

Notation "x ** y" := (prodT x y) (at level 75).
Notation "( x ,, y )" := (pairT x y) : core_scope.

Section MapT.
 Variables A B:Type.
 Variable f: A -> B.
 Fixpoint mapT (l:listT A) : listT B :=
  match l with 
  | NilT => NilT
  | ConsT a l => ConsT (f a) (mapT l)
  end.
 
 Lemma inT_mapT : forall a l, inT a l -> inT (f a) (mapT l).
 Proof.
   induction l;simpl;intros;auto.
   destruct H;auto. rewrite H;auto.
 Qed.

 Lemma mapT_inT : forall b l, inT b (mapT l) -> exists a, inT a l /\ f a = b.
 Proof.
  induction l;simpl;intros H;destruct H.
  exists a;auto.
  destruct (IHl H) as (a', (H2, H3)).
  exists a';auto.
 Qed.

End MapT.


