Add LoadPath "../EffectsBicolano".
Require Export Tactic.
Require Export Misc.
Require Export BigStep.
Require Export MapSignatures.

Open Scope type_scope.

Module Type RULES.

 Declare Module V5 : BIGSTEP.
 Declare Module MM:MAP with Definition key := V5.Dom.Prog.MethodSignature.
 Declare Module PCM:MAP with Definition key := V5.Dom.Prog.PC.
Print V5.NormalStep.
Print V5.IntraNormalState. 
(*********************************)



(**********************************)
 Load "BaseLogicLoad.v".

 Ltac instrAt Hb := match goal with
   | [H : instructionAt ?M ?l = _ |- _] => rewrite H in Hb;try (elim Hb;fail)
   | _ => idtac end.

 Parameter GP_SP_Judgement_inv : GP -> 
    forall msig M ms LST LAT, 
     MSPEC_LOOKUP msig = Some ((ms,,LST),,LAT) ->
     findMethod P msig = Some M ->
     forall b l A B I,
     SP_Judgement M LST LAT b l A B I ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic M LST LAT l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\
	     SP_Judgement_call M LST LAT l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\
	     SP_Judgement_virtualcall M LST LAT l msig A B I))
      | ReturnInstr => SP_Judgement_return M LAT l A B I
      | IfInstr o => SP_Judgement_if M LST LAT l o A B I
      | GotoInstr o => SP_Judgement_goto M LST LAT l o A B I
      | LSwitchInstr def listkey => instructionAt M l =Some(Lookupswitch def listkey) /\ SP_Judgement_lookupswitch M LST LAT def listkey l A B I
      | TSwitchInstr def low high list_offset => instructionAt M l =Some(Tableswitch def low high list_offset) /\ SP_Judgement_tableswitch M LST LAT def low high list_offset l A B I
      end.

 End SP_Judgement.
End RULES.

Module Rules (VV5 : BIGSTEP)
             (_MM:MAP with Definition key := VV5.Dom.Prog.MethodSignature)
             (_PCM:MAP with Definition key := VV5.Dom.Prog.PC)
            <: RULES
                   with Module V5 := VV5 
                   with Module MM := _MM
                   with Module PCM := _PCM.

  Module V5 := VV5.
  Module MM := _MM.
  Module PCM := _PCM.

  Load "BaseLogicLoad.v".

Print NormalStep.
Print IntraNormalState.

  Ltac instrAt Hb := match goal with
   | [H : instructionAt ?M ?l = _ |- _] => rewrite H in Hb;try (elim Hb;fail)
   | _ => idtac end.

  Lemma GP_SP_Judgement_true_false : GP ->
    forall msig M ms LST LAT,
      MSPEC_LOOKUP msig = Some ((ms,,LST),,LAT) -> 
      findMethod P msig = Some M ->
     forall  l A B I, SP_Judgement M LST LAT true l A B I -> 
     exists AA:Assertion, exists BB:LocalPost, exists II:StrongInv,
      (forall s0 s, A s0 s -> AA s0 s) /\
      (forall s0 s t, A s0 s -> BB s0 s t -> B s0 s t) /\
       (forall t s , II t s -> I t s) /\
       SP_Judgement M LST LAT false l AA BB II.
  Proof.
   intros HGP.
   assert (forall b msig M  ms  LST LAT l A B I,
     SP_Judgement M LST LAT b l A B I -> 
     MSPEC_LOOKUP msig = Some ((ms,,LST),,LAT) -> 
     findMethod P msig = Some M -> 
       b = true ->
     exists AA:Assertion, exists BB:LocalPost, exists II:StrongInv,
       (forall s0 s, A s0 s -> AA s0 s) /\
      (forall s0 s t, A s0 s -> BB s0 s t -> B s0 s t) /\
       (forall t s , II t s -> I t s) /\
       SP_Judgement M LST LAT false l AA BB II);eauto.
   induction 1;intros;try discriminate;subst;auto.
   destruct (IHSP_Judgement H3 H4 (refl_equal _)) as
         (AA0,(BB0,(II0,(HH1,(HH2,(HH3,HH4)))))).
   exists AA0;exists BB0;exists II0;repeat split;eauto.
   exists A;exists B0;exists I0;repeat split;eauto.
   exists A;exists B0;exists I0;repeat split;eauto.
   destruct ms as ((pre,post),MI). destruct HGP as (BST1,HGP1).  destruct (HGP1 _ _ _ _ _ _ H2 _ H3);auto.
  Qed.

  Lemma kind_aux : forall M LST LAT l (A AA:Assertion) (B BB:LocalPost)
                               (I II:StrongInv),
    (forall s0 s, A s0 s -> AA s0 s) /\
    (forall s0 s t, A s0 s -> BB s0 s t -> B s0 s t) /\
    (forall t s , II t s -> I t s) ->
    (exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic M LST LAT l AA BB II
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\
	     SP_Judgement_call M LST LAT l msig AA BB II) \/
           (instructionAt M l = Some(Invokevirtual msig) /\
	     SP_Judgement_virtualcall M LST LAT l msig AA BB II))
      | ReturnInstr => SP_Judgement_return M LAT l AA BB II
      | IfInstr o => SP_Judgement_if M LST LAT l o AA BB II
      | GotoInstr o => SP_Judgement_goto M LST LAT l o AA BB II
      | LSwitchInstr def listkey => instructionAt M l =Some(Lookupswitch def listkey) /\ SP_Judgement_lookupswitch M LST LAT def listkey l AA BB II
      | TSwitchInstr def low high loffset => instructionAt M l =Some(Tableswitch def low high loffset) /\ SP_Judgement_tableswitch M LST LAT def low high loffset l AA BB II
      end) ->
   exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic M LST LAT l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\
	     SP_Judgement_call M LST LAT l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\
	     SP_Judgement_virtualcall M LST LAT l msig A B I))
      | ReturnInstr => SP_Judgement_return M LAT l A B I
      | IfInstr o => SP_Judgement_if M LST LAT l o A B I
      | GotoInstr o => SP_Judgement_goto M LST LAT l o A B I
      | LSwitchInstr def listkey => instructionAt M l =Some(Lookupswitch def listkey) /\ SP_Judgement_lookupswitch M LST LAT def listkey l A B I
      | TSwitchInstr def low high loffset => instructionAt M l =Some(Tableswitch def low high loffset) /\ SP_Judgement_tableswitch M LST LAT def low high loffset l A B I
      end.
  Proof.
   intros. destruct H0 as (ins, (Heq, HH));exists ins;split;trivial.
   destruct H as (UU1,(UU2,UU3)).
   destruct (kind_of_instr ins);inversion_mine HH;auto.
   econstructor;eauto. intros;eauto.
   destruct H. 
     destruct H; exists x.  left.  split. trivial. 
       inversion_mine H0.
         econstructor;eauto.  intros;eauto. 
     destruct H; exists x.  right.  split. trivial. 
       inversion_mine H0.
         econstructor;eauto.  intros;eauto. 
   econstructor;eauto. intros;eauto.
   econstructor;eauto. intros;eauto.
   econstructor;eauto. intros;eauto.
   split. trivial.  inversion_mine H0. econstructor. eauto. apply H2. eauto. eauto. eauto. eauto. eauto. 
   split. trivial. inversion_mine H0. econstructor. eauto. apply H2. eauto. eauto. eauto. eauto. eauto. 
  Qed.
 
  Lemma SP_Judgement_inv_false_kind : 
    forall msig M ms LST LAT b l A B I,
     SP_Judgement M LST LAT b l A B I -> 
     MSPEC_LOOKUP msig = Some ((ms,,LST),,LAT) ->
     findMethod P msig = Some M -> 
     b = false ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic M LST LAT l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\
	     SP_Judgement_call M LST LAT l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\
	     SP_Judgement_virtualcall M LST LAT l msig A B I))
      | ReturnInstr => SP_Judgement_return M LAT l A B I
      | IfInstr o => SP_Judgement_if M LST LAT l o A B I
      | GotoInstr o => SP_Judgement_goto M LST LAT l o A B I
      | LSwitchInstr def listkey => instructionAt M l =Some(Lookupswitch def listkey) /\ SP_Judgement_lookupswitch M LST LAT def listkey l A B I
      | TSwitchInstr def low high loffset => instructionAt M l =Some(Tableswitch def low high loffset) /\ SP_Judgement_tableswitch M LST LAT def low high loffset l A B I
      end.
  Proof.
   induction 1;intros;simpl_eq;eauto.
   unfold isBasicInstr in H. destruct (instructionAt M l);try (elim H;fail).
     exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);
     econstructor;eauto.
   unfold isLSwitchInstr in H. destruct(instructionAt M l);try (elim H;fail). 
     destruct (kind_of_instr i); try (elim H;fail). 
     destruct H as (Ins,H); simpl_eq; subst. 
     destruct H; simpl_eq; subst. 
     exists (Lookupswitch def listkey). split. reflexivity. 
       econstructor;eauto. 
       econstructor;eauto. 
   unfold isTSwitchInstr in H. destruct (instructionAt M l);try (elim H;fail).
     destruct (kind_of_instr i); try (elim H;fail). 
     destruct H as (Ins,H); simpl_eq; subst. 
     destruct H as (Cond1, H); simpl_eq; subst. 
     destruct H as (Cond2, H); simpl_eq; subst. 
     destruct H as (Cond3, Cond4); simpl_eq; subst. 
     exists (Tableswitch def low high list_offset). split. reflexivity. 
       econstructor;eauto. 
       econstructor; eauto.  
   unfold isReturnInstr in H;destruct (instructionAt M l);try (elim H;fail).
     exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);
     econstructor;eauto.
   exists (Goto o);split;trivial;econstructor;eauto.
   unfold isIfInstr in H. destruct (instructionAt M l);try (elim H;fail).
   exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);subst;
     econstructor;eauto.
   exists (Invokestatic msig0);split;trivial.
    exists msig0.  left. split;trivial;econstructor;eauto.
   exists (Invokevirtual msig0);split;trivial.
    exists msig0.  right. split;trivial;econstructor;eauto.
  eapply kind_aux;eauto.
  discriminate H4. 
Qed. 

  Lemma GP_SP_Judgement_inv : GP -> 
    forall msig M ms LST LAT,
     MSPEC_LOOKUP msig = Some ((ms,,LST),,LAT) ->
     findMethod P msig = Some M ->
     forall b l A B I, SP_Judgement M LST LAT b l A B I ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic M LST LAT l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\
	     SP_Judgement_call M LST LAT l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\
	     SP_Judgement_virtualcall M LST LAT l msig A B I))
      | ReturnInstr => SP_Judgement_return M LAT l A B I
      | IfInstr o => SP_Judgement_if M LST LAT l o A B I
      | GotoInstr o => SP_Judgement_goto M LST LAT l o A B I
      | LSwitchInstr def listkey => instructionAt M l = Some(Lookupswitch def listkey) /\ SP_Judgement_lookupswitch M LST LAT def listkey l A B I
      | TSwitchInstr def low high loffset => instructionAt M l = Some(Tableswitch def low high loffset) /\ SP_Judgement_tableswitch M LST LAT def low high loffset l A B I
      end.
  Proof.
   intros. destruct b.
   destruct (GP_SP_Judgement_true_false H _ _ _ _ _ H0 H1 _ _ _ _ H2) as 
    (AA, (BB, (II, (HH1, (HH2, (HH3,HH4)))))).
   eapply kind_aux; eauto. 
   eapply SP_Judgement_inv_false_kind;eauto.
   eapply SP_Judgement_inv_false_kind;eauto.
  Qed.

 End SP_Judgement.
End Rules.
 
