Add LoadPath "../Bicolano".
Require Export Tactic.
(*Require Export Misc.*)
Require Export BigStep.
Require Export AssListM.

Open Scope type_scope.

Module Type RULES.

 Declare Module V5 : BIGSTEP.
 Declare Module AL : ASSLIST.

 Load "BaseLogicLoad.v".
 
 Variable subclass : ClassName -> ClassName -> Prop.

 Inductive BehavSubTyping : Prop :=
   | BehavST :  
        (forall d c (M:Method) (M1:Method) T MI Msig Msig1 mname ,
            subclass d c -> 
            MSPEC_LOOKUP MST M = SomeT (T ,, MI) ->
            METHOD.signature M = Msig ->
            METHODSIGNATURE.name Msig = (c,mname) ->
            METHODSIGNATURE.name Msig1 = (d,mname) ->
            METHOD.signature M1 = Msig1 ->
            (exists T1, exists MI1,
                    MSPEC_LOOKUP MST M1 = SomeT (T1 ,, MI1) /\
                    (forall s0 t , T1 s0 t -> T s0 t) /\
                    (forall s0 s , MI1 s0 s -> MI s0 s))) ->
     (*************************************************************)
     BehavSubTyping.

 Ltac instrAt Hb := match goal with
   | [H : instructionAt ?M ?l = _ |- _] => rewrite H in Hb;try (elim Hb;fail)
   | _ => idtac end.

 Parameter GP_SP_Judgement_inv : GP -> 
    forall C M b l A B I,
     SP_Judgement C M b l A B I ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic C M l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\ SP_Judgement_call C M l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\ SP_Judgement_virt_call C M l msig A B I))
      | ReturnInstr => SP_Judgement_return C M l A B I
      | IfInstr o => SP_Judgement_if C M l o A B I
      | GotoInstr o => SP_Judgement_goto C M l o A B I
      | SwitchInstr => False
      end.

 End SP_Judgement.
End RULES.


Module Rules (VV5 : BIGSTEP) (AAL:ASSLIST).

  Module V5 := VV5.
  Module AL := AAL.

  Load "BaseLogicLoad.v".

  Ltac instrAt Hb := match goal with
   | [H : instructionAt ?M ?l = _ |- _] => rewrite H in Hb;try (elim Hb;fail)
   | _ => idtac end.

  Lemma GP_SP_Judgement_true_false : GP ->
    forall C M l A B I,
     SP_Judgement C M true l A B I -> 
     exists AA:LocalAssn, exists BB:LocalPost, exists II:StrongInv,
       (forall s0 s , A s0 s -> AA s0 s) /\
       (forall s0 s t , BB s0 s t -> B s0 s t) /\
       (forall t s , II t s -> I t s) /\
       SP_Judgement C M false l AA BB II.
  Proof.
   intros HGP.
   assert (  forall b C M l A B I,
     SP_Judgement C M b l A B I -> 
     b = true ->
     exists AA:LocalAssn, exists BB:LocalPost, exists II:StrongInv,
       (forall s0 s , A s0 s -> AA s0 s) /\
       (forall s0 s t , BB s0 s t -> B s0 s t) /\
       (forall t s , II t s -> I t s) /\
       SP_Judgement C M false l AA BB II).
   induction 1;intros;try discriminate;subst;auto.
   destruct (IHSP_Judgement (refl_equal _)) as
         (AA0,(BB0,(II0,(HH1,(HH2,(HH3,HH4)))))).
   exists AA0;exists BB0;exists II0;repeat split;eauto.
   exists A;exists B0;exists I0;repeat split;eauto.
   exists A;exists B0;exists I0;repeat split;eauto.
   destruct HGP;auto.
   eauto.
  Qed.

  Lemma kind_aux : forall C M l (A AA:LocalAssn) (B BB:LocalPost)
                               (I II:StrongInv),
    (forall s0 s , A s0 s -> AA s0 s) ->
    (forall s0 s t , BB s0 s t -> B s0 s t) ->
    (forall t s , II t s -> I t s) ->
    (exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic C M l AA BB II
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\ SP_Judgement_call C M l msig AA BB II) \/
           (instructionAt M l = Some(Invokevirtual msig) /\ SP_Judgement_virt_call C M l msig AA BB II))
      | ReturnInstr => SP_Judgement_return C M l AA BB II
      | IfInstr o => SP_Judgement_if C M l o AA BB II
      | GotoInstr o => SP_Judgement_goto C M l o AA BB II
      | SwitchInstr => False
      end) ->
   exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic C M l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\ SP_Judgement_call C M l msig A B I)  \/ 
           (instructionAt M l = Some(Invokevirtual msig) /\ SP_Judgement_virt_call C M l msig A B I))
      | ReturnInstr => SP_Judgement_return C M l A B I
      | IfInstr o => SP_Judgement_if C M l o A B I
      | GotoInstr o => SP_Judgement_goto C M l o A B I
      | SwitchInstr => False
      end.
   
  Proof.
   intros. destruct H2 as (ins, (Heq, HH));exists ins;split;trivial.
   destruct (kind_of_instr ins);inversion_mine HH.
   econstructor;eauto. intros;eauto.
   exists x.
   destruct H2. 
     (*Invstatic*)
     left. destruct H2. split. assumption. inversion_mine H3.
       apply (SP_INVS_call C M l x A B0 I0 A0 B1 I1 m T MI (length (METHODSIGNATURE.parameters x)) H4). 
       assumption. assumption. assumption. assumption. eauto. eauto. eauto. eauto. eauto. eauto. 
       eauto. eauto.
     (*Invvirt*) 
     right. destruct H2. split. assumption. inversion_mine H3.
       apply (SP_INVV_call C M l x A B0 I0 A0 B1 I1 m T MI (length (METHODSIGNATURE.parameters x)) H4). 
       assumption. assumption. assumption. eauto. eauto. eauto. eauto. eauto. eauto. 
        eauto. eauto. eauto.
  (*Return*)
   econstructor;eauto. intros;eauto.
   econstructor;eauto. intros;eauto.
   econstructor;eauto. intros;eauto.
  Qed.
 
  Lemma SP_Judgement_inv_false_kind : 
    forall C M b l A B I,
     SP_Judgement C M b l A B I ->
     b = false ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic C M l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\ SP_Judgement_call C M l msig A B I) \/
            (instructionAt M l = Some(Invokevirtual msig) /\ SP_Judgement_virt_call C M l msig A B I))
      | ReturnInstr => SP_Judgement_return C M l A B I
      | IfInstr o => SP_Judgement_if C M l o A B I
      | GotoInstr o => SP_Judgement_goto C M l o A B I
      | SwitchInstr => False
      end.
  Proof.
   induction 1;intros;simpl_eq;eauto.
   unfold isBasicInstr in H. destruct (instructionAt M l);try (elim H;fail).
   exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);
     econstructor;eauto.
   unfold isReturnInstr in H;destruct (instructionAt M l);try (elim H;fail).
   exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);
     econstructor;eauto.
   exists (Goto o);split;trivial;econstructor;eauto.
   unfold isIfInstr in H. destruct (instructionAt M l);try (elim H;fail).
   exists i;split;trivial;destruct (kind_of_instr i);try (elim H;fail);subst;
     econstructor;eauto.
   exists (Invokestatic msig). split. trivial. econstructor. left. split. eauto. clear H10. clear H8. 
     apply (SP_INVS_call C M l msig A B0 I0 A B0 I0 m T MI N H0). assumption. assumption. 
       eauto.  assumption. assumption. assumption. assumption. assumption. assumption. eauto. eauto. eauto. 
   exists (Invokevirtual msig). split. trivial. econstructor. right. split. eauto. clear H10. clear H8. 
     apply (SP_INVV_call C M l msig A B0 I0 A B0 I0 m T MI N H0). assumption. assumption. 
       eauto.  assumption. assumption. assumption. assumption. assumption. assumption. eauto. eauto. eauto.
   eapply kind_aux;eauto.
   discriminate.
  Qed.

  Lemma GP_SP_Judgement_inv : GP -> 
    forall C M b l A B I,
     SP_Judgement C M b l A B I ->
     exists ins, instructionAt M l = Some ins /\
      match kind_of_instr ins with 
      | BasicInstr => SP_Judgement_basic C M l A B I
      | CallInstr => 
         exists msig, 
          ((instructionAt M l = Some(Invokestatic msig) /\ SP_Judgement_call C M l msig A B I) \/
           (instructionAt M l = Some(Invokevirtual msig) /\ SP_Judgement_virt_call C M l msig A B I))
      | ReturnInstr => SP_Judgement_return C M l A B I
      | IfInstr o => SP_Judgement_if C M l o A B I
      | GotoInstr o => SP_Judgement_goto C M l o A B I
      | SwitchInstr => False
      end.
  Proof.
   intros. destruct b.
   destruct (GP_SP_Judgement_true_false H _ _ _ _ _ _ H0) as 
    (AA, (BB, (II, (HH1, (HH2, (HH3,HH4)))))).
   eapply kind_aux;eauto.
   eapply SP_Judgement_inv_false_kind;eauto.
   eapply SP_Judgement_inv_false_kind;eauto.
  Qed.

 End SP_Judgement.
End Rules.
