  Export V5 Sem Prog.

  Definition Store := Sem.LocalVar.t.
  Definition Heap := Sem.Heap.t.
  Definition OpStack := Sem.OperandStack.t.

  Definition InitState := Store * Heap.
  Definition LocalState := Heap*OpStack * Store.
  Definition TermState := ReturnState.

  Definition LocalAssn := InitState -> LocalState -> Prop.

  Definition LocalPost := InitState -> LocalState -> TermState -> Prop.

  Definition StrongInv := LocalState -> LocalState -> Prop.

  Definition Label := PC.

  Definition MethSpec := InitState -> TermState -> Prop.
  Definition MethInv := InitState -> LocalState -> Prop.

  (* some abbreviations for using module ASSLIST *)
  Definition AnnoType := AL.t (Class * Method * Label) LocalAssn.
  Definition Anno_LOOKUP
     : AnnoType -> (Class * Method * Label) -> optionT LocalAssn
     := AL.lookup (Class * Method * Label) (LocalAssn).

  (*Due to the way Bicolano's invoke rules are formulated, it is
    easier to have access to the methods only, so let's simplify
    the type of method specification tables.*)
  Definition MSPEC_TYPE := AL.t Method (MethSpec ** MethInv).
  Definition MSPEC_LOOKUP
     : MSPEC_TYPE -> Method -> optionT (MethSpec ** MethInv)
     := AL.lookup Method (MethSpec ** MethInv).

  Definition Assum_TYPE
     := AL.t (Class * Method * Label) (LocalAssn ** (LocalPost ** StrongInv)).
  Definition ASSUM_LOOKUP
     : Assum_TYPE -> (Class * Method * Label) ->
                        optionT (LocalAssn ** (LocalPost ** StrongInv))
     := AL.lookup (Class * Method * Label)
                  (LocalAssn ** (LocalPost ** StrongInv)).

  Inductive kindOfInstr : Set :=
   | BasicInstr
   | CallInstr
   | ReturnInstr
   | IfInstr (o:OFFSET.t)
   | GotoInstr (o:OFFSET.t)
   | SwitchInstr (* of what  ???????????????? *).

  Definition kind_of_instr (ins:Instruction) : kindOfInstr := 
   match ins with
   | Aaload             => BasicInstr
   | Aastore            => BasicInstr 
   | Aconst_null        => BasicInstr
   | Aload _            => BasicInstr 
   | Anewarray _        => BasicInstr 
   | Areturn            => ReturnInstr
   | Arraylength        => BasicInstr 
   | Astore _           => BasicInstr
   | Athrow             => BasicInstr
   | Baload             => BasicInstr 
   | Bastore            => BasicInstr
   | Checkcast _        => BasicInstr
   | Const _ _          => BasicInstr 
   | Dup                => BasicInstr 
   | Dup_x1             => BasicInstr 
   | Dup_x2             => BasicInstr 
   | Dup2               => BasicInstr 
   | Dup2_x1            => BasicInstr 
   | Dup2_x2            => BasicInstr
   | Getfield _         => BasicInstr
   | Getstatic _        => BasicInstr
   | Goto o             => GotoInstr o
   | I2b                => BasicInstr 
   | I2s                => BasicInstr
   | Ibinop _           => BasicInstr
   | Iaload             => BasicInstr 
   | Iastore            => BasicInstr 
   | If_acmp _ o        => IfInstr o
   | If_icmp _ o        => IfInstr o
   | If0 _ o            => IfInstr o
   | Ifnull _ o         => IfInstr o
   | Iinc _ _           => BasicInstr
   | Iload _            => BasicInstr
   | Ineg               => BasicInstr 
   | Instanceof _       => BasicInstr
   | Invokeinterface _  => CallInstr
   | Invokespecial _    => CallInstr
   | Invokestatic _     => CallInstr
   | Invokevirtual _    => CallInstr
   | Ireturn            => ReturnInstr
   | Istore _           => BasicInstr
   | Lookupswitch _ _   => SwitchInstr 
   | Multianewarray _ _ => BasicInstr
   | New _              => BasicInstr
   | Newarray _         => BasicInstr
   | Nop                => BasicInstr 
   | Pop                => BasicInstr 
   | Pop2               => BasicInstr 
   | Putfield _         => BasicInstr
   | Putstatic _        => BasicInstr
   | Return             => ReturnInstr
   | Saload             => BasicInstr 
   | Sastore            => BasicInstr 
   | Swap               => BasicInstr 
   | Tableswitch _ _ _ _  => SwitchInstr
   end.

  Definition isBasicInstr (M:Method) (l:Label) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | BasicInstr => True
      | _ => False
      end
    | None => False
    end.

  Definition isCallInstr (M:Method) (l:Label) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | CallInstr => True
      | _ => False
      end
    | None => False
    end.
  

  Definition isReturnInstr (M:Method) (l:Label) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | ReturnInstr => True
      | _ => False
      end
    | None => False
    end.

  Definition isIfInstr (M:Method) (l:Label) (o:OFFSET.t) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | IfInstr o' => o = o'
      | _ => False
      end
    | None => False
    end.

  Definition isGotoInstr (M:Method) (l:Label) (o:OFFSET.t) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | GotoInstr o' => o = o' 
      | _ => False
      end
    | None => False
    end.

  Definition isSwitchInstr (M:Method) (l:Label) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | SwitchInstr => True
      | _ => False
      end
    | None => False
    end.

  Inductive SP_pre (P:Program) (M:Method) (l l':Label) (A:LocalAssn) 
                   (s0:InitState) (t:LocalState) : Prop :=
   | SP_pre_def: forall s:LocalState,
        A s0 s ->
        ElementaryIntraNormalStep P M (l,s) (l',t) ->
       (**********************************)
        SP_pre P M l l' A s0 t.

  
  Definition SP_post (P:Program) (M:Method) (l l':Label) (B:LocalPost)
                    (s0:InitState) (r:LocalState) (t:TermState) : Prop :=
     forall s:LocalState, 
        ElementaryIntraNormalStep P M (l,s) (l',r) -> B s0 s t.

  Definition SP_inv (P:Program) (M:Method) (l l' :Label) (I:StrongInv)
                   (r:LocalState) (t:LocalState) : Prop:=
     forall s:LocalState, 
        ElementaryIntraNormalStep P M (l,s) (l',r) -> I s t.

 
  (* Case of execption *)
  Inductive SP_preEC (P:Program) (M:Method) (l:Label) (e:ClassName)
                     (A:LocalAssn) (s0:InitState) (t:LocalState) : Prop :=
   | SP_preEC_def:  forall  (ops:OpStack)(s:Store)(h:Heap),
                  A s0 (h,ops,s) ->   
                  forall (h2:Heap) loc2,
                  ElementaryIntraExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  Sem.Heap.typeof h2 loc2 = Some (Sem.Heap.LocationObject e) ->
                  t = (h2,Sem.Ref loc2::Sem.OperandStack.empty, s) ->
     (*************************************************)
                  SP_preEC P M l e A s0 t.

  
  Inductive SP_postEC (P:Program) (M:Method) (l:Label) (e:ClassName)
                      (B:LocalPost) (s0:InitState) (r:LocalState)
                      (t:TermState) : Prop :=
   | SP_postEC_def: forall (h2:Heap) loc2 (s:Store),
              (forall  (ops:OpStack) (h:Heap),
                  ElementaryIntraExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  Sem.Heap.typeof h2 loc2 = Some (Sem.Heap.LocationObject e) ->
                  B s0 (h,ops,s) t) -> 
                  r = (h2,Sem.Ref loc2::Sem.OperandStack.empty, s) ->
     (*************************************************)
                  SP_postEC P M l e B s0 r t.


  Inductive SP_invEC (P:Program) (M:Method) (l:Label) (e:ClassName)
                     (I:StrongInv) (r:LocalState) (t:LocalState) : Prop:=
   | SP_invEC_def:  forall (h2:Heap) loc2  (s:Store),
		  r = (h2,Sem.Ref loc2::Sem.OperandStack.empty, s) ->
                  Sem.Heap.typeof h2 loc2 = Some (Sem.Heap.LocationObject e) ->
                  (forall (ops:OpStack) (h:Heap),
                  ElementaryIntraExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  I (h,ops,s) t) ->
     (*************************************************)
                  SP_invEC P M l e I r t.

       
  (*Operators for Static invoke-non-exceptional case, non-void return value).*)
  Inductive SINV_pre (T:MethSpec) (A:LocalAssn) (N:nat) (s0:InitState)
                      (s:LocalState)  : Prop :=
    SINV_pre_def : forall (ops:OpStack) (S:Store) (h:Heap) 
                           (H:Heap) (args:OpStack) (w:option Sem.value),
           length args = N ->
           T (Sem.stack2localvar (args++ops) N, H) (h, Sem.Normal w) ->
           A s0 (H,args++ops,S) ->
           s = (h, cons_option w ops, S) ->
     (*************************************************)
           SINV_pre T A N s0 s.


  Inductive SINV_inv (T:MethSpec) (J:StrongInv) (N:nat) (s:LocalState)
                     (t:LocalState) : Prop :=
   SINV_inv_def:  (forall ops S h H args w, 
                   length args = N ->
                   T (Sem.stack2localvar (args++ops) N,H) (h, Sem.Normal w) ->
                   s=(h,cons_option w ops,S) -> J (H,args++ops,S) t) ->
     (*************************************************)
         SINV_inv T J N s t.

  Inductive SINV_post (T:MethSpec) (B:LocalPost) (N:nat) (s0:InitState)
                      (s:LocalState) (t:TermState) : Prop :=
   SINV_post_def: (forall ops S h H args w r,
                    length args = N ->
                    T (Sem.stack2localvar (args++ops) N, H)
                      (h, Sem.Normal w) ->
                    s=(h, cons_option w ops,S) ->
                    r=(H,args++ops,S) ->
                    B s0 r t) ->
     (*************************************************)
          SINV_post T B N s0 s t.


 Inductive SINV_preEC (T:MethSpec) (A:LocalAssn) (N:nat) (l:Label) 
             (e:ClassName) (s0:InitState) (s:LocalState)  : Prop :=
   | SINV_preEC_def : forall (ops:OpStack) (S:Store) (h:Heap) 
                           (H:Heap) (loc:Sem.Location)(args:OpStack),
           length args = N ->
           T (Sem.stack2localvar (args++ops) N, H) (h, Sem.Exception loc) ->
           Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
           A s0 (H,args++ops,S) ->
           s = (h,Sem.Ref loc::Sem.OperandStack.empty, S) ->
     (*************************************************)
           SINV_preEC T A N l e s0 s.


  Inductive SINV_invEC (T:MethSpec) (J:StrongInv) (N:nat)  (l:Label) 
                        (e:ClassName) (s:LocalState) (t:LocalState) : Prop :=
   | SINV_invEC_def:  
       (forall ops S h H args loc, 
          length args = N ->
          T (Sem.stack2localvar (args++ops) N,H) (h, Sem.Exception loc) ->
          Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
          s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
          J (H,args++ops,S) t) ->
     (*************************************************)
       SINV_invEC T J N l e s t.

 
  Inductive SINV_postEC (T:MethSpec) (B:LocalPost) (N:nat) (l:Label) 
            (e:ClassName) (s0:InitState) (s:LocalState) (t:TermState) : Prop :=
  | SINV_postEC_def: 
       (forall ops S h H args r loc,
          length args = N ->
          T (Sem.stack2localvar (args++ops) N, H) (h, Sem.Exception loc) ->      
          Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
          s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
          r=(H,args++ops,S) -> 
          B s0 r t) ->
     (*************************************************)
       SINV_postEC T B N l e s0 s t.

 
  (*Operators for Virtual invoke-non-exceptional case, non-void return value).*)
  Inductive VINV_pre (T:MethSpec) (A:LocalAssn) (N:nat) (s0:InitState)
                      (s:LocalState)  : Prop :=
    VINV_pre_def : forall (ops:OpStack) (S:Store) (h:Heap) 
                           (H:Heap) (args:OpStack) (w:option Sem.value) a,
           length args = N ->
           T (Sem.stack2localvar (args++(Ref a)::ops) (1+N), H) (h, Sem.Normal w) ->
           A s0 (H,args++(Ref a)::ops,S) ->
           s = (h, cons_option w ops, S) ->
     (*************************************************)
           VINV_pre T A N s0 s.


  Inductive VINV_inv (T:MethSpec) (J:StrongInv) (N:nat) (s:LocalState)
                     (t:LocalState) : Prop :=
   VINV_inv_def:  (forall ops S h H args w a, 
                   length args = N ->
                   T (Sem.stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Sem.Normal w) ->
                   s=(h,cons_option w ops,S) -> J (H,args++(Ref a)::ops,S) t) ->
     (*************************************************)
         VINV_inv T J N s t.

  Inductive VINV_post (T:MethSpec) (B:LocalPost) (N:nat) (s0:InitState)
                      (s:LocalState) (t:TermState) : Prop :=
   VINV_post_def: (forall ops S h H args w r a,
                    length args = N ->
                    T (Sem.stack2localvar (args++(Ref a)::ops) N, H)
                      (h, Sem.Normal w) ->
                    s=(h, cons_option w ops,S) ->
                    r=(H,args++(Ref a)::ops,S) ->
                    B s0 r t) ->
     (*************************************************)
          VINV_post T B N s0 s t.


 Inductive VINV_preEC (T:MethSpec) (A:LocalAssn) (N:nat) (l:Label) 
             (e:ClassName) (s0:InitState) (s:LocalState)  : Prop :=
   | VINV_preEC_def : forall (ops:OpStack) (S:Store) (h:Heap) 
                           (H:Heap) (loc:Sem.Location)(args:OpStack) a,
           length args = N ->
           T (Sem.stack2localvar (args++(Ref a)::ops) N, H) (h, Sem.Exception loc) ->
           Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
           A s0 (H,args++(Ref a)::ops,S) ->
           s = (h,Sem.Ref loc::Sem.OperandStack.empty, S) ->
     (*************************************************)
           VINV_preEC T A N l e s0 s.


  Inductive VINV_invEC (T:MethSpec) (J:StrongInv) (N:nat)  (l:Label) 
                        (e:ClassName) (s:LocalState) (t:LocalState) : Prop :=
   | VINV_invEC_def:  
       (forall ops S h H args loc a, 
          length args = N ->
          T (Sem.stack2localvar (args++(Ref a)::ops) N,H) (h, Sem.Exception loc) ->
          Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
          s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
          J (H,args++(Ref a)::ops,S) t) ->
     (*************************************************)
       VINV_invEC T J N l e s t.

 
  Inductive VINV_postEC (T:MethSpec) (B:LocalPost) (N:nat) (l:Label) 
            (e:ClassName) (s0:InitState) (s:LocalState) (t:TermState) : Prop :=
  | VINV_postEC_def: 
       (forall ops S h H args r loc a,
          length args = N ->
          T (Sem.stack2localvar (args++(Ref a)::ops) N, H) (h, Sem.Exception loc) ->      
          Sem.Heap.typeof h loc = Some (Sem.Heap.LocationObject e) ->
          s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
          r=(H,args++(Ref a)::ops,S) -> 
          B s0 r t) ->
     (*************************************************)
       VINV_postEC T B N l e s0 s t.

 Inductive VINV_preECNull (P:Program) (A:LocalAssn) (N:nat)
                          (s0:InitState) (s:LocalState)  : Prop :=
   | VINV_preECNull_def : forall (ops:OpStack) (S:Store) (h:Heap) 
                           (H:Heap) (loc:Sem.Location)(args:OpStack),
           length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           A s0 (H,args++Null::ops,S) ->
           s = (h,Sem.Ref loc::Sem.OperandStack.empty, S) ->
     (*************************************************)
           VINV_preECNull P A N s0 s.

  Inductive VINV_invECNull (P:Program) (J:StrongInv) (N:nat)
                           (s:LocalState) (t:LocalState) : Prop :=
   | VINV_invECNull_def:  
       (forall ops S h H args loc, 
          length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
           J (H,args++Null::ops,S) t) ->
     (*************************************************)
       VINV_invECNull P J N s t.

 
  Inductive VINV_postECNull (P:Program) (B:LocalPost) (N:nat) 
                            (s0:InitState) (s:LocalState) (t:TermState) : Prop :=
  | VINV_postECNull_def: 
       (forall ops S h H args r loc,
           length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           s=(h,Sem.Ref loc::Sem.OperandStack.empty, S) -> 
           r=(H,args++Null::ops,S) -> 
           B s0 r t) ->
     (*************************************************)
       VINV_postECNull P B N s0 s t.

 Section SP_Judgement.
  Variable G   : Assum_TYPE.
  Variable Anno: AnnoType.
  Variable MST : MSPEC_TYPE.
  Variable P   : Program.

  Section METHOD.
  Variable C: Class.  
  Variable M: Method. 

  Inductive SP_Judgement :
     bool -> Label -> LocalAssn -> LocalPost -> StrongInv -> Prop :=

  | SP_INSTR : forall  (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv),

      isBasicInstr M l ->
      (* Local Invariant is satisfied *)
      (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Next instruction is valid, 3 cases : *)
         (* 1 - Normal execution, SP_pre, SP_post, SP_inv use
            ElementaryIntraNormalStep who does not contain the case of
            exception *)
         (forall ll, next M l = Some ll -> 
              SP_Judgement true ll (SP_pre P M l ll A)
                                   (SP_post P M l ll B)
                                   (SP_inv P M l ll I)) ->
   
         (* 2 - Exception caught by the handler *)
         (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                             l e ll ->
              SP_Judgement true ll (SP_preEC P M l e A)
                                    (SP_postEC P M l e B)
                                    (SP_invEC P M l e I)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 s, A s0 s ->
            forall h2 loc2,
               ElementaryIntraExceptionStep P M (l,s) (h2,loc2) ->
               Sem.UnCaughtException  P M (l,h2,loc2) ->
               B s0 s (h2,Sem.Exception loc2)) ->
     (********************************************************************)
      SP_Judgement false l A B I

  | SP_RETURN : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv),
      isReturnInstr M l ->
      (* Local Invariant is satisfied *)
      (forall Q, Anno_LOOKUP Anno (C,M,l) = SomeT Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      (forall s0 s, A s0 s ->
         forall h2 ov, ElementaryReturnStep P M (l,s) (h2,Sem.Normal ov) ->
         B s0 s (h2,Sem.Normal ov)) ->
      (********************************************************************)
      SP_Judgement false l A B I

   | SP_GOTO : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                    (o:OFFSET.t),   
      instructionAt M l = Some (Goto o) ->
    
      (* Local Assertion is satisfied *) 
      (forall Q, Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->

      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      SP_Judgement true (OFFSET.jump l o)
                        (SP_pre P M l (OFFSET.jump l o) A)
                        (SP_post P M l (OFFSET.jump l o) B)
                        (SP_inv P M l (OFFSET.jump l o) I) ->
    (********************************************************************)
      SP_Judgement false l A B I

  | SP_IF : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv)
                 (ll:Label)  (o:OFFSET.t),
      isIfInstr M l o ->
      next M l = Some ll ->
      (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->
      (forall s0 s, A s0 s -> I s s) ->

      SP_Judgement true (OFFSET.jump l o)
        (SP_pre P M l (OFFSET.jump l o) A)
        (SP_post P M l (OFFSET.jump l o) B)
        (SP_inv P M l (OFFSET.jump l o) I) -> 
      SP_Judgement true ll
        (SP_pre P M l ll A)
        (SP_post P M l ll B)
        (SP_inv P M l ll I) ->
    (********************************************************************)
      SP_Judgement false l A B I

  | SP_INVS: forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                 (msig:MethodSignature) (m:Method)
                 (T:MethSpec) (MI:MethInv) (N:nat),

     instructionAt M l = Some(Invokestatic msig) ->
     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP MST m = SomeT (T ,, MI) ->
     (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters msig) ->
     (forall s0 args ops S H t , length args = N ->
                                 A s0 (H,args++ops,S) ->
                                 MI (Sem.stack2localvar (args++ops) N,H) t -> 
                                 I (H,args++ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
           SP_Judgement true ll (SINV_pre T A N)
                                (SINV_post T B N)
                                (SINV_inv T I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l e ll ->
              SP_Judgement true ll (SINV_preEC T A N l e)
                                    (SINV_postEC T B N l e)
                                    (SINV_invEC T I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H, A s0 (H,args++ops,S) ->
            forall h loc,
              length args = N ->
              T (Sem.stack2localvar (args++ops) N,H) (h, Sem.Exception loc) ->
              Sem.UnCaughtException  P M (l,h,loc) ->
              B s0  (H,args++ops,S) (h,Sem.Exception loc)) ->
     (*******************************************************************)
      SP_Judgement false l A B I


(*********************************************************************)
(***********************INVOKEVIRTUAL**********************************)
(*ToDo: Add case for NullPointerException (see BigStepLoad.ElementarIntraJVMExceptionStep*)

  | SP_INVV: forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                 (msig:MethodSignature) (m:Method)
                 (T:MethSpec) (MI:MethInv) (N:nat),

     instructionAt M l = Some(Invokevirtual msig) ->
     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP MST m = SomeT (T ,, MI) ->
     (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters msig) ->
     (forall s0 args ops S H t a, length args = N ->
                                 A s0 (H,args++(Ref a)::ops,S) ->
                                 MI (Sem.stack2localvar (args++(Ref a)::ops) (1+N),H) t -> 
                                 I (H,args++(Ref a)::ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
           SP_Judgement true ll (VINV_pre T A N)
                                (VINV_post T B N)
                                (VINV_inv T I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l e ll ->
              SP_Judgement true ll (VINV_preEC T A N l e)
                                    (VINV_postEC T B N l e)
                                    (VINV_invEC T I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H a, A s0 (H,args++(Ref a)::ops,S) ->
            forall h loc,
              length args = N ->
              T (Sem.stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Sem.Exception loc) ->
              Sem.UnCaughtException  P M (l,h,loc) ->
              B s0  (H,args++(Ref a)::ops,S) (h,Sem.Exception loc)) ->

     (*Nullpointerexception*)
         (forall s0 args ops S H , A s0 (H,args++Null::ops,S) ->
            forall h loc,
              length args = N -> 
              Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
              ( (forall ll bm, METHOD.body M = Some bm ->
                               Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l (javaLang,NullPointerException) ll ->
                               SP_Judgement true ll (VINV_preECNull P A N)
                                                    (VINV_postECNull P B N)
                                                    (VINV_invECNull P I N)) /\
               (Sem.UnCaughtException P M (l,h,loc) ->
                B s0 (H,args++Null::ops,S) (h,Sem.Exception loc)))) ->
           

     (*******************************************************************)
      SP_Judgement false l A B I

(*********************************END OF INVOKEVIRTUAL******************)
(*********************************************************************)

  | SP_CONSEQ : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                    (b:bool) (AA:LocalAssn) (BB:LocalPost) (II:StrongInv) ,
      SP_Judgement b l AA BB II ->
      (forall s0 s , A s0 s -> AA s0 s) ->
      (forall s0 s t , BB s0 s t -> B s0 s t) ->
      (forall t s , II t s -> I t s) ->
  (********************************************************************)
      SP_Judgement b l A B I 

  | SP_INJECT : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv),
      SP_Judgement false l A B I ->
  (********************************************************************)
      SP_Judgement true l A B I

  | SP_AX : forall (l:Label) (A:LocalAssn) (B:LocalPost) (I:StrongInv), 
      ASSUM_LOOKUP G (C,M,l) = SomeT (A ,, (B ,, I)) ->
      (forall Q, Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                                   (forall s0 s, A s0 s -> Q s0 s)) ->
      (forall s0 s, A s0 s -> I s s) ->
  (********************************************************************)
      SP_Judgement true l A B I.



 Inductive SP_Judgement_call (l:Label) (msig:MethodSignature)(AA:LocalAssn)
                             (BB:LocalPost)(II:StrongInv): Prop :=
 | SP_INVS_call: forall (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                  (m:Method) (T:MethSpec) (MI:MethInv) (N:nat),
     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP MST m = SomeT (T ,, MI) ->
     (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters msig) ->
     (forall s0 args ops S H t , length args = N ->
                                 A s0 (H,args++ops,S) ->
                                 MI (Sem.stack2localvar (args++ops) N,H) t -> 
                                 I (H,args++ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
              SP_Judgement true ll (SINV_pre T A N)
                                   (SINV_post T B N)
                                   (SINV_inv T I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                       l e ll ->
              SP_Judgement true ll (SINV_preEC T A N l e)
                                    (SINV_postEC T B N l e)
                                    (SINV_invEC T I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H, A s0 (H,args++ops,S) ->
            forall h loc,
                length args = N ->
                T (Sem.stack2localvar (args++ops) N,H) (h, Sem.Exception loc) ->
                Sem.UnCaughtException  P M (l,h,loc) ->
                B s0  (H,args++ops,S) (h,Sem.Exception loc)) ->
     (*******************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
    SP_Judgement_call l msig AA BB II.

 Inductive SP_Judgement_virt_call (l:Label) (msig:MethodSignature)(AA:LocalAssn)
                                  (BB:LocalPost)(II:StrongInv): Prop :=
 | SP_INVV_call: forall (A:LocalAssn) (B:LocalPost) (I:StrongInv) 
                  (m:Method) (T:MethSpec) (MI:MethInv) (N:nat),
     findMethod P msig = Some m ->
     MSPEC_LOOKUP MST m = SomeT (T ,, MI) ->
     (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters msig) ->
     (forall s0 args ops S H t a , length args = N ->
                                   A s0 (H,args++(Ref a)::ops,S) ->
                                   MI (Sem.stack2localvar (args++(Ref a)::ops) (1+N),H) t -> 
                                   I (H,args++(Ref a)::ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
              SP_Judgement true ll (VINV_pre T A N)
                                   (VINV_post T B N)
                                   (VINV_inv T I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                       l e ll ->
              SP_Judgement true ll (VINV_preEC T A N l e)
                                    (VINV_postEC T B N l e)
                                    (VINV_invEC T I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H a, A s0 (H,args++(Ref a)::ops,S) ->
            forall h loc,
                length args = N ->
                T (Sem.stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Sem.Exception loc) ->
                Sem.UnCaughtException  P M (l,h,loc) ->
                B s0  (H,args++(Ref a)::ops,S) (h,Sem.Exception loc)) ->

     (*Nullpointerexception*)
         (forall s0 args ops S H , A s0 (H,args++Null::ops,S) ->
            forall h loc,
              length args = N -> 
              Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
              ( (forall ll bm, METHOD.body M = Some bm ->
                               Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l (javaLang,NullPointerException) ll ->
                               SP_Judgement true ll (VINV_preECNull P A N)
                                                    (VINV_postECNull P B N)
                                                    (VINV_invECNull P I N)) /\
               (Sem.UnCaughtException P M (l,h,loc) ->
                B s0 (H,args++Null::ops,S) (h,Sem.Exception loc)))) ->

     (*******************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
    SP_Judgement_virt_call l msig AA BB II.

 Inductive SP_Judgement_basic (l:Label) (AA:LocalAssn) (BB:LocalPost) (II:StrongInv) : Prop :=
  | SP_INSTR_basic : forall  (A:LocalAssn) (B:LocalPost) (I:StrongInv),

      (* Local Invariant is satisfied *)
      (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Next instruction is valid, 3 cases : *)
         (* 1 - Normal execution, SP_pre, SP_post, SP_inv use
            ElementaryIntraNormalStep who does not contain the case of
            exception *)
         (forall ll, next M l = Some ll -> 
              SP_Judgement true ll (SP_pre P M l ll A)
                                   (SP_post P M l ll B)
                                   (SP_inv P M l ll I)) ->
   
         (* 2 - Exception caught by the handler *)
         (forall ll e bm, 
              METHOD.body M = Some bm ->
              Sem.lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                                 l e ll ->
              SP_Judgement true ll (SP_preEC P M l e A)
                                    (SP_postEC P M l e B)
                                    (SP_invEC P M l e I)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 s, A s0 s ->
            forall h2 loc2,
               ElementaryIntraExceptionStep P M (l,s) (h2,loc2) ->
               Sem.UnCaughtException  P M (l,h2,loc2) ->
               B s0 s (h2,Sem.Exception loc2)) ->
     (********************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->

      SP_Judgement_basic l AA BB II.

 Inductive SP_Judgement_return (l:Label) (AA:LocalAssn) (BB:LocalPost)
                               (II:StrongInv) : Prop := 
  | SP_RETURN_return : forall (A:LocalAssn) (B:LocalPost) (I:StrongInv),
      (* Local Invariant is satisfied *)
      (forall Q, Anno_LOOKUP Anno (C,M,l) = SomeT Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      (forall s0 s, A s0 s ->
         forall h2 ov, ElementaryReturnStep P M (l,s) (h2,Sem.Normal ov) ->
         B s0 s (h2,Sem.Normal ov)) ->
      (********************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_return l AA BB II.

 Inductive SP_Judgement_goto (l:Label)(o:OFFSET.t)(AA:LocalAssn)
                             (BB:LocalPost) (II:StrongInv) : Prop := 

   | SP_GOTO_goto : forall (A:LocalAssn) (B:LocalPost) (I:StrongInv) ,
      (* Local Assertion is satisfied *) 
      (forall Q, Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->

      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      SP_Judgement true (OFFSET.jump l o)
                        (SP_pre P M l (OFFSET.jump l o) A)
                        (SP_post P M l (OFFSET.jump l o) B)
                        (SP_inv P M l (OFFSET.jump l o) I) ->
    (********************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_goto l o AA BB II.

 Inductive SP_Judgement_if (l:Label)(o:OFFSET.t)(AA:LocalAssn) (BB:LocalPost)
                           (II:StrongInv) : Prop := 
  | SP_IF_if : forall (A:LocalAssn) (B:LocalPost) (I:StrongInv) (ll:Label),
      next M l = Some ll ->
      (forall Q , Anno_LOOKUP Anno (C,M,l) = SomeT Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->
      (forall s0 s, A s0 s -> I s s) ->

      SP_Judgement true (OFFSET.jump l o)
        (SP_pre P M l (OFFSET.jump l o) A)
        (SP_post P M l (OFFSET.jump l o) B)
        (SP_inv P M l (OFFSET.jump l o) I) -> 
      SP_Judgement true ll
        (SP_pre P M l ll A)
        (SP_post P M l ll B)
        (SP_inv P M l ll I) ->
    (********************************************************************)
      (forall s0 s , AA s0 s -> A s0 s) ->
      (forall s0 s t , B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
     SP_Judgement_if l o AA BB II.


 End METHOD.

 Definition StrongInvConversion (I:MethInv) : StrongInv :=
    fun s t => (forall S H , s=(H,Sem.OperandStack.empty,S) -> I (S,H) t).

 

 Definition GP : Prop :=
    (forall  C M l A B I, 
       ASSUM_LOOKUP G (C,M,l) = SomeT (A ,, (B ,, I)) -> 
       SP_Judgement C M false l A B I)
    /\
    (forall M MB T MI l, 
       METHOD.body M = Some MB ->
       MSPEC_LOOKUP MST M = SomeT (T ,, MI) ->
       BYTECODEMETHOD.firstAddress MB = l ->
       exists C, 
       SP_Judgement C M false l
          (fun s0 s => let (S,H) := s0 in s = (H,Sem.OperandStack.empty,S))
          (fun s0 s t => let (S,H):= s0 in s = (H,Sem.OperandStack.empty,S) ->
                 T (S,H) t)
          (StrongInvConversion MI)).

 
