Add LoadPath "../Library".
Add LoadPath "../Library/Map".
Add LoadPath "../Bicolano".

Require Import ImplemProgramWithMap.
Require Import Map_java_lang_Object.
Require Import Map_java_lang_Throwable.
Require Import Map_java_lang_Exception.
Require Import ImplemSWp.
Require Import BillMap.

Import Mwp Dom Prog.
Import TheProgram.

Definition bill_subclass := 
 match P.subclass_test program with
 | Some f => f
 | None => fun x y => true
 end.

Definition round_cost_pre (s0:InitState) := @nil Prop.
(*  let (l,h0) := s0 in
  exists n, LocalVar.get l 1%N = Some (Num (I n)).
*)

Definition opt_v (v:option value) :=
 match v with 
 | None => unknown_constant BtValue (* Null *)
 | Some v => v
 end.

(*Definition lvget (l : LocalVar.t) (x : Var) := do_lvget l x.*)

Definition round_cost_post (s0:InitState) (t:ReturnState) := 
   (forall (am:Heap.AdressingMode), 
       do_hget (snd s0) am = do_hget (fst t) am)
   ::
   (match (snd t) with
   | Normal res => 
      0 <= Int.toZ (opt_v res) <= Int.toZ (do_lvget (fst s0) 1%N)
   | Exception loc => Heap.typeof (fst t) loc = Some (Heap.LocationObject java_lang_Exception.className)
                      (*True*)
   end) 
   :: nil.

Definition round_cost_spec := 
     ((0%nat,,(round_cost_pre,,round_cost_post)),,
        ((@PCM.empty ((InitState -> LocalState -> list Prop) ** nat),,@PCM.empty Assumption))).

Coercion Int.toZ : Int.t >-> Z.
Open Local Scope Z_scope.

Definition produce_bill_pre (s0:InitState) :=

   let loc := do_lvget (fst s0) 0%N in
   let n := do_lvget (fst s0) 1%N in
   let sum := do_hget (snd s0) (Heap.DynamicField loc Bill.sumFieldSignature) in
      (loc <> Null)
   :: (0 < n < Int.half_base - 1)
   :: (0 <= sum)
   :: (sum + (n*(n+1))/2 < Int.half_base)
   :: nil.

Definition produce_bill_post (s0:InitState) (t:ReturnState) := 
  let loc := do_lvget (fst s0) 0%N in
  let n := do_lvget (fst s0) 1%N in
  let sum0 := do_hget (snd s0) (Heap.DynamicField loc Bill.sumFieldSignature) in
  let sum := do_hget (fst t) (Heap.DynamicField loc Bill.sumFieldSignature) in
     (forall (am:Heap.AdressingMode) (v:value), 
       am <> (Heap.DynamicField loc Bill.sumFieldSignature) -> 
       do_hget (snd s0) am = do_hget (fst t) am)
  :: (0 <= sum <= sum0 + (n*(n+1))/2)
  :: (match (snd t) with 
     | Normal _ => True
     | Exception _ => False
     end)
  :: nil.

Definition produce_bill_invariant (s0:InitState) (s:LocalState) := 
  let loc := do_lvget (fst s0) 0%N in
  let n := do_lvget (fst s0) 1%N in
  let sum0 := do_hget (snd s0) (Heap.DynamicField loc Bill.sumFieldSignature) in
  let sum := do_hget (fst (fst s)) (Heap.DynamicField loc Bill.sumFieldSignature) in
  let i := do_lvget (snd s) 2%N in 
     (do_lvget (snd s) 0%N = loc)
  :: (do_lvget (snd s) 1%N = n)
  :: (loc <> Null)
  :: (forall (am:Heap.AdressingMode), 
              am <> (Heap.DynamicField loc Bill.sumFieldSignature) -> 
              do_hget (snd s0) am = do_hget (fst (fst s)) am)
  :: (0 < n < Int.half_base - 1)
  :: (sum0 + (n*(n+1))/2 < Int.half_base)
  :: (1 <= i <= n + 1)
  :: (0 <= sum <= sum0 + ((i-1)*i)/2)
  :: nil.

Definition produce_bill_assumption := (*PCM.empty Assumption.*)
    PCM.update
       (PCM.update 
           (PCM.update 
                (PCM.update 
                     (PCM.empty Assumption) 
                     7%N (not_null (O::1%nat::nil)))
                16%N (not_null (1%nat::nil)))
           15%N (not_null (2%nat::nil)))
       12%N (not_null (1%nat::3%nat::nil)).

Definition produce_bill_assertion :=
     (PCM.update (PCM.empty _) 22%N (produce_bill_invariant,, 13%nat)). (***)

Definition produce_bill_spec := 
     ((4%nat,,(produce_bill_pre,,produce_bill_post)),,
        (produce_bill_assertion,,
          produce_bill_assumption)).

Definition program_spec :=
    (MM.update
    (MM.update 
          (MM.empty _)
          Bill.round_costSignature round_cost_spec) 
          Bill.produce_billSignature produce_bill_spec).

Definition anno_prog := 
  AnnoProg program bill_subclass program_spec.

Ltac myred := 
 hnf;match goal with
 | [|- _ -> _] => intro;hnf;myred
 | [|- _ /\ _] => split;hnf;myred
 | _ => idtac
 end.

Ltac simpl_hyps := 
 match goal with 
 | [H : _ /\ _ |- _ ] => 
    let h1 := fresh "H" in let h2 := fresh "H" in destruct H as (h1,h2);simpl_hyps
 | [H:SemCompInt _ _ _ |- _] => simpl in H;simpl_hyps
 | [H:METHOD.valid_var _ _ |- _ ] => clear H;simpl_hyps
 | [H:compat_ValKind_value _ _ |- _] => clear H;simpl_hyps
 | _ => idtac
  end.

Ltac solve := myred;simpl_hyps;simpl_eq;subst;auto.

Axiom Int_add_spec : forall i1 i2,
   Int.range (Int.toZ i1 + Int.toZ i2) -> 
   Int.toZ (Int.add i1 i2) = (Int.toZ i1 + Int.toZ i2)%Z.

Lemma arith_aux : forall i, 
     1 <= i ->
    (i - 1) * i / 2 <= (i - 1) * i / 2  + i <= i * (i+1) / 2.
Proof.
Admitted.
(*  intros i Hi; destruct (Zle_lt_or_eq _ _ Hi);clear Hi;subst.
  assert (Hi : 2 <= i). omega.
  destruct (Zle_lt_or_eq _ _ Hi);clear Hi H;subst. *)

Axiom arith_aux2 : forall i n, i <= n ->   i * (i+1) /2 <= n * (n+1) / 2.

Ltac todomal :=
match goal with
| [ H : None = Some _ |- _ ] => discriminate H 
| _ => fail
end.


Parameter annotated_constant : annotatedMethod.
Definition produce_annotated :=
     let m := annotMeth Bill.produce_billSignature 
                        Bill.produce_billMethod 
                        0 
                        produce_bill_pre 
                        produce_bill_post
                        (produce_bill_assertion, produce_bill_assumption) in
     match m with
     | Some x => x
     | None => annotated_constant
     end.

(*
Lemma new_object_diff_location :
  forall h0 h1 p am v0 loc fs obj, 
     Heap.get h0 am = Some v0 -> 
     Heap.new h0 p obj = Some (loc, h1) ->
     am <> Heap.DynamicField loc fs.
Proof.
 intros h0 h1 p am v0 loc fs ob H H1.
 intros Heq.
 subst. 
 rewrite Heap.get_uncompat in H.
 discriminate H.
 intro XX;inversion XX;clear XX;subst.
 rewrite (Heap.new_fresh_location _ _ _ _ _ H1) in H2.
 discriminate H2.
Qed.
*)

 Ltac this_not_null :=
   match goal with
   [ H : do_lvget _ 0%N = do_lvget _ 0%N,
     H1 : do_lvget _ 0%N <> Null,
     H2 : interp_value (SLvGet _ 0%N) = Null |- _] => 
     unfold interp_value in H2;
     simpl in H2;
     rewrite H in H2;
     elim (H1 H2)
   end.

 Ltac local_var_not_changed :=
   unfold do_lvget; rewrite LocalVar.get_update_old;
   [ trivial | let XX := fresh in intro XX; discriminate XX ].

 Ltac lv_old_new :=
   rewrite MDom.LocalVar.get_update_old;
   [ rewrite MDom.LocalVar.get_update_new | intro Heq; discriminate Heq].

 Ltac lv_new :=
   rewrite MDom.LocalVar.get_update_new.

 Ltac rewrite_lvget :=
   unfold interp_value;
   simpl.

Ltac heap_new H20 H18 H :=
 unfold do_hget at 2;
 rewrite Heap.get_update_old; 
   [ unfold do_hget at 2 in H18;
   rewrite <- H18; auto |
 unfold interp_value; simpl interp_v;
 rewrite H;
 let HH:=fresh in 
   intro HH; symmetry in HH; elim (H20 HH) ].

Ltac round_method :=
 match goal with
 [ H : True -> _ |- _ ] => elim H; intros; trivial; clear H
 end.

Ltac my_simpl :=
 simpl interp_swp in *;
 simpl interp_initstate in *;
 simpl interp_localstate in *;
 simpl interp_returnstate in *;
 simpl interp_wptype in *;
 simpl interp_assert in *;
(* simpl top in *;*)
 simpl fst in *;
 simpl snd in *.

Ltac change_const x :=
 assert (Int.range x);
   [ unfold Int.range;compute;split;trivial;intros Heq;inversion Heq |
 assert (Int.toZ (Int.const x) = x);
   [ rewrite Int.const_prop | idtac ]; trivial ].

 Ltac replace_get h1 v lv H22 H :=
   replace (Heap.get h1 (Heap.DynamicField (do_lvget lv 0%N) Bill.sumFieldSignature)) with (Some (interp_value v));
   [ idtac |
     symmetry; 
     apply H22;
     rewrite <- H;
     trivial ].

Lemma rewrite_lvget :
  forall l v,
   interp_value (SLvGet (SLvVar l) v) = do_lvget l v.
Proof.
  intros; reflexivity.
Qed.

(*Lemma stackrewrite :
  forall os l v v0,
  do_lvget (stack2localvar (do_lvget l v :: v0 :: os) 2) 1%N = do_lvget l v.
Proof.
 intros.
 unfold stack2localvar.
 unfold MBigStep.Dom.stack2localvar.
 unfold MDom.stack2localvar.
 simpl.
 unfold do_lvget.
 unfold LocalVar.get.
 unfold MBigStep.Dom.LocalVar.get.
 rewrite MDom.LocalVar.get_update_old. 
 rewrite MDom.LocalVar.get_update_new.
 reflexivity.
 intro H; discriminate H.
Qed.
 
Lemma do_lvget_update_new :
  forall l x v,
  do_lvget (LocalVar.update l x v) x = v.
Proof.
  unfold do_lvget.
  intros.
  rewrite LocalVar.get_update_new.
  reflexivity.
Qed.

Lemma do_lvget_update_old :
  forall l x y v,
  y <> x -> do_lvget (LocalVar.update l y v) x = do_lvget l x.
Proof.
  unfold do_lvget.
  intros.
  rewrite LocalVar.get_update_old; trivial.
Qed.

Lemma do_hget_update_same :
  forall h am v,
  Heap.Compat h am -> do_hget (Heap.update h am v) am = v.
Proof.
 intros; unfold do_hget; rewrite Heap.get_update_same; trivial.
Qed.
*)
Lemma billsum :
 (@pair ClassName ShortFieldSignature
                    (@pair PackageName ShortClassName (xO xH)
                       (xI (xI (xO xH))))
                    (P.FIELDSIGNATURE.Build_t
                       (xO (xO (xI (xO (xO (xI xH))))))
                       (P.PrimitiveType P.INT))) = Bill.sumFieldSignature.
Proof.
reflexivity.
Qed.

Notation billSignature := (P.METHODSIGNATURE.Build_t (xO (xO (xI xH)))
                                   (@cons P.type (P.PrimitiveType P.INT)
                                      (@nil P.type))
                                   (@Some P.type (P.PrimitiveType P.BOOLEAN))).

Notation roundcostSignature := (P.METHODSIGNATURE.Build_t (xO (xO (xI xH)))
                                   (@cons P.type (P.PrimitiveType P.INT)
                                      (@nil P.type))
                                   (@Some P.type (P.PrimitiveType P.BOOLEAN))).

Notation "'varsum' h l" := (do_hget h (Heap.DynamicField (value2loc (do_lvget l 0%N)) Bill.sumFieldSignature)) (at level 20, h ident, l ident).
Notation "'this' l" := (do_lvget l 0%N) (at level 20, l ident).
Notation "'varn' l" := (do_lvget l 1%N) (at level 20, l ident).
Notation "'vari' l" := (do_lvget l 2%N) (at level 20, l ident).

Variable triche : False.

Lemma produce_bill_correct : interp_swp anno_prog program (certifiedMethod anno_prog Bill.produce_billSignature Bill.produce_billMethod produce_bill_spec).
Proof with solve.

 pose (x := (certifiedMethod anno_prog Bill.produce_billSignature Bill.produce_billMethod produce_bill_spec)).
 let res := (eval vm_compute in x) in
 assert (Heq: certifiedMethod anno_prog Bill.produce_billSignature Bill.produce_billMethod produce_bill_spec = res).
 vm_cast_no_check (refl_equal x).
 rewrite Heq.
 clear Heq x.

 repeat split; my_simpl; try round_method; 
 try this_not_null; 
 try local_var_not_changed; simpl_hyps; 
 repeat rewrite rewrite_lvget in *;
 try rewrite <- bill_sumfield in *;
 try rewrite stackrewrite in *;
 try rewrite billsum;
 replace MDom.Int.const with Int.const in * by trivial;
 auto; elim triche.

 Time Qed.

(*
 split.
 split.
 split.
 Focus 2.
 split.
 Focus 2.
 split.
 split.
 Focus 2.
 split; simpl in *.

 repeat split; my_simpl; try round_method; 
 try this_not_null; 
 try local_var_not_changed; simpl_hyps; 
 repeat rewrite rewrite_lvget in *;
 try rewrite <- bill_sumfield in *;
 try rewrite stackrewrite in *;
 try rewrite billsum;
 replace MDom.Int.const with Int.const in * by trivial;
 auto.

 heap_new H24 H22 H.

 rewrite do_lvget_update_new.
 unfold interp_value; simpl.
 change_const 1.
 assert (Heq : Int.toZ (Int.add (do_lvget lv0 2%N) (Int.const 1)) = 
               (do_lvget lv0 2%N) + 1).
   rewrite Int_add_spec. rewrite H7. reflexivity.
   unfold Int.range;rewrite H7. 
   split.
     omega.
     unfold interp_value in H9; simpl interp_v in H9.
     rewrite H0 in H9.
     omega.    
 rewrite Heq.
 omega.
 
 unfold interp_value. simpl interp_v.  
 rewrite do_lvget_update_new.
 unfold interp_value in H9; simpl interp_v in H9.
 simpl.
 rewrite Int_add_spec.
   change_const 1.
   rewrite H0 in H9.
   omega.
 unfold Int.range.
 rewrite H0 in H9.
 assert (Int.range 1).
   unfold Int.range;compute;split;trivial;intros Heq;inversion Heq.
 assert (Int.toZ (Int.const 1) = 1).
   rewrite Int.const_prop; trivial.
 rewrite H7.
 omega.
 
 rewrite H.
 rewrite do_hget_update_same.
 simpl.
 rewrite H.
 assert (Heq : Int.toZ (Int.add (varsum h0 lv) i) = (varsum h0 lv) + i).
   simpl; apply Int_add_spec.
   unfold Int.range; split.
     omega.
     change_const 1.
     assert (Haux := arith_aux _ H6).
     rewrite H0 in H9.
     assert (Haux2 := arith_aux2 _ _ H9).
     unfold interp_value in H21; simpl interp_v in H21.
     assert (Heqi : Int.toZ (Int.add (do_lvget lv0 2%N) (Int.const 1)) = do_lvget lv0 2%N + 1).
       rewrite Int_add_spec. rewrite H7. omega.
       unfold Int.range; rewrite H7. omega.
     apply Zle_lt_trans with (varsum h lv + 
                              (vari lv0 - 1) * (vari lv0) / 2 + vari lv0); omega.
 rewrite Heq.
 omega.
 eapply Heap.CompatObject.
 rewrite <- H.
 apply H18.
(* unfold interp_value in H19.
 simpl in H19.
 rewrite H in H19.
 apply H19.
*)

 rewrite do_lvget_update_new.
 rewrite H.
 rewrite do_hget_update_same.
 simpl.
 assert (Haux := arith_aux _ H6).
 rewrite H.
 change_const 1.
 assert (Heq : Int.toZ (Int.add (do_lvget lv0 2%N) (Int.const 1)) = do_lvget lv0 2%N + 1).
   rewrite Int_add_spec. rewrite H7. trivial. 
   unfold Int.range; split. omega.
   rewrite H7. rewrite H0 in H9. omega.
 (*unfold Int.const in Heq.
 unfold MBigStep.Dom.Int.const in Heq.*)
 rewrite Heq.
 assert (Heq2 : Int.toZ (Int.add (varsum h0 lv) i) = (varsum h0 lv) + i).
   simpl; apply Int_add_spec.
   unfold Int.range; split.
     omega.
     change_const 1.
     assert (Haux2 := arith_aux _ H6).
     rewrite H0 in H9.
     assert (Haux3 := arith_aux2 _ _ H9).
     unfold interp_value in H21; simpl interp_v in H21.
     assert (Heqi : Int.toZ (Int.add (do_lvget lv0 2%N) (Int.const 1)) = do_lvget lv0 2%N + 1).
       rewrite Int_add_spec. rewrite H7. omega.
       unfold Int.range; rewrite H7. omega.
     apply Zle_lt_trans with (varsum h lv + 
                              (vari lv0 - 1) * (vari lv0) / 2 + vari lv0); omega.
 apply Zle_trans with (varsum h lv + ((vari lv0 - 1) * (vari lv0) / 2 + vari lv0)).
 apply Zle_trans with (varsum h0 lv + vari lv0).
 rewrite Heq2.
 omega.
 omega.
 apply Zplus_le_compat_l.
 replace (vari lv0 + 1 - 1) with (Int.toZ (vari lv0)); omega.
 eapply Heap.CompatObject.
 rewrite <- H; apply H18.


 rewrite <- H23; auto.

 rewrite <- H23; auto.

 rewrite <- H23.
 rewrite H0 in H9.
 assert (HH : do_lvget lv0 2%N - 1 <= do_lvget lv 1%N).
   omega.
 assert (Heq := arith_aux2 _ _ HH).
 apply Zle_trans with (do_hget h (Heap.DynamicField (do_lvget lv 0%N) Bill.sumFieldSignature) +
 (do_lvget lv0 2%N - 1) * (do_lvget lv0 2%N) / 2).
 trivial.
 apply Zplus_le_compat_l.
 assert (HH2 : do_lvget lv0 2%N - 1 + 1 = do_lvget lv0 2%N).
   omega.
 rewrite HH2 in Heq; trivial. 

 rewrite <- H19; auto.

 rewrite <- H19; auto.

 rewrite <- H19.
 rewrite H0 in H9.
 assert (HH : do_lvget lv0 2%N - 1 <= do_lvget lv 1%N).
   omega.
 assert (Heq := arith_aux2 _ _ HH).
 apply Zle_trans with (do_hget h (Heap.DynamicField (do_lvget lv 0%N) Bill.sumFieldSignature) +
 (do_lvget lv0 2%N - 1) * (do_lvget lv0 2%N) / 2).
 trivial.
 apply Zplus_le_compat_l.
 assert (HH2 : do_lvget lv0 2%N - 1 + 1 = do_lvget lv0 2%N).
   omega.
 rewrite HH2 in Heq; trivial. 

 elim H18.
 rewrite H20 in H16.
 replace (@pair PackageName ShortClassName 1%positive 2%positive) with (CLASS.name java_lang_Exception.class) by reflexivity.
 injection H16; intro; clear H16; subst.
 replace java_lang_Exception.className with (CLASS.name java_lang_Exception.class) by reflexivity.
 constructor.
 unfold subclass.
 constructor 2.

 rewrite H0 in H9.
 assert (HH : do_lvget lv0 2%N - 1 <= do_lvget lv 1%N).
   omega.
 assert (Heq := arith_aux2 _ _ HH).
 apply Zle_trans with (do_hget h (Heap.DynamicField (do_lvget lv 0%N) Bill.sumFieldSignature) +
 (do_lvget lv0 2%N - 1) * (do_lvget lv0 2%N) / 2).
 trivial.
 apply Zplus_le_compat_l.
 assert (HH2 : do_lvget lv0 2%N - 1 + 1 = do_lvget lv0 2%N).
   omega.
 rewrite HH2 in Heq; trivial. 
 
 rewrite do_lvget_update_new.
 simpl.
 change_const 1.
 simpl.
 rewrite H4.
 omega.

 rewrite do_lvget_update_new.
 simpl.
 change_const 1.
 simpl.
 rewrite H4.
 omega.

 rewrite do_lvget_update_new.
 simpl.
 change_const 1.
 simpl.
 rewrite H4.
 change ((1 - 1) * 1 / 2) with 0.
 omega.

 Time Qed.
*)
