Add LoadPath "../Bicolano/ParserExperiments/Bill".
Require Export ImplemDomain.
(*Require Import TheProgram.*)
Require Import BillList. 
Require Export BigStep.
Require Export Tactic.
Require Export Misc.
Require Export AssListM.
Require Export BaseLogic.

Module Prog := ImplemProgramWithList.P.
Module D := ImplemDomain.Make(Prog).
Module BS := BigStep.MakeBigStep(D).
Module AL := AssListM.AssList.
Module R := BaseLogic.Rules(BS).
Module Bill := BillList.TheProgram.Bill. 

Open Scope type_scope.

(*ToDo: Add object invariant to both methods!
              loop invariant!*)

(*The specification of produceBill*)
  Definition pB_VarN : Prog.Var := (1%N).
  Definition rc_VarX : Prog.Var := (1%N).
  Definition pB_VarI : Prog.Var := (2%N).
  Definition VarThis : Prog.Var := (0%N).

  Parameter LEQ : D.num -> D.num -> Prop.
  Parameter plus : D.num -> D.num -> D.num.
  Parameter times : D.num -> D.num -> D.num.
  Definition One : D.num := D.I (D.Int.const 1).
  Parameter Zero : D.num.
  Parameter ZeroPointFive : D.num.

  Definition pB_Pre  := fun s0 =>
    exists n, exists loc, exists oldsum,
           D.LocalVar.get (fst s0)  pB_VarN = Some (D.Num n) /\
           D.LocalVar.get (fst s0) VarThis = Some (D.Ref loc) /\
           D.Heap.get (snd s0) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num oldsum) /\
           LEQ One n /\ LEQ Zero oldsum.

  Definition pB_Post : R.InitState -> BS.ReturnState -> Prop := fun s0 t =>
     exists loc, exists oldsum, exists sum, exists n,
              D.LocalVar.get (fst s0) pB_VarN = Some (D.Num n) /\
              D.LocalVar.get (fst s0) VarThis = Some (D.Ref loc) /\
              D.Heap.get (snd s0) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num oldsum) /\
              D.Heap.get (fst t) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num sum) /\
              LEQ Zero sum /\
              LEQ sum (plus oldsum (times n (times (plus n One) ZeroPointFive))).

  Definition pB_Inv : R.MethInv := fun s0 s => True.
  
  Definition pB_Spec : R.MethSpec := ((pB_Pre ,, pB_Post) ,, pB_Inv).

  Definition LoopPre : R.Assertion := fun s0 s =>
(*     pB_Pre s0 ->*)
     exists i, exists oldsum, exists sum, exists n, exists loc,
              D.LocalVar.get (fst s0) pB_VarN = Some (D.Num n) /\
              D.LocalVar.get (snd s) pB_VarI = Some (D.Num i) /\
              D.Heap.get (snd s0) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num oldsum) /\
              D.Heap.get (fst (fst s)) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num sum) /\
              LEQ sum (plus oldsum (times i (times (plus i One) ZeroPointFive))) /\
              LEQ i (plus n One).

  Definition LoopPost : R.LocalPost := fun s0 s t =>
  (*       pB_Pre s0 -> *)
     LoopPre s0 s -> pB_Post s0 t. 

  (*ToDo : Maybe we should in general be able to include LoopPre and/or pB_Pre 
                  here as a further assumption, like in the definition of LoopPost, 
                  but that needs changing the format of strong invariants!*)
  Definition LoopInv: R.StrongInv := fun s r => True.
         
(*Definition of a suitable context LST, an annotation table, and 
  a method specification table *)
  Definition pB_LST : R.LocalSpecTable :=
        R.AL.upd R.PCIX
                        R.LocalSpec
                        (R.AL.mty R.PCIX R.LocalSpec)
                        22%N
                        ((LoopPre,,LoopPost),,LoopInv).

  Definition pB_LAT : R.LocalAnnoTable := R.AL.mty R.PCIX R.Assertion.

  Definition rc_LST : R.LocalSpecTable := R.AL.mty R.PCIX R.LocalSpec.
  Definition rc_LAT : R.LocalAnnoTable := R.AL.mty R.PCIX R.Assertion. 
  Definition rc_Pre : R.MethPre := fun s0 => 
    exists loc, exists oldsum,
           D.LocalVar.get (fst s0) VarThis = Some (D.Ref loc) /\
           D.Heap.get (snd s0) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num oldsum) /\
           LEQ Zero oldsum.

  Definition rc_Post : R.MethPost := fun s0 t => 
     exists x, exists res, exists loc, exists sum,
           D.LocalVar.get (fst s0) VarThis = Some (D.Ref loc) /\
           D.Heap.get (fst t) (D.Heap.DynamicField loc Bill.sumFieldSignature) = Some (D.Num sum) /\
           D.LocalVar.get (fst s0) rc_VarX = Some (D.Num x) /\
           snd t = D.Normal (Some (D.Num res)) /\ LEQ res x /\
           LEQ Zero sum.

  Definition rc_Inv : R.MethInv := fun s0 s => True. 

  Definition rc_Spec : R.MethSpec := ((rc_Pre,, rc_Post),, rc_Inv).

  Definition myMST : R.MethSpecTable :=
          AL.upd R.MethIX
                       (R.MethSpec ** R.LocalSpecTable ** R.LocalAnnoTable)
                       (AL.upd R.MethIX
                                     (R.MethSpec ** R.LocalSpecTable ** R.LocalAnnoTable)
                                     (AL.mty R.MethIX (R.MethSpec ** R.LocalSpecTable ** R.LocalAnnoTable))
                                     Bill.produce_billMethod
                                     ((pB_Spec,, pB_LST) ,, pB_LAT))
                       Bill.round_costMethod
                       ((rc_Spec,, rc_LST) ,, rc_LAT).

Lemma MTY_notSome:  forall D T x X P, R.AL.lookup D T (R.AL.mty D T) x = Some X -> P.
unfold AL.mty.
unfold AL.lookup.
intros.
discriminate H.
Qed.

Lemma MTY_notSomeLAT:  forall x X P, R.AL.lookup R.PCIX R.Assertion (R.AL.mty R.PCIX R.Assertion) x = Some X -> P.
apply MTY_notSome. 
Qed.

Lemma MTY_notSomeLST:  forall x X P, R.AL.lookup R.PCIX R.LocalSpec (R.AL.mty R.PCIX R.LocalSpec) x = Some X -> P.
apply MTY_notSome. 
Qed.


Lemma MTY_notSomeLST1:  forall x X P, AL.lookup R.PCIX R.LocalSpec (AL.mty R.PCIX R.LocalSpec) x = Some X -> P.
apply MTY_notSome. 
Qed.

 Definition GP_point 
  (P   : Prog.Program) (MST : R.MethSpecTable) (m:Prog.Method)  (pre:R.MethPre) (post:R.MethPost) (MI:R.MethInv) LST LAT: Prop :=
    AL.lookup _ _ MST m = Some ((((pre,,post),,MI),,LST),,LAT) ->
    (   (forall MB, Prog.METHOD.body m = Some MB ->
          R.SP_Judgement P MST m LST LAT false 
          (Prog.BYTECODEMETHOD.firstAddress MB)
          (fun s0 s => let (S,H) := s0 in pre s0 /\ s = (H,D.OperandStack.empty,S))
          (fun s0 s t => let (S,H):= s0 in 
             s = (H,D.OperandStack.empty,S) -> post (S,H) t) 
          (R.StrongInvConversion MI))
      /\ (forall l A B I, 
            R.ASSUM_LOOKUP LST l = Some ((A,,B),,I) ->
            R.SP_Judgement P MST m LST LAT false l A B I)).

Lemma GP_point_suffices:
    forall P MST,
    (forall m pre post MI LST LAT, AL.lookup _ _ MST m  = Some ((((pre,,post),,MI),,LST),,LAT) -> GP_point P MST m pre post MI LST LAT)
     -> R.GP P MST.
    unfold R.GP.  unfold GP_point. simpl. auto. 
Qed.

Lemma rc_verified: GP_point TheProgram.program myMST Bill.round_costMethod rc_Pre rc_Post rc_Inv rc_LST rc_LAT.
unfold GP_point. intros.  split.
(*Method*)
  intros. inversion_mine H0. intros. 
Admitted.

Ltac simpl_InstrAt := 
  repeat match goal with
  | [H : R.V5.Dom.Prog.instructionAt Bill.produce_billMethod ?l = Some ?ins |- _ ] => 
                 unfold R.V5.Dom.Prog.instructionAt in H; simpl in H; 
                 unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt in H; simpl in H;
                 unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt in H; simpl in H;
                 unfold D.Prog.BYTECODEMETHOD.instructionAt in H; simpl in H;
                 unfold Prog.BYTECODEMETHOD.instructionAt in H; simpl in H; inversion_mine H
  end.

Lemma Arith1: forall x,
   LEQ Zero x ->                                
  LEQ x (plus x (times One (times (plus One One) ZeroPointFive))).
Admitted.
                         
Lemma Arith2: forall n,
   LEQ One n -> LEQ One (plus n One).      
Admitted.

Definition CHEAT1 : R.V5.Dom.num := R.V5.Dom.I (R.V5.Dom.Int.const 1).

Lemma AuxVstore: forall t1 x1 t2 x2,
  Some (ImplemProgramWithList.P.Vstore t1 x1) =
  Some (R.V5.Dom.Prog.Vstore t2 x2) -> 
  t1=t2 /\ x1=x2.
intros. simpl.
Admitted.

Lemma AuxConst: forall t1 x1 t2 x2,
  Some (ImplemProgramWithList.P.Const t1 x1) =
  Some (R.V5.Dom.Prog.Const t2 x2) -> 
  t1=t2 /\ x1=x2.
intros. simpl.
Admitted.

Lemma AuxIinc: forall t1 x1 t2 x2,
  Some (ImplemProgramWithList.P.Iinc t1 x1) =
  Some (R.V5.Dom.Prog.Iinc t2 x2) -> 
  t1=t2 /\ x1=x2.
intros. simpl.
Admitted.

Lemma AuxVload: forall t1 x1 t2 x2,
  Some (ImplemProgramWithList.P.Vload t1 x1) =
  Some (R.V5.Dom.Prog.Vload t2 x2) -> 
  t1=t2 /\ x1=x2.
intros. simpl.
assert (ImplemProgramWithList.P.Vload=R.V5.Dom.Prog.Vload).
reflexivity.
Admitted.

Lemma AuxIf_icmp: forall t1 x1 t2 x2,
  Some (ImplemProgramWithList.P.If_icmp t1 x1) =
  Some (R.V5.Dom.Prog.If_icmp t2 x2) -> 
  t1=t2 /\ x1=x2.
intros. simpl.
Admitted.

Lemma AuxGetfield: forall x1 x2,
  Some (ImplemProgramWithList.P.Getfield x1) =
  Some (R.V5.Dom.Prog.Getfield x2) -> 
  x1=x2.
intros. simpl.
Admitted.

Lemma AuxPutfield: forall x1 x2,
  Some (ImplemProgramWithList.P.Putfield x1) =
  Some (R.V5.Dom.Prog.Putfield x2) -> 
  x1=x2.
intros. simpl.
Admitted.

Lemma AuxIbinop: forall x1 x2,
  Some (ImplemProgramWithList.P.Ibinop x1) =
  Some (R.V5.Dom.Prog.Ibinop x2) -> 
  x1=x2.
intros. simpl.
Admitted.

Lemma pB_verified: GP_point TheProgram.program myMST Bill.produce_billMethod pB_Pre pB_Post pB_Inv pB_LST pB_LAT.
unfold GP_point. intros. simpl. split.
(*The condition on MBody*)
    intros; inversion_mine H0. 
    unfold Bill.produce_billMethod in H. 
    unfold Prog.BYTECODEMETHOD.firstAddress.
    simpl. 
   inversion_mine H. 
    eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
     (*First goal: is_basic instr*)
           unfold R.isBasicInstr. simpl. auto. 
     (*Second goal*)
          intros.  unfold pB_LAT in H. 
          unfold R.Anno_LOOKUP in H. 
          eapply MTY_notSomeLAT.   apply H. 
     (*Third goal*)
     intros. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
     (*Fourth goal: verification of the successor instruction*)
     intros. 
        unfold R.V5.Dom.Prog.next in H. simpl in H. 
        unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
        unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
        unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
        unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
        injection H; intros; subst.
        eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
        eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
        (*First goal: is_basic instr*)
             unfold R.isBasicInstr. simpl. auto. 
        (*Second goal*)
             intros.  unfold pB_LAT in H0. 
             unfold R.Anno_LOOKUP in H0. 
             eapply MTY_notSomeLAT.   apply H0. 
        (*Third goal*)
        intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
        (*Fourth goal: verification of the successor instruction*)
          clear H; intros. 
          unfold R.V5.Dom.Prog.next in H. simpl in H. 
          unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
          unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
          unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
          unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
          injection H; intros; subst.
          clear H. 
          eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
          eapply (R.SP_GOTO TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
          (*First goal: is goto instr*)
             unfold R.V5.Dom.Prog.instructionAt; simpl. 
             unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
         (*Second goal*)
             intros.  unfold pB_LAT in H. 
             unfold R.Anno_LOOKUP in H. 
             eapply MTY_notSomeLAT.   apply H. 
         (*Third goal*)
            intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
         (*Fourth goal: verification of the successor instruction*)
              simpl. 
             unfold R.V5.Dom.Prog.OFFSET.jump; simpl. 
             unfold BS.Dom.Prog.OFFSET.jump; simpl. 
             unfold D.Prog.OFFSET.jump; simpl. 
             unfold Prog.OFFSET.jump; simpl.              
             eapply (R.SP_CONSEQ TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
             (*First goal*)
                eapply (R.SP_AX TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                  (*First goal*) simpl. auto.  
                  (*Second goal*)
                    intros.  unfold pB_LAT in H. 
                    unfold R.Anno_LOOKUP in H. 
                    eapply MTY_notSomeLAT.   apply H. 
                 (*Third goal *) unfold LoopInv. auto. 
            (*Second goal:LoopPre*)
                 intros.
                 inversion_mine H. 
                 inversion_mine H0. 
                 inversion_mine H. 
                 destruct s0.  unfold pB_Pre in H0.  destruct H0. destruct H. subst s3. destruct H. 
                 inversion_mine H3. simpl. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 inversion_mine H2. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 inversion_mine H1. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 simpl in H. destruct H as (oldsum,H). 
                   destruct H as (VarN_is_x, H). destruct t. simpl in VarN_is_x. 
                   destruct H as (Loc_is_x0, H). simpl in Loc_is_x0. subst. 
                   destruct H as (Oldsum_isoldsum, H). 
                   destruct H as (N_pre, sum_pre). 
                 unfold LoopPre. 
                 exists One. 
                 exists oldsum. exists oldsum. (*in the first step, no assignment has happened yet*)
                 exists x; exists x0. split.
                 simpl. exact VarN_is_x. split. 
                 assert (H2_2 := AuxVstore _ _ _ _ H2).  destruct H2_2 as (H2_3,H2_4). simpl_eq. subst. 
                             assert (H0_2 := AuxConst _ _ _ _ H0).  destruct H0_2 as (H0_3,H0_4). simpl_eq. subst. 
                 unfold One. 
                    cut (D.LocalVar.get (R.V5.Dom.LocalVar.update (Some (D.Ref x0),t) 2%N
                                                      (R.V5.Dom.Num (R.V5.Dom.I (R.V5.Dom.Int.const 1)))) pB_VarI =
                            Some(D.Num(R.V5.Dom.I (R.V5.Dom.Int.const 1)))).  intros. auto. 
                 apply (R.V5.Dom.LocalVar.get_update_new). 
                 split. simpl. exact Oldsum_isoldsum. 
                 split. simpl. exact Oldsum_isoldsum. 
                 split.  apply Arith1. exact sum_pre. 
                           apply Arith2. exact N_pre.
               (*Third goal: LoopPost*)
                 intros.
                 inversion_mine H. 
                 inversion_mine H1. 
                 inversion_mine H. 
                 destruct s0.  unfold pB_Pre in H1.  destruct H1. 
                 destruct H as (n,H). 
                 destruct H as (loc,H). 
                 destruct H as (oldsum,H). 
                 destruct H as (N_is_n,H). 
                 destruct H as (This_is_loc,H). 
                 destruct H as (Oldsum_isoldsum,H). 
                 destruct H as (N_pre,sum_pre). 
                 inversion_mine H4. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 inversion_mine H5. 
                 inversion_mine H3. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 inversion_mine H2. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 destruct H0. 
                   unfold LoopPre.
                   exists One. 
                   exists oldsum. exists oldsum. (*in the first step, no assignment has happened yet*)
                   exists n; exists loc. split.
                   exact N_is_n. split. 
                   assert (H2_2 := AuxVstore _ _ _ _ H1).  destruct H2_2 as (H2_3,H2_4). simpl_eq. subst. 
                               assert (H0_2 := AuxConst _ _ _ _ H2).  destruct H0_2 as (H0_3,H0_4). simpl_eq. subst. 
                   unfold One. 
                      cut (D.LocalVar.get (R.V5.Dom.LocalVar.update t0 2%N
                                                        (R.V5.Dom.Num (R.V5.Dom.I (R.V5.Dom.Int.const 1)))) pB_VarI =
                              Some(D.Num(R.V5.Dom.I (R.V5.Dom.Int.const 1)))).  intros. auto. 
                   apply (R.V5.Dom.LocalVar.get_update_new). 
                   split. simpl. exact Oldsum_isoldsum. 
                   split. simpl. exact Oldsum_isoldsum. 
                   split.  apply Arith1. exact sum_pre. 
                             apply Arith2. exact N_pre.
                unfold R.SP_post. intros. 
                 destruct H0. 
                 destruct H0 as (sum,H0).
                 destruct H0 as (n1,H0). subst. 
                 destruct H0 as (N_is_n1,H0). simpl_eq. 
                 destruct H0 as (This_is_x0,H0). simpl_eq. 
                 destruct H0 as (Oldsum_isoldsum1,H0). simpl_eq. 
                 destruct H0 as (Sum_issum,H0). 
                 destruct H0 as (sum_pre1,sum_post). 
                 inversion_mine H5. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 inversion_mine H4.
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 inversion_mine H3.
                 simpl_InstrAt.
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 unfold pB_Post. 
                 exists x0. exists oldsum. exists sum. exists n1. split.
                 exact N_is_n. split. 
                 exact This_is_loc. split. 
                exact Oldsum_isoldsum1. split. 
                 exact Sum_issum. split. 
                 exact sum_pre1. exact sum_post. 
           (*Fourth goal: LoopInv*)
                 intros. unfold R.SP_inv. intros. unfold R.StrongInvConversion. intros; subst. 
                 unfold pB_Inv. auto. 
     (*postpone all Exception side conditions*)
 Focus 5.  intros. 
  cut (l = 22%N /\ Some ((LoopPre,,LoopPost),,LoopInv) = Some((A,,B),,I)).
  intros.  destruct H1.  inversion H2. subst.
             rename H into BillSpec. clear H2. clear H0. 
             eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
             (*First goal: is_basic instr*)
                unfold R.isBasicInstr. simpl. auto. 
             (*Second goal*)
                 intros.  unfold pB_LAT in H. 
                 unfold R.Anno_LOOKUP in H. 
                 eapply MTY_notSomeLAT.   apply H. 
             (*Third goal*)
                 intros. unfold LoopInv. auto.
            (*Fourth goal: verification of the successor instruction*)
               intros; simpl. 
               unfold R.V5.Dom.Prog.next in H.  simpl in H. 
               unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
               unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
               unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
               unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
               injection H. intros; subst. 
               clear H. 
               eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
               eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
              (*First goal: is_basic instr*)
                unfold R.isBasicInstr. simpl. auto. 
              (*Second goal*)
                intros.  unfold pB_LAT in H. 
                unfold R.Anno_LOOKUP in H. 
                eapply MTY_notSomeLAT.   apply H. 
             (*Third goal*)
               intros. unfold R.SP_inv. unfold LoopInv. auto.
             (*Fourth goal: verification of the successor instruction*)
                intros; simpl.  
                unfold R.V5.Dom.Prog.next in H. simpl in H. 
                unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                injection H; intros; subst. 
                clear H. 
                eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                eapply (R.SP_IF TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                (*First goal: is_if instr*)
                   unfold R.isIfInstr. simpl. auto. 
                (*Second goal*)
                   unfold R.V5.Dom.Prog.next. simpl. 
                   unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold D.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold Prog.BYTECODEMETHOD.nextAddress. simpl. auto. 
                (*Third goal*)
                    intros.  unfold pB_LAT in H.   
                    unfold R.Anno_LOOKUP in H. 
                    eapply MTY_notSomeLAT.   apply H. 
                (*Fourth goal*)
                    intros. unfold R.SP_inv. unfold LoopInv. auto. 
                (*Fifth goal: jump case*)
                    simpl. 
                   unfold R.V5.Dom.Prog.OFFSET.jump; simpl. 
                   unfold BS.Dom.Prog.OFFSET.jump; simpl. 
                   unfold D.Prog.OFFSET.jump; simpl. 
                   unfold Prog.OFFSET.jump; simpl. 
                   eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                   eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                   (*First goal: is_basic instr*)
                      unfold R.isBasicInstr. simpl. auto. 
                   (*Second goal*)
                      intros.  unfold pB_LAT in H. 
                      unfold R.Anno_LOOKUP in H. 
                      eapply MTY_notSomeLAT.   apply H. 
                   (*Third goal*)
                      intros. unfold R.SP_inv. unfold LoopInv. auto. 
                   (*Fourth goal: verification of the successor instruction*)
                     intros. 
                     unfold R.V5.Dom.Prog.next in H. simpl in H. 
                     unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                     unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                     unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                     unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                     injection H; intros; subst. 
                     clear H. 
                     eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                     eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                     (*First goal: is_basic instr*)
                         unfold R.isBasicInstr. simpl. auto. 
                     (*Second goal*)
                         intros.  unfold pB_LAT in H. 
                         unfold R.Anno_LOOKUP in H. 
                         eapply MTY_notSomeLAT.   apply H. 
                     (*Third goal*)
                        intros. unfold R.SP_inv. unfold LoopInv. auto. 
                     (*Fourth goal: verification of the successor instruction*)
                        intros. 
                        unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                        unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                        unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                        unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                        unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                        injection H. intros; subst. 
                        clear H. 
                        eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                        eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                       (*First goal: is_basic instr*)
                           unfold R.isBasicInstr. simpl. auto. 
                       (*Second goal*)
                           intros.  unfold pB_LAT in H. 
                           unfold R.Anno_LOOKUP in H. 
                           eapply MTY_notSomeLAT.   apply H. 
                       (*Third goal*)
                          intros. unfold R.SP_inv. unfold LoopInv. auto. 
                       (*Fourth goal: verification of the successor instruction*)
                          intros. 
                         unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                         unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                         unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                         unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                         unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                         injection H. intros; subst. 
                         clear H. 
                         eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                         eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                        (*First goal: is_basic instr*)
                            unfold R.isBasicInstr. simpl. auto. 
                        (*Second goal*)
                            intros.  unfold pB_LAT in H. 
                            unfold R.Anno_LOOKUP in H. 
                            eapply MTY_notSomeLAT.   apply H. 
                        (*Third goal*)
                            intros. unfold R.SP_inv. unfold LoopInv. auto. 
                        (*Fourth goal: verification of the successor instruction*)
                            intros. 
                            unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                            unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                            unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                            unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                            unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                            injection H. intros; subst. 
                            clear H. 
                            eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                            eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                            (*First goal: is_basic instr*)
                                unfold R.isBasicInstr. simpl. auto. 
                            (*Second goal*)
                                intros.  unfold pB_LAT in H. 
                                unfold R.Anno_LOOKUP in H. 
                                eapply MTY_notSomeLAT.   apply H. 
                            (*Third goal*)
                                intros. unfold R.SP_inv. unfold LoopInv. auto. 
                            (*Fourth goal: verification of the successor instruction (1/22)*)
                                intros. 
                                unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                                unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                injection H. intros; subst. 
                                clear H. 
                                eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                eapply (R.SP_INVV TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                                (*First goal: is invvokevirtual instruction (1/33)*)
                                    unfold R.V5.Dom.Prog.instructionAt; simpl. 
                                    unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
                               (*Second goal: findMethod*)
                                    unfold R.V5.Dom.Prog.findMethod; simpl. 
                                    unfold R.V5.Dom.Prog.CLASS.method; simpl. 
                                    unfold BS.Dom.Prog.CLASS.method; simpl. 
                                    unfold D.Prog.CLASS.method; simpl. 
                                    unfold Prog.CLASS.method; simpl. auto. 
                               (*Third goal: isNative*)
                                    unfold R.V5.Dom.Prog.METHOD.isNative; simpl. auto.                                    
                               (*Fourth goal: MSPEC_lookup*)
                                     unfold R.MSPEC_LOOKUP. unfold myMST. 
                                    unfold R.AL.lookup.  unfold rc_Spec. simpl. auto. 
                               (*Fifith goal*)
                                    intros.  unfold pB_LAT in H. 
                                    unfold R.Anno_LOOKUP in H. 
                                    eapply MTY_notSomeLAT.   apply H. 
                               (*Sixth goal*) 
                                    intros. unfold R.SP_inv. unfold LoopInv. auto. 
                               (*Seventh goal*)
                                    auto. 
                               (*Eighth goal*)
                                    intros. unfold R.SP_inv. unfold LoopInv. auto. 
                               (*Nineth goal*)
                                    intros; simpl.          
                                    unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                                    unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                    unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                    unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                    unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                    injection H. intros; subst. 
                                    clear H. 
                                    eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                    eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).                            
                                   (*First goal: is_basic instr*)
                                      unfold R.isBasicInstr. simpl. auto. 
                                   (*Second goal*)
                                      intros.  unfold pB_LAT in H. 
                                      unfold R.Anno_LOOKUP in H. 
                                      eapply MTY_notSomeLAT.   apply H. 
                                  (*Third goal*)
                                     intros. unfold R.SP_inv. unfold LoopInv. simpl. 
                                     constructor. intros. auto. 
                                  (*Fourth goal: verification of the successor instruction*)
                                     intros. 
                                     unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                                     unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                     unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                     unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                     unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                     injection H. intros; subst. 
                                     clear H. 
                                     eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                     eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                                    (*First goal: is_basic instr*)
                                        unfold R.isBasicInstr. simpl. auto. 
                                    (*Second goal*)
                                         intros.  unfold pB_LAT in H. 
                                         unfold R.Anno_LOOKUP in H. 
                                         eapply MTY_notSomeLAT.   apply H. 
                                    (*Third goal*)
                                         intros. unfold R.SP_inv. unfold LoopInv. simpl. 
                                          constructor. intros. auto. 
                                   (*Fourth goal: verification of the successor instruction*)
                                        intros. 
                                        unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                                        unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                        unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                        unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                        unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                        injection H. intros; subst. 
                                        clear H. 
                                        eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                        eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                       (*First goal: is_basic instr*)
                                          unfold R.isBasicInstr. simpl. auto. 
                                       (*Second goal*)
                                           intros.  unfold pB_LAT in H. 
                                           unfold R.Anno_LOOKUP in H. 
                                           eapply MTY_notSomeLAT.   apply H. 
                                       (*Third goal*)
                                           intros. unfold R.SP_inv. unfold LoopInv. simpl. 
                                            constructor. intros. auto. 
                                       (*Fourth goal: verification of the successor instruction*)
                                           intros. 
                                           unfold R.V5.Dom.Prog.next in H.  simpl in H. 
                                           unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                           unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                           unfold D.Prog.BYTECODEMETHOD.nextAddress in H. simpl in H. 
                                           unfold Prog.BYTECODEMETHOD.nextAddress in H. simpl in H.
                                           injection H. intros; subst. 
                                           clear H. 
                                           eapply (R.SP_CONSEQ TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).   
             (*First goal*)
                eapply (R.SP_AX TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                  (*First goal*) simpl. auto.  
                  (*Second goal*)
                    intros.  unfold pB_LAT in H. 
                    unfold R.Anno_LOOKUP in H. 
                    eapply MTY_notSomeLAT.   apply H. 
                 (*Third goal *) unfold LoopInv. auto. 
            (*Second goal:LoopPre*)
                 intros.
                 inversion_mine H. 
                 inversion_mine H0. 
                 inversion_mine H. 
                 inversion_mine H0. 
                 inversion_mine H6. 
                 inversion_mine H0. 
                 inversion_mine H6. 
                 inversion_mine H0. 
                 inversion_mine H6. 
                 inversion_mine H0. 
                 inversion_mine H6. 
                 inversion_mine H0. 
                 inversion_mine H14. simpl. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt. 
                   inversion_mine H13. simpl. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                    simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                    simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.                    
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                     inversion_mine H12. simpl. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                       inversion_mine H11. simpl. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                        inversion_mine H10. simpl. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                          inversion_mine H9. simpl. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                            inversion_mine H8. simpl. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                             simpl_InstrAt. 
                             simpl_InstrAt.
                              inversion_mine H7. simpl. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt. 
                                inversion_mine H3. simpl. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                  inversion_mine H2. simpl. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt.
                                   simpl_InstrAt.
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                    inversion_mine H1. simpl. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt.
                                     simpl_InstrAt.
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                       assert (H0_1 := AuxVload _ _ _ _ H0).  destruct H0_1 as (H0_2,H0_3). simpl_eq. subst. clear H0. 
                                       assert (H13_1 := AuxVload _ _ _ _ H13).  destruct H13_1 as (H13_2,H13_3). simpl_eq. subst. clear H13. 
                                       assert (H11_1 := AuxVload _ _ _ _ H11).  destruct H11_1 as (H11_2,H11_3). simpl_eq. subst. clear H11. 
                                       assert (H12_1 := AuxIf_icmp _ _ _ _ H12).  destruct H12_1 as (H12_2,H12_3). simpl_eq. subst. clear H12. 
                                       assert (H1_1 := AuxIinc _ _ _ _ H1).  destruct H1_1 as (H1_2,H1_3). simpl_eq. subst. clear H1. 
                                       assert (H2_2 := AuxPutfield _ _ H2).  destruct H2_2 as (H2_3,H2_4). simpl_eq. subst. clear H2. 
                                       assert (H9_2 := AuxGetfield _ _ H9).  destruct H9_2 as (H9_3,H9_4). simpl_eq. subst. clear H9. 
                                       assert (H3_2 := AuxIbinop _ _ H3).  destruct H3_2 as (H3_3,H3_4). simpl_eq. subst. clear H3.                                        
                                       unfold LoopPre in H6. 
                                       destruct H6 as (Pre_i, H6).  destruct H6 as (Pre_oldsum, H6).  destruct H6 as (Pre_sum, H6).  destruct H6 as (Pre_n, H6).  destruct H6 as (Pre_loc, H6).  
                                       destruct H6 as (Pre_VarN, H6). destruct H6 as (Pre_VarI, H6). destruct H6 as (Pre_GetOldSum, H6). destruct H6 as (Pre_GetSum, H6). 
                                       destruct H6 as (Pre_LEQ_sum, PreLEQ_i). 
                                       unfold LoopPre. simpl. 
                                         exists (plus (D.Num i) One). 
Qed: 150
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.
                                     simpl_InstrAt. 
                                     simpl_InstrAt.


Qed. 130                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt. 
                                   simpl_InstrAt.
                                   simpl_InstrAt.
                                   simpl_InstrAt.
Qed: 122
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                                 simpl_InstrAt.
                                 simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt.
                               simpl_InstrAt. 
                               simpl_InstrAt. 

                 assert (H2_2 := AuxVstore _ _ _ _ H2).  destruct H2_2 as (H2_3,H2_4). simpl_eq. subst. 
                             assert (H0_2 := AuxConst _ _ _ _ H0).  destruct H0_2 as (H0_3,H0_4). simpl_eq. subst. 
                                
Qed. 93                              
Qed: 92
Qed. 91
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                           simpl_InstrAt.
                           simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 
                         simpl_InstrAt.
                         simpl_InstrAt. 

Qed. 57

Qed. 56
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
                       simpl_InstrAt. 
                       simpl_InstrAt.
Qed: 55
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 
                     simpl_InstrAt.
                     simpl_InstrAt. 

Qed: 29
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt.
                   simpl_InstrAt. 
                   simpl_InstrAt. 
                   
Qed.
                28: simpl_InstrAt.
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 inversion_mine H2. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 inversion_mine H1. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 Focus 2. simpl_InstrAt. 
                 Focus 2. simpl_InstrAt.
                 simpl in H. destruct H as (oldsum,H). 
                   destruct H as (VarN_is_x, H). destruct t. simpl in VarN_is_x. 
                   destruct H as (Loc_is_x0, H). simpl in Loc_is_x0. subst. 
                   destruct H as (Oldsum_isoldsum, H). 
                   destruct H as (N_pre, sum_pre). 
                 unfold LoopPre. 
                 exists One. 
                 exists oldsum. exists oldsum. (*in the first step, no assignment has happened yet*)
                 exists x; exists x0. split.
                 simpl. exact VarN_is_x. split. 
                 assert (H2_2 := AuxVstore _ _ _ _ H2).  destruct H2_2 as (H2_3,H2_4). simpl_eq. subst. 
                             assert (H0_2 := AuxConst _ _ _ _ H0).  destruct H0_2 as (H0_3,H0_4). simpl_eq. subst. 
                 unfold One. 
                    cut (D.LocalVar.get (R.V5.Dom.LocalVar.update (Some (D.Ref x0),t) 2%N
                                                      (R.V5.Dom.Num (R.V5.Dom.I (R.V5.Dom.Int.const 1)))) pB_VarI =
                            Some(D.Num(R.V5.Dom.I (R.V5.Dom.Int.const 1)))).  intros. auto. 
                 apply (R.V5.Dom.LocalVar.get_update_new). 
                 split. simpl. exact Oldsum_isoldsum. 
                 split. simpl. exact Oldsum_isoldsum. 
                 split.  apply Arith1. exact sum_pre. 
                           apply Arith2. exact N_pre.
               (*Third goal: LoopPost*)
                 intros.
                 inversion_mine H. 
                 inversion_mine H1. 
                 inversion_mine H. 
                 destruct s0.  unfold pB_Pre in H1.  destruct H1. 
                 destruct H as (n,H). 
                 destruct H as (loc,H). 
                 destruct H as (oldsum,H). 
                 destruct H as (N_is_n,H). 
                 destruct H as (This_is_loc,H). 
                 destruct H as (Oldsum_isoldsum,H). 
                 destruct H as (N_pre,sum_pre). 
                 inversion_mine H4. 
                 simpl_InstrAt.
                 simpl_InstrAt. 
                 simpl_InstrAt.
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                 Focus 2.  simpl_InstrAt. 
                                            

                   
Qed.
                (*Sixth goal: fallthrough case (2/12)*)
             unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
        (*Second goal*)

     intros.  

Qed.
   Focus 2. intros. inversion_mine H1. 
   inversion_mine H8. clear H2. 
   inversion_mine H. 
(*        inversion_mine H8. *)
(*          unfold Bill.produce_billBody. unfold Bill.produce_billInstructions. simpl. *)
          unfold R.V5.Dom.Prog.BYTECODEMETHOD.firstAddress.
          unfold BS.Dom.Prog.BYTECODEMETHOD.firstAddress.
          unfold D.Prog.BYTECODEMETHOD.firstAddress. 
          unfold Bill.produce_billBody. 
          unfold Bill.produce_billInstructions. 
          unfold Prog.BYTECODEMETHOD.firstAddress. 
          unfold Prog.BYTECODEMETHOD.instr.
          unfold Bill.produce_billMethod. 
          unfold Bill.produce_billSignature. 
          unfold R.isBasicInstr. simpl. 
          inversion_mine MB. 
simpl_eq. 
          unfold Bill.produce_billInstructions. auto.     
          unfold ImplemProgramWithList.P.BYTECODEMETHOD.Build_t.   
          unfold ImplemProgramWithList.P.BYTECODEMETHOD.Build_t.   
Qed.

Lemma Bill_verified: R.GP TheProgram.program myMST.
apply GP_point
(*The proof of the GP property*)
  Lemma L1:  R.GP TheProgram.program myMST.
  unfold R.GP.
  intros.
  (*Since MST has only one entry, the following cut should be trivial to
     prove - but how?*)
  case(m = Bill.produce_billMethod /\ LAT = pB_LAT /\ LST = pB_LST 
          /\ MI = pB_Inv /\ pre = pB_Pre /\ post = pB_Post); intros; simpl_eq. 
  destruct H0; simpl_eq.  destruct H1; simpl_eq.  destruct H2; simpl_eq.  destruct H3; simpl_eq.  destruct H4; simpl_eq. 
  split.
  (*The condition on MBody*)
    intros. inversion_mine H6. 
    unfold Bill.produce_billMethod in H8. 
   inversion_mine H8. 
    unfold myMST in H.
(*    cut (MB = Bill.produce_billBody). intros; simpl_eq. *)
(*    eapply (R.SP_CONSEQ TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). *)
    eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
     (*First goal: is_basic instr*)
          unfold R.isBasicInstr. simpl. auto. 
     (*Second goal*)
          intros.  unfold pB_LAT in H0. 
          unfold R.Anno_LOOKUP in H0. 
          eapply MTY_notSomeLAT.   apply H0. 
     (*Third goal*)
     intros. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
     (*Fourth goal: verification of the successor instruction*)
     intros. 
        unfold R.V5.Dom.Prog.next in H0. simpl in H0. 
        unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
        unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
        unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
        unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
        injection H0; intros; subst.
        eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
        eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
        (*First goal: is_basic instr*)
             unfold R.isBasicInstr. simpl. auto. 
        (*Second goal*)
             intros.  unfold pB_LAT in H1. 
             unfold R.Anno_LOOKUP in H1. 
             eapply MTY_notSomeLAT.   apply H1. 
        (*Third goal*)
        intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
        (*Fourth goal: verification of the successor instruction*)
          clear H0; intros. 
          unfold R.V5.Dom.Prog.next in H0. simpl in H0. 
          unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
          unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
          unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
          unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H.
          injection H0; intros; subst.
          eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
          eapply (R.SP_GOTO TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
          (*First goal: is goto instr*)
             unfold R.V5.Dom.Prog.instructionAt; simpl. 
             unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
         (*Second goal*)
             intros.  unfold pB_LAT in H1. 
             unfold R.Anno_LOOKUP in H1. 
             eapply MTY_notSomeLAT.   apply H1. 
         (*Third goal*)
            intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
         (*Fourth goal: verification of the successor instruction*)
              simpl. 
             unfold R.V5.Dom.Prog.OFFSET.jump; simpl. 
             unfold BS.Dom.Prog.OFFSET.jump; simpl. 
             unfold D.Prog.OFFSET.jump; simpl. 
             unfold Prog.OFFSET.jump; simpl. 
             eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
             eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
             (*First goal: is_basic instr*)
                unfold R.isBasicInstr. simpl. auto. 
             (*Second goal*)
                 intros.  unfold pB_LAT in H1. 
                 unfold R.Anno_LOOKUP in H1. 
                 eapply MTY_notSomeLAT.   apply H1. 
             (*Third goal*)
                 intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
            (*Fourth goal: verification of the successor instruction*)
               clear H0; intros. 
               unfold R.V5.Dom.Prog.next in H0.  simpl in H0. 
               unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
               unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
               unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
               unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
               injection H0. intros; subst. 
               eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
               eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
              (*First goal: is_basic instr*)
                unfold R.isBasicInstr. simpl. auto. 
              (*Second goal*)
                intros.  unfold pB_LAT in H1. 
                unfold R.Anno_LOOKUP in H1. 
                eapply MTY_notSomeLAT.   apply H1. 
             (*Third goal*)
               intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
             (*Fourth goal: verification of the successor instruction*)
                clear H0; intros. 
                unfold R.V5.Dom.Prog.next in H0. simpl in H0. 
                unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                injection H0; intros; subst. 
                clear H0. 
                eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                eapply (R.SP_IF TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                (*First goal: is_if instr*)
                   unfold R.isIfInstr. simpl. auto. 
                (*Second goal*)
                   unfold R.V5.Dom.Prog.next. simpl. 
                   unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold D.Prog.BYTECODEMETHOD.nextAddress. simpl. 
                   unfold Prog.BYTECODEMETHOD.nextAddress. simpl. auto. 
                (*Third goal*)
                    intros.  unfold pB_LAT in H0.   
                    unfold R.Anno_LOOKUP in H0. 
                    eapply MTY_notSomeLAT.   apply H0. 
                (*Fourth goal*)
                    intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                (*Fifth goal: jump case*)
                    simpl. 
                   unfold R.V5.Dom.Prog.OFFSET.jump; simpl. 
                   unfold BS.Dom.Prog.OFFSET.jump; simpl. 
                   unfold D.Prog.OFFSET.jump; simpl. 
                   unfold Prog.OFFSET.jump; simpl. 
                   eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                   eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                   (*First goal: is_basic instr*)
                      unfold R.isBasicInstr. simpl. auto. 
                   (*Second goal*)
                      intros.  unfold pB_LAT in H0. 
                      unfold R.Anno_LOOKUP in H0. 
                      eapply MTY_notSomeLAT.   apply H0. 
                   (*Third goal*)
                      intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                   (*Fourth goal: verification of the successor instruction*)
                     intros. 
                     unfold R.V5.Dom.Prog.next in H0. simpl in H0. 
                     unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                     unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                     unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                     unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                     injection H0; intros; subst. 
                     clear H0. 
                     eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                     eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                     (*First goal: is_basic instr*)
                         unfold R.isBasicInstr. simpl. auto. 
                     (*Second goal*)
                         intros.  unfold pB_LAT in H0. 
                         unfold R.Anno_LOOKUP in H0. 
                         eapply MTY_notSomeLAT.   apply H0. 
                     (*Third goal*)
                        intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                     (*Fourth goal: verification of the successor instruction*)
                        intros. 
                        unfold R.V5.Dom.Prog.next in H0.  simpl in H0. 
                        unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                        unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                        unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                        unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                        injection H0. intros; subst. 
                        eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                        eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                       (*First goal: is_basic instr*)
                           unfold R.isBasicInstr. simpl. auto. 
                       (*Second goal*)
                           intros.  unfold pB_LAT in H1. 
                           unfold R.Anno_LOOKUP in H1. 
                           eapply MTY_notSomeLAT.   apply H1. 
                       (*Third goal*)
                          intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                       (*Fourth goal: verification of the successor instruction*)
                          clear H0; intros. 
                         unfold R.V5.Dom.Prog.next in H0.  simpl in H0. 
                         unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                         unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                         unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                         unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                         injection H0. intros; subst. 
                         eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                         eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                        (*First goal: is_basic instr*)
                            unfold R.isBasicInstr. simpl. auto. 
                        (*Second goal*)
                            intros.  unfold pB_LAT in H1. 
                            unfold R.Anno_LOOKUP in H1. 
                            eapply MTY_notSomeLAT.   apply H1. 
                        (*Third goal*)
                            intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                        (*Fourth goal: verification of the successor instruction*)
                            clear H0; intros. 
                            unfold R.V5.Dom.Prog.next in H0.  simpl in H0. 
                            unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                            unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                            unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                            unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                            injection H0. intros; subst. 
                            eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                            eapply (R.SP_INSTR TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                            (*First goal: is_basic instr*)
                                unfold R.isBasicInstr. simpl. auto. 
                            (*Second goal*)
                                intros.  unfold pB_LAT in H1. 
                                unfold R.Anno_LOOKUP in H1. 
                                eapply MTY_notSomeLAT.   apply H1. 
                            (*Third goal*)
                                intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                            (*Fourth goal: verification of the successor instruction (1/22)*)
                                clear H0; intros. 
                                unfold R.V5.Dom.Prog.next in H0.  simpl in H0. 
                                unfold R.V5.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                                unfold BS.Dom.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                                unfold D.Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0. 
                                unfold Prog.BYTECODEMETHOD.nextAddress in H0. simpl in H0.
                                injection H0. intros; subst. 
                                clear H0. 
                                eapply (R.SP_INJECT TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT).  
                                eapply (R.SP_INVV TheProgram.program myMST Bill.produce_billMethod pB_LST pB_LAT). 
                                (*First goal: is invvokevirtual instruction (1/33)*)
                                    unfold R.V5.Dom.Prog.instructionAt; simpl. 
                                    unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
                                    unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
                               (*Second goal: findMethod*)
                                    unfold R.V5.Dom.Prog.findMethod; simpl. 
                                    unfold R.V5.Dom.Prog.CLASS.method; simpl. 
                                    unfold BS.Dom.Prog.CLASS.method; simpl. 
                                    unfold D.Prog.CLASS.method; simpl. 
                                    unfold Prog.CLASS.method; simpl. auto. 
                               (*Third goal: isNative*)
                                    unfold R.V5.Dom.Prog.METHOD.isNative; simpl. auto.                                    
                               (*Fourth goal: MSPEC_lookup*)
                                     unfold R.MSPEC_LOOKUP. 
                                    unfold R.AL.lookup. 
                                    unfold AL.lookup. simpl. auto. 
Qed.
                               (*Fifith goal*)
                                    intros.  unfold pB_LAT in H1. 
                                    unfold R.Anno_LOOKUP in H1. 
                                    eapply MTY_notSomeLAT.   apply H1. 
                               (*Third goal*)
                                   intros. unfold R.SP_inv. unfold R.StrongInvConversion; unfold pB_Inv. auto. 
                                    

                   
Qed.
                (*Sixth goal: fallthrough case (2/12)*)
             unfold R.V5.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold BS.Dom.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold D.Prog.BYTECODEMETHOD.instructionAt; simpl. 
             unfold Prog.BYTECODEMETHOD.instructionAt; simpl. auto. 
        (*Second goal*)

     intros.  

Qed.
   Focus 2. intros. inversion_mine H1. 
   inversion_mine H8. clear H2. 
   inversion_mine H. 
(*        inversion_mine H8. *)
(*          unfold Bill.produce_billBody. unfold Bill.produce_billInstructions. simpl. *)
          unfold R.V5.Dom.Prog.BYTECODEMETHOD.firstAddress.
          unfold BS.Dom.Prog.BYTECODEMETHOD.firstAddress.
          unfold D.Prog.BYTECODEMETHOD.firstAddress. 
          unfold Bill.produce_billBody. 
          unfold Bill.produce_billInstructions. 
          unfold Prog.BYTECODEMETHOD.firstAddress. 
          unfold Prog.BYTECODEMETHOD.instr.
          unfold Bill.produce_billMethod. 
          unfold Bill.produce_billSignature. 
          unfold R.isBasicInstr. simpl. 
          inversion_mine MB. 
simpl_eq. 
          unfold Bill.produce_billInstructions. auto.     
          unfold ImplemProgramWithList.P.BYTECODEMETHOD.Build_t.   
          unfold ImplemProgramWithList.P.BYTECODEMETHOD.Build_t.   

qed.   (*Verification 
(*
              //@ ensures sum <= \old(sum) + n * (n + 1) / 2;
*)

(* 
 Local Variables: 
 coq-prog-name: "coqtop -I ../Bicolano -I ../Library/ -I ../Library/Map -I ../Bicolano/ParserExperiments/Bill -emacs"
 End:
*)