Add LoadPath "../Library".
Add LoadPath "../Library/Map".
Add LoadPath "../Bicolano".

Require Import ImplemProgramWithMap.
Require Import Map_java_lang_Object.
Require Import Map_java_lang_Throwable.
Require Import Map_java_lang_Exception.
Require Import ImplemWp.
Require Import BillMap.

Import Mwp R V5 Dom Prog.
Import TheProgram.

Definition bill_subclass := 
 match P.subclass_test program with
 | Some f => f
 | None => fun x y => true
 end.

Definition round_cost_pre (s0:InitState) := 
  let (l,h0) := s0 in
  exists n, LocalVar.get l 1%N = Some (Num (I n)).

Definition round_cost_post (s0:InitState) (t:ReturnState) := 
   let (l0, h0) := s0 in
   let (ht,vt) := t in
   (forall (am:Heap.AdressingMode) (v:value), 
       Heap.get h0 am = Some v -> Heap.get ht am = Some v)
   /\ 
   match vt with
   | Normal res => 
        forall x, LocalVar.get l0 1%N = Some (Num (I x)) ->  
           exists n, res = Some (Num (I n)) /\ 0 <= Int.toZ n <= Int.toZ x
   | Exception loc => Heap.typeof ht loc = Some (Heap.LocationObject java_lang_Exception.className)
   end.

Definition round_cost_spec := 
     ((0%nat,,(round_cost_pre,,round_cost_post)),,
        @PCM.empty (R.Assertion ** nat)).

Definition mk_bill_var loc n :=
       LocalVar.update 
         (LocalVar.update MDom.LocalVar.empty
                  0%N (Ref loc))
                  1%N (Num (I n)).

Coercion Int.toZ : Int.t >-> Z.
Open Local Scope Z_scope.

Definition mk_produce_bill_pre (s0:R.InitState) loc (n sum:Int.t) := 
   let (l0,h0) := s0 in
        l0 = mk_bill_var loc n
     /\ 0 < n < Int.half_base - 1
     /\ Heap.get h0 (Heap.DynamicField loc Bill.sumFieldSignature) = Some (Num (I sum))
     /\ 0 <= sum 
     /\ sum + (n*(n+1))/2 < Int.half_base.

Definition produce_bill_pre (s0:R.InitState) := 
  exists loc, exists n, exists sum, mk_produce_bill_pre s0 loc n sum.

Definition produce_bill_post (s0:R.InitState) (t:ReturnState) := 
  let (l0, h0) := s0 in
  let (ht,vt) := t in
    forall loc n sum, mk_produce_bill_pre s0 loc n sum ->
           (forall (am:Heap.AdressingMode) (v:value), 
              am <> (Heap.DynamicField loc Bill.sumFieldSignature) -> 
              Heap.get h0 am = Some v -> Heap.get ht am = Some v) 
       /\ (exists sum', 
              Heap.get ht (Heap.DynamicField loc Bill.sumFieldSignature) = Some (Num (I sum')) /\
              0 <= sum' <= sum + (n*(n+1))/2)
       /\ match vt with 
           | Normal _ => True
           | Exception _ => False
           end.

Definition produce_bill_invariant (s0:R.InitState) (s:R.LocalState) :=
   let (l0,h0) := s0 in
   let (hs,l) := s in
   let (h,s) := hs in
     exists loc, exists n, exists sum0, exists sum, exists i, 
          mk_produce_bill_pre s0 loc n sum0 
       /\ (forall (am:Heap.AdressingMode) (v:value), 
              am <> (Heap.DynamicField loc Bill.sumFieldSignature) -> 
              Heap.get h0 am = Some v -> Heap.get h am = Some v)
       /\ l = LocalVar.update (mk_bill_var loc n) 2%N (Num (I i))
       /\ 1 <= i <= n + 1
       /\ Heap.get h (Heap.DynamicField loc Bill.sumFieldSignature) = Some (Num (I sum))
       /\ 0 <= sum <= sum0 + ((i-1)*i)/2.

Definition produce_bill_spec := 
     ((4%nat,,(produce_bill_pre,,produce_bill_post)),,
        PCM.update (PCM.empty _) 22%N (produce_bill_invariant,, 14%nat)).

Definition program_spec :=
    (MM.update
    (MM.update 
          (MM.empty _)
          Bill.round_costSignature round_cost_spec) 
          Bill.produce_billSignature produce_bill_spec).

Definition anno_prog := 
  AnnoProg program bill_subclass program_spec.

Ltac myred := 
 hnf;match goal with
 | [|- _ -> _] => intro;hnf;myred
 | [|- _ /\ _] => split;hnf;myred
 | _ => idtac
 end.

Ltac simpl_hyps := 
 match goal with 
 | [H : _ /\ _ |- _ ] => 
    let h1 := fresh "H" in let h2 := fresh "H" in destruct H as (h1,h2);simpl_hyps
 | [H:SemCompInt _ _ _ |- _] => simpl in H;simpl_hyps
 | [H:METHOD.valid_var _ _ |- _ ] => clear H;simpl_hyps
 | [H:compat_ValKind_value _ _ |- _] => clear H;simpl_hyps
 | _ => idtac
  end.

Ltac solve := myred;simpl_hyps;simpl_eq;subst;auto.

Axiom Int_add_spec : forall i1 i2,
   Int.range (Int.toZ i1 + Int.toZ i2) -> 
   Int.toZ (Int.add i1 i2) = (Int.toZ i1 + Int.toZ i2)%Z.

Lemma arith_aux : forall i, 
     1 <= i ->
    (i - 1) * i / 2 <= (i - 1) * i / 2  + i <= i * (i+1) / 2.
Proof.
Admitted.
(*  intros i Hi; destruct (Zle_lt_or_eq _ _ Hi);clear Hi;subst.
  assert (Hi : 2 <= i). omega.
  destruct (Zle_lt_or_eq _ _ Hi);clear Hi H;subst. *)

Axiom arith_aux2 : forall i n, i <= n ->   i * (i+1) /2 <= n * (n+1) / 2.
   
Lemma produce_bill_correct : certifiedMethod anno_prog Bill.produce_billMethod produce_bill_spec.
Proof with solve.
 myred ...
 destruct s0 as (l0,h0);destruct s as ((h,s),l).
 destruct H as (loc, (n, (sum0, (sum, (i, (Hpre, (Hmod, (Heq, HH))))))))...
 myred.
 assert (Rpre : round_cost_pre
             (stack2localvar (Num (I i) :: Ref loc :: Num (I sum) :: Ref loc :: s)
                (length
                   (METHODSIGNATURE.parameters (snd Bill.round_costSignature)) + 1), h)). 
    myred;exists i;trivial. 
 case rv...
 assert (Rpost := H17 Rpre). clear Rpre H17... 
 simpl in Rpost ...
 assert (RP2':= H0 _ (refl_equal _));clear H0;
     destruct RP2' as (n1,(Heq1, Hbound));inversion_mine Heq1...
 clear H13 H12.
 injection H11;clear H11;intros ...
 unfold mk_produce_bill_pre in Hpre...
 assert (Int.range 1).
   unfold Int.range;compute;split;trivial;intros Heq;inversion Heq.
 assert (Int.toZ (Int.const 1) = 1).
    rewrite Int.const_prop;trivial.
 assert (Heqi : Int.toZ (Int.add i0 (Int.const 1)) = i0 + 1).
  rewrite Int_add_spec. rewrite H12. omega.
  unfold Int.range;rewrite H12. omega.
 assert (Haux := arith_aux _ H3).
 assert (Haux2 := arith_aux2 _ _ H7).
 assert (Heq2 : Int.toZ (SemBinopInt AddInt sum n1) = sum + n1).
  simpl;apply Int_add_spec. unfold Int.range;split. omega. 
  apply Zle_lt_trans with (sum0 + (i0 - 1) * i0 / 2 + i0);try omega.

 exists loc;exists n;exists sum0;exists (SemBinopInt AddInt sum n1);exists (Int.add i0 (Int.const 1));try rewrite Heqi;try rewrite Heq2;repeat split;auto;try rewrite Heqi;try omega.
 intros;rewrite Heap.get_update_old;auto.
 rewrite Heap.get_update_same;trivial.
 econstructor;eauto. 
 replace (i0 + 1 - 1) with (Int.toZ i0);omega.
 assert (Rpost := H18 Rpre). clear Rpre H18... 
 simpl in Rpost...
 myred ...
 intro XX;hnf;intros;hnf;clear XX;intros.
 simpl in Hpre, H8;simpl_hyps;subst;simpl_eq.
 injection H9;clear H9;intros;subst ...
 exists sum;repeat split;auto.
 assert (XX:= arith_aux _ H3); assert (XX2 := arith_aux2 _ _ H7);omega.
 intro XX;clear XX;simpl in H7;hnf;intros;hnf;intros.
 simpl in Hpre, H10;simpl_hyps;subst;simpl_eq.
 injection H11;clear H11;intros;subst ...
 exists sum;repeat split;auto.
 assert (n0 + 1 =  i). omega. 
 rewrite <- H in H4.
 pattern n0 at -1;replace (Int.toZ n0) with  (n0 + 1 - 1);[trivial|ring].

 (* precondiction implies wp of first address *)

 intro XX;clear XX.
 simpl in H.
 destruct H as (loc, (n, (sum, (Heq1,XX)))).
 subst l.
 hnf.
 intros ...
 assert (1 = Int.const 1).
   rewrite Int.const_prop;trivial. 
   compute;split;auto. intro H;discriminate H.
 exists loc;exists n; exists sum; exists sum; exists (Int.const 1);
 simpl;rewrite <- H;repeat split;auto;simpl;try omega.
 change (0/2) with 0;omega.
Time Qed.


