Add LoadPath "../EffectsBicolano".

Require Export BaseLogic.

Open Scope type_scope. 

Module Validity (Rules:RULES).

 Import Rules.

 Section SEM.
  Variable (P: Program).

  Lemma exec_call_isCallInstr : forall M l i s ret' m' s' r' S, 
    instructionAt M l = Some i ->
    exec_call P M ((l, s),S) ret' m' s' r' -> kind_of_instr i = CallInstr.
  Proof.
   intros M l i s ret' m' s' r' S Hi Hcall.
    inversion_mine Hcall.
    match goal with 
    | [H2: CallStep _ _ _ _ |- _] =>
      inversion_mine H2;simpl_eq;trivial
    | _ => idtac
    end.
    match goal with 
    | [H4: _CallStep _ _ _ _ |- _] =>
      inversion_mine H4;simpl_eq;trivial
    | _ => idtac
    end.
    match goal with 
    | [H2: CallStep _ _ _ _ |- _] =>
      inversion_mine H2;simpl_eq;trivial
    | _ => idtac
    end.
    match goal with 
    | [H4: _CallStep _ _ _ _ |- _] =>
      inversion_mine H4;simpl_eq;trivial
    | _ => idtac
    end.
    match goal with 
    | [H2: CallStep _ _ _ _ |- _] =>
      inversion_mine H2;simpl_eq;trivial
    | _ => idtac
    end.
    match goal with 
    | [H4: _CallStep _ _ _ _ |- _] =>
      inversion_mine H4;simpl_eq;trivial
    | _ => idtac
    end.
  Qed.

  Lemma _elemExcep_isBasicInstr : forall M l i s h' loc,
     instructionAt M l = Some i ->
     _ExceptionStep P M (l, s) (h', loc) ->
     kind_of_instr i = BasicInstr \/ 
     (exists mid, i = Invokevirtual mid) \/
     (exists mid, i =(Invokestatic mid) /\ 
        exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true).
  Proof.
    intros M l i s h' loc Hi Hel;inversion_mine Hel; simpl_eq;simpl;auto.
    right;right;eauto.
    inversion_mine H3;simpl_eq;auto;eauto.
   Qed.

  Lemma elemExcep_isBasicInstr : forall M l i s h' loc S R,
     instructionAt M l = Some i ->
     ExceptionStep P M ((l, s),S) ((h', loc),R) ->

     (R = bookkeepException P M (l,s)  (h',loc) S /\
       (kind_of_instr i = BasicInstr \/ 
         (exists mid, i = Invokevirtual mid) \/
         (exists mid, i =(Invokestatic mid) /\ 
            exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true))).
  Proof.
    intros M l i s h' loc S R Hi Hel.
    inversion_mine Hel.
    split; trivial.
    apply ( _elemExcep_isBasicInstr M l i s h' loc Hi H3).
   Qed.

  Lemma _elemIntra_kind : forall M l i s1 s3, 
     instructionAt M l = Some i ->
     _NormalStep P M (l, s1) s3 ->
      match kind_of_instr i with
      | BasicInstr => 
         exists ll, next M l = Some ll /\ exists s2, s3 = (ll, s2)
      | CallInstr => 
         exists mid, i =(Invokestatic mid) /\ 
           exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true
      | ReturnInstr => False
      | IfInstr o => 
            (exists ll, next M l = Some ll /\ exists s2, s3 = (ll,s2))
         \/ (exists s2, s3 = (OFFSET.jump l o, s2))
      | GotoInstr o => exists s2, s3 = (OFFSET.jump l o, s2)
      | LSwitchInstr def listkey =>  
             ((exists ii, exists i', exists o', exists s, exists lv, exists h, 
                 i' = Int.toZ ii /\
                List.In (pair i' o') listkey /\ 
                s1 = (h,(Num (I ii)::s),lv) /\
                s3 = (OFFSET.jump l o',(h,s,lv))) \/
              (exists ii, exists s, exists lv, exists h, 
                 (forall i' o', List.In (pair i' o')listkey ->  i' <> Int.toZ ii) /\
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l def,(h,s,lv))))
       | TSwitchInstr def low high list_offset => 
             ((exists ii, exists h, exists s, exists lv,
                 Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                 (Int.toZ ii < low \/ high < Int.toZ ii) /\
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l def,(h,s,lv))) \/
              (exists ii, exists n, exists o, exists h, exists s, exists lv,
                 Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                 low <= Int.toZ ii <= high /\
                 Z_of_nat n = (Int.toZ ii) - low /\
                 nth_error list_offset n = Some o /\
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l o,(h,s,lv))))
      end.
  Proof.
   intros M l i s1 s3 H Hi;inversion_mine Hi;simpl_eq;simpl;eauto.
   left. exists i0. exists (Int.toZ i0). exists o'. exists s. exists l0.
             exists h. auto. 
   right. exists i0. exists s. exists l0. exists h. auto. 
    left. exists i0. exists h. exists s. exists l0. auto. 
   right. exists i0. exists n. exists o. exists h. exists s. exists l0.  split. auto. auto. 
  Qed.

  Lemma elemIntra_kind : forall M l i s1 s3 S T, 
     instructionAt M l = Some i ->
     NormalStep P M ((l, s1),S) (s3,T) ->
     ( T = bookkeepNormal P M (l,s1) s3 S /\ 
      (match kind_of_instr i with
       | BasicInstr => 
          exists ll, next M l = Some ll /\ exists s2, s3 = (ll, s2)
       | CallInstr => 
          exists mid, i =(Invokestatic mid) /\ 
            exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true
       | ReturnInstr => False
       | IfInstr o => 
             (exists ll, next M l = Some ll /\ exists s2, s3 = (ll,s2))
          \/ (exists s2, s3 = (OFFSET.jump l o, s2))
       | GotoInstr o => exists s2, s3 = (OFFSET.jump l o, s2)
       | LSwitchInstr def listkey =>  
              ((exists ii, exists i', exists o', exists s, exists lv, exists h, 
                  i' = Int.toZ ii /\
                 List.In (pair i' o') listkey /\ 
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l o',(h,s,lv))) \/
               (exists ii, exists s, exists lv, exists h, 
                  (forall i' o', List.In (pair i' o')listkey ->  i' <> Int.toZ ii) /\
                  s1 = (h,(Num (I ii)::s),lv) /\
                  s3 = (OFFSET.jump l def,(h,s,lv))))
        | TSwitchInstr def low high list_offset => 
             ((exists ii, exists h, exists s, exists lv,
                 Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                 (Int.toZ ii < low \/ high < Int.toZ ii) /\
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l def,(h,s,lv))) \/
              (exists ii, exists n, exists o, exists h, exists s, exists lv,
                 Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                 low <= Int.toZ ii <= high /\
                 Z_of_nat n = (Int.toZ ii) - low /\
                 nth_error list_offset n = Some o /\
                 s1 = (h,(Num (I ii)::s),lv) /\
                 s3 = (OFFSET.jump l o,(h,s,lv))))
      end) ).
  Proof.
   intros M l i s1 s3 S T H Hi.
    inversion_mine Hi.
    split; trivial. 
   apply ( _elemIntra_kind M l i s1 s3 H H4).
  Qed.
    
  Lemma _elemReturn_isBasicReturn : forall M l i s r,
      instructionAt M l = Some i ->
      _ReturnStep P M (l,s) r ->
      kind_of_instr i = ReturnInstr.
  Proof. 
   intros M l i s r Hi Hr;inversion_mine Hr;simpl_eq;simpl;trivial.
  Qed.

  Lemma elemReturn_isBasicReturn : forall M l i s r S R,
      instructionAt M l = Some i ->
      ReturnStep P M ((l,s),S) (r,R) ->
      (R = bookkeepReturn P M (l,s) r S /\
       kind_of_instr i = ReturnInstr).
  Proof. 
   intros M l i s r S R Hi Hr.
   inversion_mine Hr.
   split; trivial. 
   apply (_elemReturn_isBasicReturn M l i s r Hi H3).
 Qed.

 End SEM.

 Section VALID.

  Variables (P: Program) 
            (MST: MethSpecTable).

  Definition startState MB (s0:LocalVar.t) (h0:Heap.t) :=
     (BYTECODEMETHOD.firstAddress MB, (h0, OperandStack.empty,s0)).
 
  Hypothesis HGP: GP P MST.
 
  (*A simple fact about object allocation in Bicolano*)
  Lemma NullException: forall h1 h2 loc2 e, 
              Heap.new h1 P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc2,h2) ->
              Heap.typeof h2 loc2 = Some(Heap.LocationObject e) ->
              e = (javaLang, NullPointerException).
  intros.
  assert (X := Heap.new_typeof _ _ _ _ _ H); simpl_eq.
  eauto. 
  Qed.
  
  Lemma lookup_findMethod : forall cn mid cl m,
     lookup P cn mid (cl, m) -> findMethod P (cl,METHOD.signature m) = Some m.
  Proof.
    assert (forall cn mid M,
     lookup P cn mid M -> forall cl m, M =(cl, m) -> findMethod P (cl,METHOD.signature m) = Some m).
     induction 1;simpl;intros.
     injection H0;intros;subst.
     inversion_mine H. simpl in H1.
     destruct (PROG.class P cl);try discriminate.
     rewrite <-  (CLASS.method_signature_prop _ _ _ H1);trivial.
     exact (IHlookup _ _ H2).
    intros; eapply H;eauto.
  Qed.

  Lemma SOUND_POST:  forall (M: Method) s r, 
      V5.BigStep P M s r ->
      forall Msig ms LST LAT , MSPEC_LOOKUP MST Msig = Some ((ms,,LST),,LAT) ->
      findMethod P Msig = Some M ->
      forall MB , METHOD.body M = Some MB ->
      forall b A B I l s0 s1 R ss,
        s = ((l,s1),R) ->
        ss = (s1,R) ->
        SP_Judgement P MST M LST LAT b l A B I -> 
        A s0 ss ->
        B s0 ss r.
  Proof.
  induction 1 using V5.BigStep_ind;intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 
   l s0 s1 R ss Heq1 Heq2 Hsp HA;subst;
   destruct (GP_SP_Judgement_inv _ _  HGP _ _ _ _ _ Hlookup Hfind _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP)); clear Hsp.
(* exec_return *)  
  inversion_mine H.
   (* return *)
    assert (H0_0 := elemReturn_isBasicReturn _ _ _ _ _ _ _ act  HinstrAt H0). destruct H0_0. subst.
    assert (H0_1 := H0).
     inversion_mine H0. 
    rewrite (_elemReturn_isBasicReturn _ _ _ _ _ _ HinstrAt H5) in HSP;
     inversion_mine HSP;eauto.
   (* uncaught exception *)
     assert (H4_0 := elemExcep_isBasicInstr _  _ _ _ _ _ _ _ _ HinstrAt H4). destruct H4_0; subst. 
    assert (H4_1 := H4).
     inversion_mine H4. 
   destruct (_elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H6) as 
    [Heq| [(mid,Heq) | (mid,(Heq,(m0,(HH1,HH2))))]]; simpl_eq;
    rewrite Heq in HSP.
    (* basic instruction *)
    inversion_mine HSP;eauto.
    (*InvokeVirtual*)
    destruct HSP as (msig, X).
    destruct X.  destruct H. simpl_eq. discriminate H3.
    destruct H as (Heq1,SP_virt).  simpl_eq. inversion_mine H0. 
    inversion_mine H1; simpl_eq.
    inversion_mine H; simpl_eq.
    inversion_mine SP_virt.
    clear H8. clear H10. clear H11. clear H12. 
    destruct H. destruct H.  inversion_mine H; subst. inversion_mine H1; subst. 
    assert (HAA := H15 _ _ HA). 
    inversion_mine H6; simpl_eq.
    inversion_mine H9; simpl_eq.
    clear H13.
    destruct (H14 _ _ _ _ _ _  (bookkeepException P m (l,((h1,args ++ Null :: s),l1)) (h2,loc2) R) HAA _ _ H18 H19) as (H14_1,H14_2). clear H14. clear H14_1.
    assert (H14_3 := H14_2 H5 (refl_equal _)). eauto. 

   destruct H. destruct H. discriminate H. 
    (* invokestatic*)
    inversion_mine HSP. simpl_eq.
    destruct H.
       destruct H; simpl_eq. 
       inversion_mine H1; simpl_eq. 
    destruct H; simpl_eq.
(* exec_intra *)
  inversion_mine H.
   (* No exception *)
     assert (H_1 := H1).
     inversion_mine H1.
    assert (HH := _elemIntra_kind _ _ _ _ _ _ HinstrAt H4);
    destruct (kind_of_instr instr);simpl in HH, HSP;
    match goal with [H:False |- _ ] => elim H | _ => idtac end.
    (* isBasicInstr *)
    destruct HH as (ll, (Heq1, (s2, Heq2))); inversion_mine HSP;simpl_eq;eauto.
    assert (Hpre:SP_pre P m l ll A0 s0 (s2,bookkeepNormal P m (l,s1) (ll,s2) R)). econstructor;eauto. 
    assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) (refl_equal _)
                  (H2 _ Heq1) Hpre);unfold SP_post in UU;eauto.
    (* A native call*)
    destruct HH as (msig, (Heq1, Hsp));inversion_mine Hsp;simpl_eq.
    destruct H; simpl_eq. 
    destruct HSP. 
    destruct H2. 
       destruct H2; simpl_eq. inversion_mine H3; simpl_eq.
       destruct H2; simpl_eq. 
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    assert (Hpre:SP_pre P m l ll A0 s0 (s2,bookkeepNormal P m (l,s1) (ll,s2) R)). econstructor;eauto. 
    assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _)  (refl_equal _)
                  H5 Hpre);unfold SP_post in UU;eauto.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2,bookkeepNormal P m (l,s1) (OFFSET.jump l o,s2) R)). 
      econstructor;eauto.
    assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) (refl_equal _)
                  H3 Hpre);unfold SP_post in UU;eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (s2,Heq);simpl_eq;subst.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2,bookkeepNormal P m (l,s1) (OFFSET.jump l o,s2) R)).
      econstructor;eauto.
    assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _)  (refl_equal _)
                  H2 Hpre);unfold SP_post in UU;eauto.
   (*LSwitch*)
     destruct HSP as (Instr,HSP). inversion_mine HSP. 
     destruct HH. 
     (*Entry exists*)
        destruct H8 as (ii,H8).  
        destruct H8 as (i',H8). 
           destruct H8 as (o',H8).   
           destruct H8 as (s,H8). 
           destruct H8 as (lv,H8). 
           destruct H8 as (h,H8). 
           destruct H8 as (i_ii,H8). 
           destruct H8 as (entry,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o') A0 s0 ((h,s,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l o', ((h,s), lv)) R)).
             econstructor;eauto.
          assert (exists i' : Z, (exists i : Int.t, In (i', o') l0 /\ i' = Int.toZ i)). 
            exists (Int.toZ ii). exists ii. split. trivial.  trivial. 
          assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _  _ _ _ (refl_equal _)  (refl_equal _) (H2 o' H8) Hpre).  eauto. 
     (*No entry exists - default def is taken*)
        destruct H8 as (ii,H8).  
           destruct H8 as (s,H8). 
           destruct H8 as (lv,H8). 
           destruct H8 as (h,H8). 
           destruct H8 as (no_entry,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l def) A0 s0 
                       ((h,s,lv),
                        bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l def, ((h,s), lv)) R)).
             econstructor; eauto.
          assert (H3_1 := H3 ii no_entry); clear H3. 
          assert (LS: SP_LSwitch_Def_pre P m l (OFFSET.jump l def) A0 ii s0 ((h,s,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l def, ((h,s), lv)) R)).
            econstructor. eauto. eauto. eauto. trivial.
          assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ s0  _ _ _ (refl_equal _) (refl_equal _) H3_1 LS). 
          unfold SP_LSwitch_Def_post in UU. 
          assert (U_ : exists h0, exists ss, exists lv0, (h, Num(I ii)::s, lv) = (h0, Num(I ii) :: ss, lv0)). 
            exists h; exists s; exists lv. trivial. 
          assert (UU1 := UU ((h,Num(I ii)::s, lv),R)).  
          assert (UU2 := UU1 _ _ _ _  U_ (refl_equal _) (refl_equal _) H_1).
          eauto.
   (*TSwitch*)
    destruct HSP as (Instr, HSP). inversion_mine HSP. 
     destruct HH. 
     (*Defaulf jump*)
        destruct H8 as (ii,H8).  
        destruct H8 as (h,H8). 
           destruct H8 as (s,H8).   
           destruct H8 as (lv,H8). 
           destruct H8 as (arith1,H8). 
           destruct H8 as (arith2,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l def) A0 s0 ((h,s,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l def, ((h,s), lv)) R)).
             econstructor;eauto.
          assert (exists i : Int.t, Z_of_nat (length loffset) = (high - low +1)%Z /\ (Int.toZ i < low \/ high < Int.toZ i)). 
            exists ii. split. trivial. trivial. 
          assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _  _ _ _ (refl_equal _) (refl_equal _) (H2 H8) Hpre).  eauto. 
     (*No entry exists - default def is taken*)
        destruct H8 as (ii,H8).  
           destruct H8 as (n,H8). 
           destruct H8 as (o,H8). 
           destruct H8 as (h,H8).  
           destruct H8 as (s,H8). 
           destruct H8 as (lv,H8). 
           destruct H8 as (arith1,H8). 
           destruct H8 as (arith2,H8). 
           destruct H8 as (arith3,H8). 
           destruct H8 as (loff,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 ((h,s,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l o, ((h,s), lv)) R)).
             econstructor;eauto.
           assert (X : (exists iii:Int.t, exists n:nat, 
                 Z_of_nat (length loffset) = (high - low + 1)%Z /\
                 low <= Int.toZ iii <= high /\
                 Z_of_nat n = (Int.toZ iii) - low /\
                 nth_error loffset n = Some o)).
            exists ii. exists n. split. trivial. split. trivial. split. trivial. trivial. 
          assert (H3_1 := H3 o X); clear H3. 
          assert (UU := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _  _ _ _ (refl_equal _) (refl_equal _) H3_1 Hpre).  unfold SP_post in UU;eauto. 
   (* Exception Caught *)
     assert (H5_0 := elemExcep_isBasicInstr _  _ _ _ _ _ _ _ _ HinstrAt H5). destruct H5_0; subst. 
     assert (H5_1 := H5).
     inversion_mine H5.   
     destruct (_elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H7) as 
    [Heq| [(mid,Heq) | (mid,(Heq,(m0,(HH1,HH2))))]]; simpl_eq;
    rewrite Heq in HSP.
    (* basic instruction *)
    clear H1.
    inversion_mine HSP.
    clear H4. clear H2. inversion_mine H6.
    cut (B1 s0 (((h1, s2),l1),R) r). eauto. clear H8. 
    assert (H3_1 := H3 _ e _ H13 H16). clear H3.
    assert (IHB := IHBigStep _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ s0 _ _ _ (refl_equal _) (refl_equal _) H3_1).
    assert (SP_preEC P m l e A0 s0 (((h2,Ref loc2 :: OperandStack.empty),l1), bookkeepException P m (l,(h1,s2,l1)) (h2,loc2) R)). 
       econstructor. apply (H5 _ _ HA).
       apply H5_1. assumption. trivial.
    assert (IHB_1 := IHB H2). clear IHB.
    unfold SP_postEC in IHB_1.
    assert (IHB_2 := IHB_1 _ _ _ _ _ _ _ _ (refl_equal _) H5_1 H15 (refl_equal _)). assumption.  
    (* A virtual call absurd up to now *)
    destruct HSP. destruct H; simpl_eq.
      destruct H; simpl_eq. discriminate H4. 
      destruct H; simpl_eq.  inversion_mine H4. 
         inversion_mine H1;simpl_eq.  
         inversion_mine H; simpl_eq.
         inversion_mine H. destruct H1. inversion_mine H. 
         inversion_mine H2; simpl_eq. 
         inversion_mine H7; simpl_eq.
         inversion_mine H19; simpl_eq.
         assert (H15_1 := H15 _ _ HA).
         assert (B1_1 : B1 s0 (((h1, args ++ Null :: s),l1),R) r).  
           assert (IHBS := IHBigStep _ _ _ _ Hlookup Hfind _ HMB). clear IHBigStep.
           clear H10. clear H11. clear H12. clear H13.
           assert (H14_1 := H14 _ _ _ _ _ _ (bookkeepException P m (l,((h1,args ++ Null :: s),l1)) (h2,loc2) R) H15_1 _ _ H23 H24). clear H14.
           destruct H14_1. clear H7.
           inversion_mine H6; simpl_eq. 
           cut (e=(javaLang,NullPointerException)). intros; simpl_eq. subst.
             assert (H2_1 := H2 _ _ HMB H18). clear H2.
             cut (VINV_preECNull P m l A0 (length (METHODSIGNATURE.parameters (snd mid)))  s0
                   ((h2,Ref loc2 :: OperandStack.empty, l1), bookkeepException P m (l,((h1,args ++ Null :: s),l1)) (h2,loc2) R)) . intros.
                assert (IHBS1 := IHBS _ _ _ _ _ _ _ _ _ (refl_equal _) (refl_equal _) H2_1 H2). 
                inversion_mine IHBS1; simpl_eq.  eauto.
             econstructor. eauto. eauto. eauto. eauto. trivial. 
           apply (NullException _ _ _ _ H24 H14).
       eauto.
     destruct H1; simpl_eq.
     destruct H; simpl_eq. discriminate H. 
   (*Invokestatic*)
    simpl_eq. destruct HSP. simpl_eq.  destruct H.  
         simpl_eq. destruct H; simpl_eq.  inversion_mine H4; simpl_eq. inversion_mine H1; simpl_eq. 
           simpl_eq. destruct H; simpl_eq.
           inversion_mine H2; simpl_eq.
         destruct H. destruct H. discriminate H.
         destruct H. destruct H. destruct H; simpl_eq. destruct H1; simpl_eq. destruct H.
           inversion_mine H7; simpl_eq.
           inversion_mine H2; simpl_eq.
           inversion_mine H8; simpl_eq.
         destruct H; simpl_eq. discriminate H4. 

  (* exec_call, uncaught exception *)
  assert (CALL:= exec_call_isCallInstr _ _ _ _ _ _ _ _ _ _ HinstrAt H).
  rewrite CALL in HSP.
   destruct HSP as (msig, Invoke). 
   destruct Invoke. 
   (*InvStatic*)
     destruct H1 as (Heq1,HSP); simpl_eq. 
     inversion_mine HSP; simpl_eq.
     inversion_mine H; simpl_eq.
    assert (H15_1 := H11 _ _ HA).
    assert (B1_1 : B1 s0 (((h1, s2),l1),R) ((h2,Exception loc), act3)).
    Focus 2. eauto.

    assert (H20_0 := H20). 
    inversion_mine H20;simpl_eq.
    inversion_mine H16;simpl_eq.
    assert (SPEC : mkMethSpec mpre mpost ((stack2localvar (args ++ os) (length args),h1),
                                  bookkeepCall P m (l,((h1,args ++ os) ,l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) ((h2,Exception loc), act3)).
    Focus 2. assert (H10_1 := H10 _ _ _ _ _ _ _ H15_1 _ _ _ _ _ _ H26 H20_0 SPEC H22). assumption.

     unfold mkMethSpec.
    destruct HGP as (BST0,HGP0). clear HGP.
       destruct (HGP0 _ _ _ _ _ _ H3 _ H24) as (HGP0_1,HGP0_2). 
       assert (HGP0_1_1 := HGP0_1 _ H21). clear HGP0_1.
       intros MPRE.
       assert (IHBS := IHBigStep _ _ _ _ H3 H24 _ H21 _ _ _ _ _ (((stack2localvar (args ++ os) (length args)),h1),bookkeepCall P m (l,((h1,args ++ os) ,l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) _ _ _ (refl_equal _) (refl_equal _) HGP0_1_1). clear IHBigStep.
       simpl in IHBS. eauto.
  (*InvVirt*)
    destruct H1 as (Heq,HSP). inversion_mine HSP.
    simpl_eq. clear H7. clear H11.
    clear H9. clear H8. 
    cut (B1 s0 (s1,R) r).  intro B1_1. eauto.
    assert (HHH := H).  
    inversion_mine H.
    assert (H17HH := H17).  inversion_mine H17;simpl_eq. inversion_mine H9. simpl_eq.
    assert (H10_1 := H10 _ _ _ _ _ _ _ act3 (H12 _ _ HA) h2 loc m' os); clear H10. simpl_eq.
    cut (mkMethSpec mpre mpost 
             ((stack2localvar (args ++ Ref loc0 :: os)
               (S(length args)), h1), 
               bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                  (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)
             ((h2,Exception loc), act3)). 
      intros MSPEC. eauto.
   clear H10_1.
    (*Situation: cn:name of dynamic dynamic class of object at loc0 in h1
                       (cl,m') : class and method found when searching in
                                    cn for method mid
                                   (so m' is the dynamically invoked method)
                       m0: Method statically associated with mid*)
    destruct HGP as (BST0,HGP0). 
    assert (BST_1 := BST0 _ _ _ _ _ _ _ _ _  HinstrAt H1 H3 H21). 
        destruct BST_1 as (ms',(LST',(LAT',(MSTlookup',SUBCLASS)))).  inversion_mine SUBCLASS. 
        clear H6. 
        cut (A2 = mpre); intros; subst. Focus 2. congruence. 
        cut (B3 = mpost); intros; subst. Focus 2. congruence. 
        cut (mkMethSpec A1 B2 
                   ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)), h1), 
                    bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                  (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)                    
                   ((h2,Exception loc), act3)). eauto. 
        clear H; clear H8. 
        cut (A1 ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)), h1), 
                    bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                  (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R) ->
                    B2 ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)), h1),
                           bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                  (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)
                         ((h2,Exception loc), act3)). intros. eauto.
        intro A1_1.

    destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H21)) as (HGP1,HGP2); clear HGP2.     
    assert (HGP1_1 := HGP1 bM H18); clear HGP1. 
    assert (IHBS_1 := IHBigStep _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H21) _ H18); clear IHBigStep.
    assert (IHBS_2 := IHBS_1 _ _ _ _ _ 
                                ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)), h1),
                                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)
                                _ _ _ (refl_equal _) (refl_equal _) HGP1_1). clear IHBS_1.
     simpl in IHBS_2.
   cut (A1 ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)), h1),
                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R) /\ 
           (((h1,OperandStack.empty),stack2localvar (args ++ Ref loc0 :: os) (S(length args))),
                                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R) =    
           (((h1,OperandStack.empty),stack2localvar (args ++ Ref loc0 :: os) (S(length args))),
                                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)). intros X. 
   assert (IHBS_3 := IHBS_2 X); clear IHBS_2.   
   cut ((((h1,OperandStack.empty),stack2localvar (args ++ Ref loc0 :: os) (S(length args))),
                                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R) =    
           (((h1,OperandStack.empty),stack2localvar (args ++ Ref loc0 :: os) (S(length args))),
                                  bookkeepCall P m (l,((h1,args ++ (Ref loc0) :: os) ,l1))
                                    (m', (os, stack2localvar (args ++ (Ref loc0) :: os) (S(length args)))) R)). intros Y. 
   assert (IHBS_4 := IHBS_3 Y); clear IHBS_3.   eauto. 
   congruence. 
   split. eauto. congruence.   

 (* exec_call *)
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP.
     destruct HSP as (msig,HSP).  inversion_mine HSP.
     (*InvStatic*)
     destruct H2; simpl_eq. (*herePatryk*)
     inversion_mine H.    
        (*1/2*)   
           inversion_mine H3;simpl_eq.
           assert (H10H := H10). 
           inversion_mine H10. inversion_mine H20; simpl_eq.
           clear H8. clear H13. clear H14.
           cut (mkMethSpec mpre mpost
                       ((stack2localvar (args ++ os) (length args),h1), 
                           bookkeepCall P m (l, ((h1, args ++ os) ,l1))
                                                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                       ((h2, Normal ov), act3)). intro SPEC.
              assert (H9_1 := H9 _ H12); clear H9. 
              assert(IHBS2_1 := IHBigStep2 _ _ _ _ Hlookup Hfind _ HMB); clear IHBigStep2.           
              assert (IHBS2_2 := IHBS2_1 _ _ _ _ _ s0 _ _ _ (refl_equal _) (refl_equal _) H9_1); clear IHBS2_1.
              assert (Hpre: SINV_pre mpre mpost A0 (length(METHODSIGNATURE.parameters (snd mid)))
                         P m l s0 (((h2, cons_option ov os), l1), act3)). 
                    econstructor.  eauto. eauto. eauto. rewrite H27. rewrite H27 in SPEC. eauto. eauto.
              assert (IHBS2_3 := IHBS2_2 Hpre); clear IHBS2_2. 
                   inversion_mine IHBS2_3. 
                   assert (H_1 := H _ _ _ _ _ _ _ _ _ _ _ _ _ H27 H10H SPEC (refl_equal _) (refl_equal _)). eauto.
           (*discharge of mkMethSpec*) 
           destruct HGP as (BST0,HGP0). 
           destruct (HGP0 _ _ _ _ _ _ H4 _ H25) as (HGPmeth, HGPctxt).
           assert (HGP_m0 := HGPmeth _ H11). clear HGPmeth; clear HGPctxt.
           clear H9; clear IHBigStep2. 
           assert (IHBS1 := IHBigStep1 _ _ _ _ H4 H25 _ H11); clear IHBigStep1. 
           unfold mkMethSpec; intro MPRE.
           assert (IHBS2 := IHBS1 _ _ _ _ _ 
                                               ((stack2localvar (args ++ os) (length args), h1), 
                                                 bookkeepCall P m (l, ((h1, args ++ os) ,l1))
                                                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                                               _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear IHBS1.
           assert ( (fun (s0:InitState) (s:LocalState) => let (ssH,R) := s0 in
                                    let (ss,H) := ssH in
                                    mpre s0 /\ s = (((H,OperandStack.empty),ss),R))
                  ((stack2localvar (args ++ os) (length args) , h1),
                    bookkeepCall P m (l,((h1,args ++ os), l1))
                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                  (((h1, OperandStack.empty), stack2localvar(args ++ os) (length args)),
                    bookkeepCall P m (l, ((h1, args ++ os), l1)) 
                            (m0,(os, stack2localvar(args ++ os)(length args))) R)).  split.  eauto.  eauto.
           assert (IHBS3 := IHBS2 H).  clear IHBS2. eauto.                  
        (*2/2: caught exception*) 
        inversion_mine H3;simpl_eq. 
           assert (H10H := H10). 
           inversion_mine H10. inversion_mine H20; simpl_eq.
           clear H8. clear H9. clear H14.
           cut (mkMethSpec mpre mpost
                       ((stack2localvar (args ++ os) (length args),h1),bookkeepCall P m (l, ((h1, args ++ os) ,l1))
                                                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                       ((h2, Exception loc), act3)). intro SPEC.
              inversion_mine H12; simpl_eq. 
              assert (H13_1 := H13 _ _ _ HMB H18); clear H13. 
              assert(IHBS2_1 := IHBigStep2 _ _ _ _ Hlookup Hfind _ HMB); clear IHBigStep2.           
              assert (IHBS2_2 := IHBS2_1 _ _ _ _ _ s0 _ _ _ (refl_equal _) (refl_equal _) H13_1); clear IHBS2_1.
              assert (Hpre: SINV_preEC mpre mpost A0 (length(METHODSIGNATURE.parameters (snd mid)))
                         P m l e s0 (((h2, Ref loc :: nil), l1), act3)). 
                    econstructor.  eauto. eauto. eauto. eauto. eauto. eauto. 
              assert (IHBS2_3 := IHBS2_2 Hpre); clear IHBS2_2. 
                   inversion_mine IHBS2_3. 
                   assert (H_1 := H _ _ _ _ _ _ _ _ _ _ _ _  _ H27 H10H SPEC H14 (refl_equal _) (refl_equal _)). eauto.
           (*discharge of mkMethSpec*) 
           destruct HGP as (BST0,HGP0). 
           destruct (HGP0 _ _ _ _ _ _ H4 _ H25) as (HGPmeth, HGPctxt).
           assert (HGP_m0 := HGPmeth _ H11). clear HGPmeth; clear HGPctxt.
           clear H13; clear IHBigStep2. 
           assert (IHBS1 := IHBigStep1 _ _ _ _ H4 H25 _ H11); clear IHBigStep1. 
           unfold mkMethSpec; intro MPRE.
           assert (IHBS2 := IHBS1 _ _ _ _ _ 
                                               ((stack2localvar (args ++ os) (length args), h1), 
                                                 bookkeepCall P m (l, ((h1, args ++ os) ,l1))
                                                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                                               _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear IHBS1.
           assert ( (fun (s0:InitState) (s:LocalState) => let (ssH,R) := s0 in
                                    let (ss,H) := ssH in
                                    mpre s0 /\ s = (((H,OperandStack.empty),ss),R))
                  ((stack2localvar (args ++ os) (length args) , h1),
                    bookkeepCall P m (l,((h1,args ++ os), l1))
                      (m0, (os, stack2localvar(args ++ os) (length args))) R)
                  (((h1, OperandStack.empty), stack2localvar(args ++ os) (length args)),
                    bookkeepCall P m (l, ((h1, args ++ os), l1)) 
                            (m0,(os, stack2localvar(args ++ os)(length args))) R)).  split.  eauto.  eauto.
           assert (IHBS3 := IHBS2 H).  clear IHBS2. eauto.
     (*InvVirtual*)
     destruct H2; simpl_eq.
     inversion_mine H.    
        (*1/2*)   
           inversion_mine H3;simpl_eq.
           assert (H10H := H10). 
           inversion_mine H10. inversion_mine H21; simpl_eq.
           clear H8. clear H13. clear H14.
           cut (mkMethSpec mpre mpost
                       ((stack2localvar (args ++ Ref loc :: os) (S(length args)),h1), 
                           bookkeepCall P m (l, ((h1, args ++ Ref loc :: os) ,l1))
                                                      (m', (os, stack2localvar(args ++ Ref loc :: os) (S(length args)))) R)
                       ((h2, Normal ov), act3)). intro SPEC.
              assert (H9_1 := H9 _ H12); clear H9. 
              assert(IHBS2_1 := IHBigStep2 _ _ _ _ Hlookup Hfind _ HMB); clear IHBigStep2.           
              assert (IHBS2_2 := IHBS2_1 _ _ _ _ _ s0 _ _ _ (refl_equal _) (refl_equal _) H9_1); clear IHBS2_1.
              assert (Hpre: VINV_pre mpre mpost A0 (length args)
                         P m l s0 (((h2, cons_option ov os), l1), act3)). 
                    econstructor.  eauto. eauto. eauto. rewrite H28. rewrite H28 in SPEC. eauto. eauto.
              rewrite H28 in Hpre. 
              assert (IHBS2_3 := IHBS2_2 Hpre); clear IHBS2_2. 
                   inversion_mine IHBS2_3.
                   rewrite H28 in SPEC. 
                   assert (H3_1 := H3 _ _ h2 _ _ ov (((h1,args ++ Ref loc :: os),l1),R) _ _ _ _ _ _ act3 H28 H10H). clear H3.
                   rewrite H28 in H3_1. eauto. 
           (*discharge of mkMethSpec*) 
            destruct HGP as (BST0,HGP0). 
            assert (BST_1 := BST0 _ _ _ _ _ _ _ _ _ HinstrAt H H4 H26). 
            destruct BST_1 as (ms',(LST',(LAT',(MSTlookup',SUBCLASS)))).  inversion_mine SUBCLASS. 
            clear H6. 
            cut (A2 = mpre); intros; subst. Focus 2. congruence.  
            cut (B3 = mpost); intros; subst. Focus 2. congruence. 
            cut (mkMethSpec A1 B2 
                       ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc :: os), l1)) 
                             (m',(os, stack2localvar(args ++ Ref loc :: os)(S(length args)))) R)
                       ((h2,Normal ov), act3)). eauto. 
            clear H3; clear H9. 
            cut (A1 ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc :: os), l1)) 
                             (m',(os, stack2localvar(args ++ Ref loc :: os)(S(length args)))) R) ->
                        B2 ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                              bookkeepCall P m (l, ((h1, args ++ Ref loc :: os), l1)) 
                               (m',(os, stack2localvar(args ++ Ref loc :: os)(S(length args)))) R)
                       ((h2,Normal ov),act3)). intros. eauto.
            intro A1_1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup'  _ (lookup_findMethod _ _ _ _ H26)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H11). clear HGPmeth; clear HGPctxt.
            clear H12; clear IHBigStep2. 
            assert (IHBS1 := IHBigStep1 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H26) _ H11); clear IHBigStep1. 
            assert (IHBS2 := IHBS1 _ _ _ _ _ 
                                                ((stack2localvar (args ++ Ref loc :: os) (1+length args),h1), bookkeepCall P m (l, ((h1, args ++ Ref loc :: os), l1)) 
                                                 (m',(os, stack2localvar(args ++ Ref loc :: os)(S(length args)))) R) 
                                               _ _ _ (refl_equal _) (refl_equal _) HGP_m0). clear IHBS1.
            rewrite H28.
            rewrite H28 in IHBS2. 
            rewrite H28 in A1_1. eauto.          
        (*2/2: caught exception*) 
        inversion_mine H3;simpl_eq. 
           assert (H10H := H10). 
           inversion_mine H10. inversion_mine H21; simpl_eq.
           clear H8. clear H9. clear H14. clear H15.
           cut (mkMethSpec mpre mpost
                       ((stack2localvar (args ++ Ref loc0 :: os) (S(length args)),h1), 
                           bookkeepCall P m (l, ((h1, args ++ Ref loc0 :: os) ,l1))
                                                      (m', (os, stack2localvar(args ++ Ref loc0 :: os) (S(length args)))) R)
                       ((h2, Exception loc), act3)). intro SPEC.
              inversion_mine H12; simpl_eq. 
              assert (H13_1 := H13 _ _ _ HMB H19); clear H13. 
              assert(IHBS2_1 := IHBigStep2 _ _ _ _ Hlookup Hfind _ HMB); clear IHBigStep2.           
              assert (IHBS2_2 := IHBS2_1 _ _ _ _ _ s0 _ _ _ (refl_equal _) (refl_equal _) H13_1); clear IHBS2_1.
              assert (Hpre: VINV_preEC mpre mpost A0 (length(METHODSIGNATURE.parameters (snd (cn,mid))))
                         P m l e s0 (((h2, Ref loc :: nil), l1), act3)).
                    econstructor.  eauto. eauto. eauto. rewrite H28. rewrite H28 in SPEC. eauto. eauto. eauto.
              assert (IHBS2_3 := IHBS2_2 Hpre); clear IHBS2_2. 
                   inversion_mine IHBS2_3. 
                   rewrite H28 in SPEC.
                   assert (H3_1 := H3 _ _ h2 _ _ (((h1,args ++ Ref loc0 :: os),l1),R) loc _ _ _ _ _ _ act3 H28 H10H). clear H3.
                   rewrite H28 in H3_1. eauto. 
           (*discharge of mkMethSpec*) 
            destruct HGP as (BST0,HGP0). 
            assert (BST_1 := BST0 _ _ _ _ _ _ _ _ _ HinstrAt H H4 H26). 
            destruct BST_1 as (ms',(LST',(LAT',(MSTlookup',SUBCLASS)))).  inversion_mine SUBCLASS. 
            clear H6. 
            cut (A2 = mpre); intros; subst. Focus 2. congruence.  
            cut (B3 = mpost); intros; subst. Focus 2. congruence. 
            cut (mkMethSpec A1 B2 
                       ((stack2localvar (args ++ Ref loc0:: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc0 :: os), l1)) 
                             (m',(os, stack2localvar(args ++ Ref loc0 :: os)(S(length args)))) R)
                       ((h2,Exception loc), act3)). eauto. 
            clear H3; clear H9. 
            cut (A1 ((stack2localvar (args ++ Ref loc0 :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc0 :: os), l1)) 
                             (m',(os, stack2localvar(args ++ Ref loc0 :: os)(S(length args)))) R) ->
                        B2 ((stack2localvar (args ++ Ref loc0 :: os) (1 + length args), h1),
                              bookkeepCall P m (l, ((h1, args ++ Ref loc0 :: os), l1)) 
                               (m',(os, stack2localvar(args ++ Ref loc0 :: os)(S(length args)))) R)
                       ((h2,Exception loc),act3)). intros. eauto.
            intro A1_1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup'  _ (lookup_findMethod _ _ _ _ H26)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H11). clear HGPmeth; clear HGPctxt.
            clear H12; clear IHBigStep2. 
            assert (IHBS1 := IHBigStep1 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H26) _ H11); clear IHBigStep1. 
            assert (IHBS2 := IHBS1 _ _ _ _ _ 
                                                ((stack2localvar (args ++ Ref loc0 :: os) (1+length args),h1), 
                                                 bookkeepCall P m (l, ((h1, args ++ Ref loc0 :: os), l1)) 
                                                 (m',(os, stack2localvar(args ++ Ref loc0 :: os)(S(length args)))) R) 
                                               _ _ _ (refl_equal _) (refl_equal _) HGP_m0). clear IHBS1.
            rewrite H28.
            rewrite H28 in IHBS2. 
            rewrite H28 in A1_1. eauto.      
Qed.


 Lemma SOUND_ANNO:  forall (M: Method) s s', 
      V5.IntraStepStar_intra P M s s' ->
      forall Msig ms LST LAT, MSPEC_LOOKUP MST Msig = Some ((ms,,LST),,LAT) ->
      findMethod P Msig = Some M ->
      forall MB , METHOD.body M = Some MB ->
      forall b A B I s0 l s1 l' s1' ss ss' R S,
        s = ((l,s1),R) ->
        ss = (s1,R) ->
        s' = ((l',s1'),S) ->
        ss' = (s1',S) -> 
        SP_Judgement P MST M LST LAT b l A B I -> 
        A s0 ss ->
        forall (Q:Assertion), Anno_LOOKUP LAT l' = Some Q -> 
        Q s0 ss'.
 Proof.
  induction 1 using V5.IntraStepStar_intra_ind;
   intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 s0 l 
          ss1 l' ss1' ss ss' R S Heq1 Heq1a Heq2 Heq2a Hsp HA Q HQ;subst;
   destruct (GP_SP_Judgement_inv _ _ HGP _ _ _ _ _ Hlookup Hfind _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  (* Base case : s = s' *)
  injection Heq2;clear Heq2;intros;subst.
  destruct (kind_of_instr instr);inversion_mine HSP;eauto.
   eapply H;eauto.
   destruct H. 
   (*InvStatic*)
     destruct H as (H1,H2);inversion_mine H2;eauto. eapply H4;eauto.
   (*InvVirtual*)
     destruct H as (H1,H2);inversion_mine H2;eauto. eapply H4;eauto.
   (*Return*)
   eapply H;eauto.
   (*If*)
   eapply H0;eauto.
   (*Goto*)
   eapply H;eauto.
  (*Lswitch*)
   simpl_eq. subst. destruct H0;eauto. eapply H;eauto. 
  (*Tswitch*)
   simpl_eq. subst. destruct H0;eauto. eapply H;eauto.
  (* exec_intra *)
  inversion_mine H.
   (* No exception *)
    destruct s' as (s',U). 
   assert (HH := elemIntra_kind _ _ _ _ _ _ _ _ HinstrAt H1);
    destruct (kind_of_instr instr);simpl in HH, HSP;
    match goal with [H:False |- _ ] => elim H | _ => idtac end.
    (* isBasicInstr *)
    destruct HH as (HU,(ll, (Heq1, (s2, Heq2)))); inversion_mine HSP;simpl_eq;eauto.
    assert (Hpre:SP_pre P m l ll A0 s0 (s2, bookkeepNormal P m (l,ss1)(ll,s2) R)). econstructor;eauto.
    eapply IHIntraStepStar_intra;eauto.
    (* A native call, absurd up to now *)
     destruct HH as (HU,(mid, (Heq2,(m0,(Heq3,Heq4))))). 
     destruct HSP as (msig, HSP1).
     destruct HSP1. 
       (*InvStat*)
       destruct H; simpl_eq. inversion_mine H4. 
       inversion_mine H2. simpl_eq. 
       (*InvVirt*)
       destruct H; simpl_eq. discriminate H4.
     destruct HH as (HU,FALSE). contradiction. 
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as (HU,HH).
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    assert (Hpre:SP_pre P m l ll A0 s0 (s2,bookkeepNormal P m (l,ss1)(ll,s2) R)). econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2,bookkeepNormal P m (l,ss1)(OFFSET.jump l o,s2) R)).
      econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (HU,HH);destruct HH as (s2,Heq);simpl_eq;subst.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2,bookkeepNormal P m (l,ss1)(OFFSET.jump l o,s2) R)). 
      econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    destruct HH as (HU,HH).
    (* Lswitch *)
    destruct HSP as (Instr, HSP); inversion_mine HSP. 
      destruct HH. 
      (*extry1 exists*) 
        destruct H8 as (ii,H8).  
        destruct H8 as (i',H8). 
           destruct H8 as (o',H8).   
           destruct H8 as (s,H8). 
           destruct H8 as (lv,H8). 
           destruct H8 as (h,H8). 
           destruct H8 as (i_ii,H8). 
           destruct H8 as (entry,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o') A0 s0
                                                      ((h,s,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: s), lv)) (OFFSET.jump l o', ((h,s), lv)) R)). econstructor; eauto.
         eapply IHIntraStepStar_intra;eauto.
      (*extry2 exists*) 
        destruct H8 as (ii,H8).  
        destruct H8 as (ops,H8). 
           destruct H8 as (lv,H8).
           destruct H8 as (h,H8). 
           destruct H8 as (i_ii,H8). 
           destruct H8 as (entry,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst.
           destruct s' as (pc,SS).
           assert (Hpre:SP_pre P m l pc A0 s0
                                                      (SS,bookkeepNormal P m (l, ((h, Num(I ii) :: ops), lv)) (pc,SS) R)). econstructor; eauto.
           assert (H1H := H1).
           inversion_mine H1; simpl_eq.
           inversion_mine H12; simpl_eq.   
             eapply IHIntraStepStar_intra;eauto.
           eapply IHIntraStepStar_intra;eauto. econstructor.  eauto. eauto. eauto. eauto.
   (*TSwitch*)
    destruct HSP as (Instr, HSP). inversion_mine HSP. 
      destruct HH as (HU,HH).
     destruct HH. 
     (*Defaulf jump*)
        destruct H8 as (ii,H8).  
        destruct H8 as (h,H8). 
           destruct H8 as (ops,H8).   
           destruct H8 as (lv,H8). 
           destruct H8 as (arith1,H8). 
           destruct H8 as (arith2,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l def) A0 s0 
                                            ((h,ops,lv), bookkeepNormal P m (l, ((h, Num(I ii) :: ops), lv)) (OFFSET.jump l def,(h,ops,lv)) R)).
             econstructor;eauto.
           eapply  IHIntraStepStar_intra;eauto.
     (*No entry exists - default def is taken*)
        destruct H8 as (ii,H8).  
           destruct H8 as (n,H8). 
           destruct H8 as (o,H8). 
           destruct H8 as (h,H8).  
           destruct H8 as (ops,H8). 
           destruct H8 as (lv,H8). 
           destruct H8 as (arith1,H8). 
           destruct H8 as (arith2,H8). 
           destruct H8 as (arith3,H8). 
           destruct H8 as (loff,H8). 
           destruct H8 as (s1IS,s'Is); simpl_eq; subst. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 
                                           ((h,ops,lv),bookkeepNormal P m (l, ((h, Num(I ii) :: ops), lv)) (OFFSET.jump l o,(h,ops,lv))R)).
             econstructor;eauto.
           assert (X : (exists iii:Int.t, exists n:nat, 
                 Z_of_nat (length loffset) = (high - low + 1)%Z /\
                 low <= Int.toZ iii <= high /\
                 Z_of_nat n = (Int.toZ iii) - low /\
                 nth_error loffset n = Some o)).
            exists ii. exists n. split. trivial. split. trivial. split. trivial. trivial. 
           eapply  IHIntraStepStar_intra;eauto.

 (* Exception Caught *)
   destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ _ _ HinstrAt H5) as 
     (Hact, [Heq|[(mid,Heq)|(mid,(Heq,(m0,(HH1,HH2))))]]) ; rewrite Heq in HSP.
    (* basic instruction *)
      subst.
      inversion_mine HSP.
      assert (H5H := H5).
      inversion_mine H5; simpl_eq. inversion_mine H14; simpl_eq. 
      (*1/3*)
        assert (Hpre:SP_preEC P m l cn A0 s0 
                     ((h2, Ref loc2 :: OperandStack.empty, l1),
                     bookkeepException P m (l, ((h2, Ref loc2 :: s), l1)) (h2, loc2) R)). econstructor;eauto. 
        inversion_mine H6; simpl_eq.
        assert (H3_1:=H3 _ _ _ HMB H17).
        eapply  IHIntraStepStar_intra;eauto.
      (*2/3*) inversion_mine Heq.
      (*2/3*)
       inversion_mine H6; simpl_eq.
       assert (H3_1:=H3 _ _ _ HMB H18).
       eapply  IHIntraStepStar_intra;eauto.
       econstructor. eauto. eauto. eauto. eauto. 
    (* A virtual call *)
      destruct HSP as (msig,HSP1).
      destruct HSP1.
       (*InvStat*)
        destruct H; simpl_eq. discriminate H3. 
      (*InvVirt*)
        destruct H; simpl_eq. inversion_mine H3. 
        inversion_mine H1.
        inversion_mine H6; simpl_eq.
        assert (H5H := H5). inversion_mine H5; simpl_eq.
        clear H9. clear H10.
        inversion_mine H18; simpl_eq.
        inversion_mine H8; simpl_eq.
        clear H12; clear H11.
        assert (EXC := NullException _ _ _ _ H23 H21). subst.
        destruct (H13 _ _ _ _ _ _ S (H14 _ _ HA) _ _ H19 H23) as (H13_1,H13_2). clear H13; clear H13_2.
        assert (H13_1_1 := H13_1 _ _ HMB H22). clear H13_1.
        eapply (IHIntraStepStar_intra _ _ _ _ Hlookup Hfind _ HMB); eauto. clear  IHIntraStepStar_intra.
        econstructor.  eauto. eauto. eauto. eauto. eauto. 
     (*InvStatic*)
      destruct HSP as (msig,HSP1).
      destruct HSP1.
       (*InvStat*)
        destruct H; simpl_eq. inversion_mine H3. 
        inversion_mine H1.
        inversion_mine H6; simpl_eq.
      (*InvVirt*)
        destruct H; simpl_eq. discriminate H3.

   (* exec_call *) 
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP. 
  destruct HSP as (msig,HSP). destruct HSP. 
    (*InvStat*)
      destruct H2 as (Heq1,HSP);inversion_mine HSP.
      destruct HGP as (BST0,HGP0). 
      destruct (HGP0 _ _ _ _ _ _ H4 _ H2) as (HGP1,HGP2).
      inversion_mine H. 
      clear H8; clear H11.
      (* no exception *)
        assert (H21H := H21).
        inversion_mine H21;simpl_eq. 
        inversion_mine H15;simpl_eq. 
        assert (HJ:= HGP1 _ H22).
        assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ H4 H21 _ H22 _ _ _ _ _ 
           ((stack2localvar (args ++ os) (length args),h1), 
               bookkeepCall P m (l,((h1, args++os),l1)) (m0, (os, stack2localvar(args++os)(length args))) R)
           _ _ _ (refl_equal _) (refl_equal _) HJ).
        rewrite H25 in HH.
        assert (Hpre : SINV_pre mpre mpost A0 (length (METHODSIGNATURE.parameters (snd mid))) 
                              P m l s0 ((h2, cons_option ov os, l1),act3)).  
        econstructor;unfold mkMethSpec; eauto.
        intros PRE. 
        cut (mpre ((stack2localvar(args++os)(length args),h1),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length args))) R) /\
                (((h1, OperandStack.empty), stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid)))),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid))))) R) = 
                (((h1, OperandStack.empty), stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid)))),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid))))) R)). 
              intros HYP. rewrite H25 in HYP. assert (HH1 := HH HYP). clear HH. destruct HYP as (HYP1,HYP2). assert (HH2 := HH1 HYP2). rewrite H25. eauto.
        split. eauto. eauto.

        eapply (IHIntraStepStar_intra Msig);eauto.
      (* exception caught *)
        assert (H21H := H21).
        inversion_mine H21;simpl_eq. 
       inversion_mine H17;simpl_eq.
        inversion_mine H23;simpl_eq.  
       assert (HJ := HGP1 _ H22).
        assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ H4 H25 _ H22 _ _ _ _ _ 
           ((stack2localvar (args ++ os) (length args),h1), 
               bookkeepCall P m (l,((h1, args++os),l1)) (m0, (os, stack2localvar(args++os)(length args))) R)
           _ _ _ (refl_equal _) (refl_equal _) HJ).
       rewrite H27 in HH.
       assert (Hpre : SINV_preEC mpre mpost A0 (length (METHODSIGNATURE.parameters (snd mid))) 
                P m l e s0 ((h2, Ref loc :: nil, l1), act3)). 
       econstructor;unfold mkMethSpec;eauto.
        intros PRE. 
        cut (mpre ((stack2localvar(args++os)(length args),h1),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length args))) R) /\
                (((h1, OperandStack.empty), stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid)))),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid))))) R) = 
                (((h1, OperandStack.empty), stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid)))),
                         bookkeepCall P m (l,((h1, args++os),l1))
                          (m0, (os, stack2localvar(args++os)(length (METHODSIGNATURE.parameters(snd mid))))) R)). 
              intros HYP. rewrite H27 in HYP. assert (HH1 := HH HYP). clear HH. destruct HYP as (HYP1,HYP2). assert (HH2 := HH1 HYP2). rewrite H27. eauto.
        split. eauto. eauto.

       eapply  (IHIntraStepStar_intra Msig);eauto.
    (*InvVirt*)
      destruct H2 as (Heq1,HSP);inversion_mine HSP.
      simpl_eq. 
      destruct HGP as (BST0,HGP0). 
      inversion_mine H. 
      (* no exception *)
        assert (H22H := H22).
        inversion_mine H22;simpl_eq. 
        inversion_mine H18;simpl_eq. 
        clear H8; clear H10; clear H11; clear H12. 
        assert (IHS1 := IHIntraStepStar_intra _ _ _ _ Hlookup Hfind _ HMB).
        assert (H9_1 := H9 _ H24); clear H9. 
        assert (IHS2 := IHS1 _ _ _ _ s0 _ _ _ _ _ _ _ _ (refl_equal _)(refl_equal _) (refl_equal _) (refl_equal _)H9_1); clear IHS1. 
        cut (VINV_pre mpre mpost A0 (length (METHODSIGNATURE.parameters (snd (cn,mid))))
                                P m l s0 ((h2, cons_option ov os,l1), act3)). intros VINV. 
           assert (IHS3 := IHS2  VINV Q HQ); clear IHS2. assumption. 
        clear H9_1. 
        econstructor. eauto. eauto. eauto.
        destruct (BST0 _ _ _ _ _ _ _ _ _ HinstrAt H2 H4 H26) as (((PRE1,POST1),MI1),(LST1,(LAT1,(MSTlookup',Implies)))). 
        inversion_mine Implies. simpl_eq.
            cut (A1 = PRE1); intros; subst. Focus 2. congruence.
            cut (B2 = POST1); intros; subst. Focus 2. congruence.
            cut (A2 = mpre); intros; subst. Focus 2. congruence.
            cut (B3 = mpost); intros; subst. Focus 2. congruence.
            eapply H. 
            clear H; clear H7; clear H8.  
            unfold mkMethSpec. intros H_PRE1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H26)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H23). clear HGPmeth; clear HGPctxt.
            assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H26)_ H23). 
            assert (HH_1 := HH _ _ _ _ _ ((stack2localvar (args ++ Ref loc :: os) (1+length args), h1), 
                                      bookkeepCall P m (l,((h1, args++Ref loc :: os),l1))
                                      (m', (os, stack2localvar(args++Ref loc::os)(Datatypes.S(length args)))) R)
                            _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear HGP_m0; clear HH.
            simpl in HH_1.
            rewrite H28 in H_PRE1.
            rewrite H28 in HH_1.
            simpl in H_PRE1.
            simpl in HH_1.
            rewrite H28. simpl.
            eauto.
            eauto.
      (* exception caught *)
        assert (H22H := H22).
        inversion_mine H22;simpl_eq. 
        inversion_mine H18;simpl_eq.
        clear H8; clear H9; clear H11; clear H12. 
        assert (IHS1 := IHIntraStepStar_intra _ _ _ _ Hlookup Hfind _ HMB).
        inversion_mine H24. simpl_eq. 
        assert (H10_1 := H10 _ _ _ HMB H17); clear H10. 
        assert (IHS2 := IHS1 _ _ _ _ s0 _ _ _ _ _ _ _ _ (refl_equal _) (refl_equal _) (refl_equal _) (refl_equal _) H10_1); clear IHS1. 
        cut (VINV_preEC mpre mpost A0 (length (METHODSIGNATURE.parameters mid)) P m l e
                                s0 ((h2, Ref loc :: nil,l1), act3)). intros VINV. 
           assert (IHS3 := IHS2  VINV Q HQ); clear IHS2. assumption. 
        clear H10_1. 
        econstructor. eauto. eauto. eauto. Focus 2. eauto. Focus 2. eauto.
        destruct (BST0 _ _ _ _ _ _ _ _ _ HinstrAt H2 H4 H26) as (((PRE1,POST1),MI1),(LST1,(LAT1,(MSTlookup',Implies)))). 
        inversion_mine Implies. simpl_eq. 
            cut (A1 = PRE1); intros; subst. Focus 2. congruence.
            cut (B2 = POST1); intros; subst. Focus 2. congruence.
            cut (A2 = mpre); intros; subst. Focus 2. congruence.
            cut (B3 = mpost); intros; subst. Focus 2. congruence.
            eapply H. 
            clear H; clear H7; clear H8.  
            unfold mkMethSpec. intros H_PRE1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H26)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H23). clear HGPmeth; clear HGPctxt.
            assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H26)_ H23). 
            assert (HH_1 := HH _ _ _ _ _ ((stack2localvar (args ++ Ref loc0 :: os) (1+length args), h1), 
                                      bookkeepCall P m (l,((h1, args++Ref loc0 :: os),l1))
                                      (m', (os, stack2localvar(args++Ref loc0::os)(Datatypes.S(length args)))) R)
                            _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear HGP_m0; clear HH.
            simpl in HH_1.
            rewrite H28 in H_PRE1.
            rewrite H28 in HH_1.
            simpl in H_PRE1.
            simpl in HH_1.
            rewrite H28. simpl.
            eauto.
Qed.

Lemma SOUND_STRONG:  forall M l s1 R l' s1' R', 
      V5.Reachable P M ((l,s1),R) ((l',s1'),R')  ->
      forall Msig ms LST LAT, MSPEC_LOOKUP MST Msig = Some ((ms,,LST),,LAT) ->
      findMethod P Msig = Some M -> 
      forall MB , METHOD.body M = Some MB ->
      forall b A B I s0  ss ss',
        ss = (s1,R) ->
        ss' = (s1',R') ->
        SP_Judgement P MST M LST LAT b l A B I -> 
        A s0 ss ->
        I ss ss'.
intros M l s1 R l' s1' R' (M',HR).
refine (V5.ReachableStar_ind P 
    (fun (ms ms': Method * IntraNormalState) =>
     let (M,lsR) := ms in 
     let (ls,R) := lsR in 
     let (l,s1) := ls in 
     let (M',lsR') := ms' in 
     let (ls',R') := lsR' in 
     let (l',s1') := ls' in 
    forall Msig ms LST LAT, MSPEC_LOOKUP MST Msig = Some ((ms,,LST),,LAT) ->
      findMethod P Msig = Some M ->
     forall MB , METHOD.body M = Some MB ->
     forall b A B I s0 ss ss',
        ss = (s1,R) ->
        ss' = (s1',R') ->
       SP_Judgement P MST M LST LAT b l A B I -> 
       A s0 ss ->
       I ss ss') _ _ _ _ _ _  HR);clear M l s1 R M' l' s1' R' HR; intros.
  (* Base case refl *)
    destruct s as ((l,s),R).
    intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 s0 ss ss' Hss Hss' Hsp HA.
    destruct (GP_SP_Judgement_inv _ _ HGP _ _ _ _  _ Hlookup Hfind _ _ _ _ _ Hsp)
      as (instr, (HinstrAt, HSP));clear Hsp.
    destruct (kind_of_instr instr). inversion_mine HSP;eauto.
    destruct HSP as (msig,H). destruct H. 
    (*InvStatic*)
       destruct H as (H1,H2). inversion_mine H2;eauto. 
    (*InvVirt*)
      destruct H as (H1,H2). inversion_mine H2;eauto. 
    (*Return*)
      inversion_mine HSP. eauto. 
    (*If*)
      inversion_mine HSP. eauto. 
    (*Goto*)
      inversion_mine HSP. eauto. 
    (*Lswitch*)
      destruct HSP as (Instr, HSP); inversion_mine HSP. eauto. 
    (*Tswitch*)
      destruct HSP as (Instr, HSP); inversion_mine HSP. eauto. 
  (* exec_intra *)
    destruct s as (ls,R);destruct s' as (ls',R');destruct s'' as (ls'',R'').
    destruct ls as (l,s);destruct ls' as (l',s');destruct ls'' as (l'',s'').
    intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 s0 ss ss' Hss Hss' Hsp HA;
     destruct (GP_SP_Judgement_inv _ _ HGP _ _ _ _ _ Hlookup Hfind _ _ _ _ _ Hsp)
       as (instr, (HinstrAt, HSP));clear Hsp.
    inversion_mine H.
     (* No exception *)
      destruct (elemIntra_kind _ _ _ _ _ _ _ _ HinstrAt H2) as (HR,HH).
      destruct (kind_of_instr instr);simpl in HH, HSP;
      match goal with [H:False |- _ ] => elim H | _ => idtac end.
      (* isBasicInstr *)
      destruct HH as (ll, (Heq1, (s2, Heq2))); inversion_mine HSP;simpl_eq;eauto.
      inversion_mine Heq2.
      assert (Hpre:SP_pre P m l ll A0 s0 (s2,bookkeepNormal P m (l,s) (ll,s2) R)). econstructor. eauto. eauto. intuition.
      assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                            (refl_equal _) (H4 _ Heq1) Hpre);
      eauto.
    (* A native call, absurd up to now *)
     destruct HH as (mid, (Heq2,(m0,(Heq3,Heq4)))).
     destruct HSP as (msig,Invoke).
     destruct Invoke.
     (*InvStat*)
        simpl_eq.
        destruct H as (Heq1,Hsp). 
        inversion_mine Hsp. simpl_eq. 
     (*InvVirt*)
        simpl_eq.
        destruct H as (Heq1,Hsp).
        simpl_eq.  discriminate H3. 
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    inversion_mine Heq2.
    assert (Hpre:SP_pre P m l ll A0 s0 (s2,bookkeepNormal P m (l,s) (ll,s2) R)). econstructor;eauto.
    assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _)  H6 Hpre); eauto.
    inversion_mine Heq2.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2, bookkeepNormal P m (l,s) (OFFSET.jump l o,s2) R)).
       econstructor;eauto.
    assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _) H5 Hpre);eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (s2,Heq);
      simpl_eq;subst;inversion_mine Heq.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0 (s2, bookkeepNormal P m (l,s) (OFFSET.jump l o, s2) R)). 
      econstructor;eauto.
    assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _) H4 Hpre);eauto.
    (* Lswitch *)
    destruct HSP as (Instr, HSP); inversion_mine HSP. 
      destruct HH. 
      (*extry exists*) 
        destruct H9 as (ii,H9).  
        destruct H9 as (i',H9). 
           destruct H9 as (o',H9).   
           destruct H9 as (ops,H9). 
           destruct H9 as (lv,H9). 
           destruct H9 as (h,H9). 
           destruct H9 as (i_ii,H9). 
           destruct H9 as (entry,H9). 
           destruct H9 as (s1IS,s'Is); simpl_eq; subst. 
           inversion_mine s'Is. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o') A0 s0 
                               ((h,ops,lv), bookkeepNormal P m (l,((h,Num (I ii) ::ops),lv)) (OFFSET.jump l o', ((h,ops),lv)) R)).
             econstructor;eauto.

            assert (exists i' : Z, (exists i : Int.t, In (i', o') l0 /\ i' = Int.toZ i)). 
              exists (Int.toZ ii). exists ii. split. trivial.  trivial. 
            assert (H4_1 := H4 o' H9); clear H4. 
            assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _)  H4_1 Hpre). 
            unfold SP_inv in UU. eauto. 
     (*No entry exists - default def is taken*)
        destruct H9 as (ii,H9).  
           destruct H9 as (ops,H9). 
           destruct H9 as (lv,H9). 
           destruct H9 as (h,H9). 
           destruct H9 as (no_entry,H9). 
           destruct H9 as (sIS,l's'Is); simpl_eq; subst. 
           inversion_mine l's'Is.
          assert (H5_1 := H5 ii no_entry); clear H5. 
           assert (LS: SP_LSwitch_Def_pre P m l (OFFSET.jump l def) A0 ii s0 
                                  ((h,ops,lv), bookkeepNormal P m (l,((h,Num (I ii) ::ops),lv)) (OFFSET.jump l def, ((h,ops),lv)) R)).
             econstructor. eauto. eauto. eauto.  trivial.
            assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _)  H5_1 LS). 
          unfold SP_LSwitch_Def_inv in UU. 
          assert (U_ : exists h0, exists ss, exists lv0, (h, Num(I ii)::ops, lv) = (h0, Num(I ii) :: ss, lv0)). 
            exists h; exists ops; exists lv. trivial. 
          assert (UU1 := UU _ (h,Num(I ii)::ops, lv) R _ _ U_ (refl_equal _) (refl_equal _) H2); clear UU.
          eauto.
   (*TSwitch*)
    destruct HSP as (Instr, HSP); inversion_mine HSP. 
      destruct HH. 
      (*Default jump*) 
        destruct H9 as (ii,H9).  
        destruct H9 as (h,H9). 
           destruct H9 as (ops,H9).    
           destruct H9 as (lv,H9). 
           destruct H9 as (arith1,H9). 
           destruct H9 as (arith2,H9). 
           destruct H9 as (s1IS,s'Is); simpl_eq; subst. 
           inversion_mine s'Is. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l def) A0 s0 
                              ((h,ops,lv), bookkeepNormal P m (l,((h,Num (I ii) ::ops),lv)) (OFFSET.jump l def, ((h,ops),lv)) R)).
             econstructor;eauto.
            assert (exists i : Int.t, Z_of_nat (length loffset) = (high - low + 1)%Z /\ (Int.toZ i < low \/ high < Int.toZ i)). 
              exists ii. split. trivial.  trivial. 
            assert (H4_1 := H4 H9); clear H4. 
            assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _) H4_1 Hpre). 
            unfold SP_inv in UU. eauto. 
     (*Other case*)
        destruct H9 as (ii,H9).  
           destruct H9 as (n,H9). 
           destruct H9 as (o,H9). 
           destruct H9 as (h,H9).  
           destruct H9 as (ops,H9). 
           destruct H9 as (lv,H9). 
           destruct H9 as (arith1,H9). 
           destruct H9 as (arith2,H9). 
           destruct H9 as (arith3,H9). 
           destruct H9 as (loff,H9). 
           destruct H9 as (sIS,l's'Is); simpl_eq; subst. 
           inversion_mine l's'Is. 
           assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 s0
                     ((h,ops,lv), bookkeepNormal P m (l,((h,Num (I ii) ::ops),lv)) (OFFSET.jump l o, ((h,ops),lv)) R)).
             econstructor;eauto.
           assert (X : (exists iii:Int.t, exists n:nat, 
                 Z_of_nat (length loffset) = (high - low + 1)%Z /\
                 low <= Int.toZ iii <= high /\
                 Z_of_nat n = (Int.toZ iii) - low /\
                 nth_error loffset n = Some o)).
            exists ii. exists n. split. trivial. split. trivial. split. trivial. trivial. 
           assert (H5_1 := H5 o X); clear H5. 
           assert (LS: SP_LSwitch_Def_pre P m l (OFFSET.jump l o) A0 ii s0
                              ((h,ops,lv), bookkeepNormal P m (l,((h,Num (I ii) ::ops),lv)) (OFFSET.jump l o, ((h,ops),lv)) R)).
             econstructor. eauto. eauto. eauto.  trivial.
            assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _)
                             (refl_equal _) H5_1 Hpre). 
          unfold SP_inv in UU. eauto.
   (* Exception Caught *)
   destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ _ _ HinstrAt H4)
    as (HR',[Heq|[(mid,Heq)|(mid,(Heq,(m0,(HH1,HH2))))]]); rewrite Heq in HSP.
   subst.
    (* basic instruction *)
    inversion_mine HSP. inversion_mine H9. simpl_eq.
    assert (Hpre:SP_preEC P m l e A0 s0
                 ((h2, Ref loc2 :: OperandStack.empty, l1), 
                  bookkeepException P m (l, ((h1,s1),l1)) (h2,loc2) R)).
      econstructor; eauto.
    assert (UU := H1 _ _ _ _ Hlookup Hfind _ HMB  _ _ _ _ s0 _ _ (refl_equal _)
                             (refl_equal _) (H5 _ _ _ HMB H18) Hpre). eauto. 
    (* A virtual call  *)
      destruct HSP as (msig, Invoke). destruct Invoke. 
      (*InvStat, absurd*)
      destruct H; simpl_eq. discriminate H5. 
      (*InvVirt*)
      destruct H as (HH1,HH2); subst; simpl_eq. 
       inversion_mine H9.
       assert (H4H := H4). 
       inversion_mine H4.
       inversion_mine H7. simpl_eq. simpl_eq. simpl_eq. 
       inversion_mine H4. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. 
         simpl_eq. Focus 2. simpl_eq. Focus 2. simpl_eq. Focus 2. simpl_eq. 
         Focus 2. simpl_eq. Focus 2. simpl_eq. Focus 2. simpl_eq. Focus 2. simpl_eq. 
      inversion_mine HH2. clear H7; clear H9; clear H13; clear H15.
      assert (NULL:= NullException _ _ _ _ H14 H8). subst; simpl_eq. 
      destruct ( H16 _ _ _ _ _ _ R (H17 _ _ HA) _ _ H12 H14) as (Caught,Uncaught); clear Uncaught; clear H16.
      assert (Caught1 :=  Caught _ _ HMB H10); clear Caught.
      assert (H1_1 := H1 _ _ _ _ Hlookup Hfind _ HMB); clear H1. 
      cut (I1 (((h1, args ++ Null ::s), l1),R) (s'',R'')). eauto. 
      assert (H1_2 := H1_1 _ _ _ _ s0 _ _ (refl_equal _) (refl_equal _) Caught1).
      cut (VINV_preECNull P m l A0 (length args) s0 
                     ((h2, Ref loc2 :: nil,l1), 
                      bookkeepException P m (l, ((h1, args ++ Null ::s), l1)) (h2,loc2) R)). 
         rewrite H12. intros VINV.
         assert (H1_3 := H1_2 VINV); clear H1_2. 
         inversion_mine H1_3.  eauto. 
      econstructor. eauto. eauto. eauto. eauto. eauto. 
    (* A native call, absurd up to now *)
      destruct HSP as (msig, Invoke). destruct Invoke. 
      (*InvStat, absurd*)
      destruct H; simpl_eq. inversion_mine H5.
      inversion_mine H2. simpl_eq. 
      (*InvVirt*)
      destruct H.  simpl_eq. discriminate H5. 
  (* exec_call *) 
    destruct s as (ls,R);destruct s2 as (ls2,R2);destruct s3 as (ls3,R3).
    destruct ls as (l,s);destruct ls2 as (l2,s2);destruct ls3 as (l3,s3).
  intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 s0 ss ss' Hss Hss' Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ HGP _ _ _ _ _ Hlookup Hfind _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP.
  destruct HSP as (msig, Invoke). destruct Invoke. 
   (*InvStatic*)
     destruct H3 as (Heq1,HSP);inversion_mine HSP.
     destruct HGP as (BST0,HGP0). 
     destruct (HGP0 _ _ _ _ _ _ H5 _ H3) as (HGP1,HGP2).
    simpl_eq;inversion_mine H. 
     (* no exception *)
       assert (H22H := H22).
       inversion_mine H22;simpl_eq. 
       inversion_mine H18;simpl_eq. 
       assert (HJ := HGP1 _ H25).
       assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ H5 H24 _ H25 _ _ _ _ _ 
           ((stack2localvar (args ++ os) (length args), h1), bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) _ _ _ (refl_equal _) (refl_equal _) HJ).
       rewrite H28 in HH.
       assert (Hpre : SINV_pre mpre mpost A0 (length (METHODSIGNATURE.parameters (snd mid))) P m l s0
                          ((h2, cons_option ov os, l1), R2)).  
        econstructor;unfold mkMethSpec;eauto.
        intros MPRE.
        cut (    (((h1, OperandStack.empty),
                            stack2localvar (args ++ os) (length args)),
                           bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) = 
                    (((h1, OperandStack.empty),
                            stack2localvar (args ++ os) (length args)),
                           bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R)). intros X.
        rewrite H28 in MPRE. rewrite H28 in X. rewrite H28. eauto.
        eauto.
      assert (UU:= H2 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _ (refl_equal _) (refl_equal _)
                 (H10 _ H26) Hpre).
      apply H15. inversion_mine UU.
      eapply H. eauto. eauto. 
        unfold mkMethSpec;eauto.
        intros MPRE.
        cut (    (((h1, OperandStack.empty),
                            stack2localvar (args ++ os) (length args)),
                           bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) = 
                    (((h1, OperandStack.empty),
                            stack2localvar (args ++ os) (length args)),
                           bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R)). intros X.
        rewrite H28 in MPRE. rewrite H28 in X. rewrite H28. eauto.
      eauto.
     intuition.
   (* exception caught *)
     assert (H22H := H22).
     inversion_mine H22;simpl_eq.
     inversion_mine H18;simpl_eq. 
     assert (HJ := HGP1 _ H25).
     assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ H5 H24 _ H25 _ _ _ _ _ 
         ((stack2localvar (args ++ os) (length args), h1) , bookkeepCall P m (l, ((h1, args ++ os),l1))
                                  (m0, (os, stack2localvar (args ++ os) (length args))) R) 
        _ _ _ (refl_equal _)  (refl_equal _) HJ).
     rewrite H28 in HH.
     inversion_mine H26.
     assert (Hpre : SINV_preEC mpre mpost A0 (length (METHODSIGNATURE.parameters (snd mid))) P m l 
              e s0 ((h2, Ref loc :: nil, l1),R2)).
       econstructor. eauto. eauto. eauto. 
         unfold mkMethSpec;eauto.
         intros MPRE.
         cut (    (((h1, OperandStack.empty),
                             stack2localvar (args ++ os) (length args)),
                            bookkeepCall P m (l, ((h1, args ++ os),l1))
                                   (m0, (os, stack2localvar (args ++ os) (length args))) R) = 
                     (((h1, OperandStack.empty),
                             stack2localvar (args ++ os) (length args)),
                             bookkeepCall P m (l, ((h1, args ++ os),l1))
                                   (m0, (os, stack2localvar (args ++ os) (length args))) R)). intros X.
         rewrite H28 in MPRE. rewrite H28 in X. rewrite H28. eauto.
         eauto.
       eauto.
     eauto.
     assert (UU:= H2 _ _ _ _ Hlookup Hfind _ HMB _ _ _ _ _ _ _
               (refl_equal _) (refl_equal _) (H11 _ _ _ H17 H20) Hpre).
     apply H15. inversion_mine UU.
    eapply H. eauto. eauto. 
         unfold mkMethSpec;eauto.
         intros MPRE.
         cut (    (((h1, OperandStack.empty),
                             stack2localvar (args ++ os) (length args)),
                            bookkeepCall P m (l, ((h1, args ++ os),l1))
                                   (m0, (os, stack2localvar (args ++ os) (length args))) R) = 
                     (((h1, OperandStack.empty),
                             stack2localvar (args ++ os) (length args)),
                             bookkeepCall P m (l, ((h1, args ++ os),l1))
                                   (m0, (os, stack2localvar (args ++ os) (length args))) R)). intros X.
         rewrite H28 in MPRE. rewrite H28 in X. rewrite H28. eauto.
         eauto.
       eauto.
     eauto. 
   (*InvStatic*)
     destruct H3 as (Heq1,HSP);inversion_mine HSP.
    simpl_eq;inversion_mine H. 
     (* no exception *)
       clear H11; clear H12; clear H13. 
        assert (H23H := H23).
       inversion_mine H23;  simpl_eq. 
       inversion_mine H13;  simpl_eq.
       assert (H10_1 := H10 _ H27); clear H10. 
       assert (H2_1 := H2 _ _ _ _ Hlookup Hfind _ HMB); clear H2.
       clear H9. 
       cut (I1 (((h1, args ++ Ref loc :: os),l1),R) (s3,R3)). intros INV. eauto. 
       assert (H2_2 := H2_1 _ _ _ _ s0 _ _ (refl_equal _) (refl_equal _) H10_1).
       cut (mkMethSpec mpre mpost 
                  ((stack2localvar (args++Ref loc::os) (1+length args),h1),
                    bookkeepCall P m (l, ((h1, args ++ Ref loc ::os),l1))
                                   (m', (os, stack2localvar (args ++ Ref loc :: os) (1+length args))) R)
                  ((h2,Normal ov),R2)). intros SPEC; rewrite H24 in SPEC. 
         cut (VINV_pre mpre mpost A0 (length args) P m l s0 (((h2, cons_option ov os),l1),R2)). 
            intros INV; rewrite H24 in INV; assert (H2_3 := H2_2 INV); clear H2_2. 
            inversion_mine H2_3. eapply H. eauto. eauto.  rewrite H24. eauto. eauto. 
         econstructor. eauto. eauto. eauto.  rewrite H24; eauto. eauto.
      destruct HGP as (BST0,HGP0). 
      destruct (BST0 _ _ _ _ _ _ _ _ _ HinstrAt H3 H5 H22) as (((PRE1,POST1),MI1),(LST1,(LAT1,(MSTlookup',Implies)))). 
             inversion_mine Implies. simpl_eq. 
            cut (A1 = PRE1); intros; subst. Focus 2. congruence.
            cut (B2 = POST1); intros; subst. Focus 2. congruence.
            cut (A2 = mpre); intros; subst. Focus 2. congruence.
            cut (B3 = mpost); intros; subst. Focus 2. congruence.
            cut (mkMethSpec PRE1 POST1 
                       ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc ::os),l1))
                                   (m', (os, stack2localvar (args ++ Ref loc :: os) (1+length args))) R)
                       ((h2,Normal ov),R2)). rewrite H24.  eauto.  
            clear H; clear H2; clear H8. 
            cut (PRE1 ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc ::os),l1))
                                   (m', (os, stack2localvar (args ++ Ref loc :: os) (1+length args))) R) ->
                   POST1 ((stack2localvar (args ++ Ref loc :: os) (1 + length args), h1),
                        bookkeepCall P m (l, ((h1, args ++ Ref loc ::os),l1))
                                   (m', (os, stack2localvar (args ++ Ref loc :: os) (1+length args))) R)
                       ((h2,Normal ov),R2)). intros. eauto.
            intro PRE1_1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H22)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H26). clear HGPmeth; clear HGPctxt.
            assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H22)_ H26). 
            assert (HH_1 := HH _ _ _ _ _ 
                                        ((stack2localvar (args ++ Ref loc :: os) (1+length args), h1),
                                              bookkeepCall P m (l, ((h1, args ++ Ref loc ::os),l1))
                                                   (m', (os, stack2localvar (args ++ Ref loc :: os) (1+length args))) R)
                                        _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear HH.
            simpl in HH_1. eauto. 
   (* exception caught *)
     assert (H23H := H23). 
     inversion_mine H23;simpl_eq. 
     inversion_mine H19;simpl_eq. 
     inversion_mine H27;simpl_eq. 
     clear H9; clear H10; clear H12; clear H13.
     cut (I1 (((h1, args ++ Ref loc0 ::os),l1),R) (s3,R3)). eauto.
     assert (H11_1 := H11 _ _ _ HMB H22); clear H11.
     assert (H2_1 := H2 _ _ _ _ Hlookup Hfind _ HMB); clear H2.
     assert (H2_2 := H2_1 _ _ _ _ s0 _ _ (refl_equal _) (refl_equal _) H11_1); clear H2_1.
     cut (mkMethSpec mpre mpost 
                   ((stack2localvar (args++Ref loc0::os) (1+length args),h1),
                      bookkeepCall P m (l, ((h1, args ++ Ref loc0 ::os),l1))
                                             (m', (os, stack2localvar (args ++ Ref loc0 :: os) (1+length args))) R)
                   ((h2,Exception loc),R2)). 
             intros SPEC; rewrite H29 in SPEC. 
         cut (VINV_preEC mpre mpost A0 (length args) P m l e s0 
                      ((h2, Ref loc :: nil,l1),R2)). 
            intros INV; rewrite H29 in INV; assert (H2_3 := H2_2 INV); clear H2_2. 
            inversion_mine H2_3. eapply H. eauto. eauto. rewrite H29. eauto. eauto. eauto. 
         econstructor. eauto. eauto. eauto. rewrite H29; eauto. eauto. eauto. 
      destruct HGP as (BST0,HGP0). 
      destruct (BST0 _ _ _ _ _ _ _ _ _ HinstrAt H3 H5 H25) as (((PRE1,POST1),MI1),(LST1,(LAT1,(MSTlookup',Implies)))). 
             inversion_mine Implies. simpl_eq. 
            cut (A1 = PRE1); intros; subst. Focus 2. congruence.
            cut (B2 = POST1); intros; subst. Focus 2. congruence.
            cut (A2 = mpre); intros; subst. Focus 2. congruence.
            cut (B3 = mpost); intros; subst. Focus 2. congruence.
            cut (mkMethSpec PRE1 POST1 
                       ((stack2localvar (args ++ Ref loc0 :: os) (1 + length args), h1),
                      bookkeepCall P m (l, ((h1, args ++ Ref loc0 ::os),l1))
                                             (m', (os, stack2localvar (args ++ Ref loc0 :: os) (1+length args))) R)
                       ((h2,Exception loc),R2)). rewrite H29.  eauto.  
            clear H; clear H2; clear H8. 
            cut (PRE1 ((stack2localvar (args ++ Ref loc0 :: os) (1 + length args), h1),
                      bookkeepCall P m (l, ((h1, args ++ Ref loc0 ::os),l1))
                                             (m', (os, stack2localvar (args ++ Ref loc0 :: os) (1+length args))) R) ->
                   POST1 ((stack2localvar (args ++ Ref loc0 :: os)(1 + length args), h1),
                      bookkeepCall P m (l, ((h1, args ++ Ref loc0 ::os),l1))
                                             (m', (os, stack2localvar (args ++ Ref loc0 :: os) (1+length args))) R)
                       ((h2,Exception loc),R2)). intros. eauto.
            intro PRE1_1.
            destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H25)) as (HGPmeth, HGPctxt).
            assert (HGP_m0 := HGPmeth _ H26). clear HGPmeth; clear HGPctxt.
            assert (HH := SOUND_POST _ _ _ H0 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H25)_ H26). 
            assert (HH_1 := HH _ _ _ _ _ 
                                    ((stack2localvar (args ++ Ref loc0 :: os) (1+length args), h1),
                                      bookkeepCall P m (l, ((h1, args ++ Ref loc0 ::os),l1))
                                             (m', (os, stack2localvar (args ++ Ref loc0 :: os) (1+length args))) R)
                       _ _ _ (refl_equal _) (refl_equal _) HGP_m0); clear HH.
            simpl in HH_1. eauto. 
  (* Call step *)
  destruct s'' as (ls'',R').
  destruct ls'' as (l'',s'').
   intros Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 s0 ss ss' Hss Hss' Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ HGP _ _ _ _ _ Hlookup Hfind _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
   subst.
  assert (H_H := H).
  inversion_mine H; simpl_eq. 
  inversion_mine H7; simpl_eq. 
  (*InvStatic*)
      destruct HSP as (msig, Invoke). destruct Invoke.
      (*InvStatic*)
        destruct H as (Heq1,HSP);simpl_eq.
        inversion_mine HSP.
       destruct HGP as (BST0,HGP0). 
       destruct (HGP0 _ _ _ _ _ _ H4 _ H) as (HGP1,HGP2);simpl_eq.
       assert (HJ := HGP1 _ H0).
       assert (HH:= H2 _ _ _ _ H4 H12 _ H0 _ _ _ _
                               (((stack2localvar (args ++ os') (length args)),h),
                                   bookkeepCall P m (pc, ((h, args ++ os'),l))
                                             (m0, (os', stack2localvar (args ++ os') (length args))) act)
                                _ _ (refl_equal _) (refl_equal _) HJ).
       clear HJ H2.
       rewrite H14 in HH.
       apply H17.
       eapply H8;eauto.
       rewrite H14.
       intros;unfold StrongInvConversion in HH;eauto.
      (*InvVirt absurd*)
        destruct H as (Heq1,HSP);simpl_eq.
  (*InvVirt*)
      destruct HSP as (msig, Invoke). destruct Invoke.
      (*InvStatic absurd*)
        destruct H as (Heq1,HSP);simpl_eq.
      (*InvVirt*)
        destruct H as (Heq1,HSP);simpl_eq.
        inversion_mine HSP.
        cut (I1 (((h, args ++ Ref loc ::os'),l),act) (s'',R')). eauto.
        clear H9; clear H10; clear H11; clear H15. 
        assert (H8_1 := H8 _ _ _ _ _ (s'',R') _ _ _ _ _ _ H14 (H16 _ _ HA) H_H); clear H8.
        cut (mpre ((stack2localvar (args++Ref loc :: os') (S(length args)),h),
                         bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                       (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args))))  act) ->
              MI ((stack2localvar (args++Ref loc :: os') (1+length args),h),
                    bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                                (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args)))) act)
                    (s'',R')). intros. 
         assert (H8_2 := H8_1 H7); clear H8_1. eauto.
        destruct HGP as (BST0,HGP0). 
        destruct (BST0 _ _ _ _ _ _ _ _ _ HinstrAt H H4 H12) as (((PRE1,POST1),MI1),(LST1,(LAT1,(MSTlookup',Implies)))). 
         inversion_mine Implies. simpl_eq. 
         clear H7. 
         cut (A1 = PRE1); intros; subst. Focus 2. congruence.
         cut (I2 = MI1); intros; subst. Focus 2. congruence.
         cut (A2 = mpre); intros; subst. Focus 2. congruence.
         cut (I3 = MI); intros; subst. Focus 2. congruence.
         cut (mkMethInv PRE1 MI1 
                       ((stack2localvar (args ++ Ref loc :: os')
                          (1 + length args), h),bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                                (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args))))  act) (s'',R')). rewrite H14. intros MINV. 
             assert (H8_2 := H8 _ _ MINV); clear H8. 
               unfold mkMethInv in H8_2. rewrite H14 in H11. eauto. 
         clear H8; clear H9; clear H10. clear H8_1.
         cut (PRE1 ((stack2localvar (args ++ Ref loc :: os') (1 + length args), h),
                          bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                                (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args)))) act) ->
                   MI1 ((stack2localvar (args ++ Ref loc :: os') (1 + length args), h), 
                            bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                                (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args)))) act) 
                         (s'',R')).
             intros. eauto.
         intro PRE1_1.
         destruct (HGP0 _ _ _ _ _ _ MSTlookup' _ (lookup_findMethod _ _ _ _ H12)) as (HGPmeth, HGPctxt).
         assert (HGP_m0 := HGPmeth _ H0). clear HGPmeth; clear HGPctxt.
         assert (H2_1 := H2 _ _ _ _ MSTlookup' (lookup_findMethod _ _ _ _ H12) _ H0); clear H2.
          simpl in HGP_m0.
         assert (H2_2 := H2_1 _ _ _ _  ((stack2localvar (args ++ Ref loc :: os')(1+length args), h),
                            bookkeepCall P m (pc,((h, args ++ Ref loc ::os'),l))
                                                                (m',(os',stack2localvar(args ++ Ref loc::os')(S(length args)))) act) 
                                _ _ (refl_equal _) (refl_equal _) HGP_m0); clear H2_1. 
         unfold StrongInvConversion in H2_2. simpl in H2_2. eauto. 
Qed.

  Definition valid M (LAT:LocalAnnoTable)(B:LocalPost) (I:StrongInv) s0 l s S := 
   (forall rR, V5.BigStep P M ((l,s),S) rR -> B s0 (s,S) rR) /\
   (forall l' s' S',  V5.IntraStepStar_intra P M ((l,s),S)  ((l',s'),S') ->
        forall (Q:Assertion),  Anno_LOOKUP LAT l' = Some Q ->  Q s0 (s',S')) /\
   (forall l' s' S', V5.Reachable P M ((l,s),S) ((l',s'),S') -> I (s,S) (s',S')).

  Lemma SOUND:  
      forall M Msig ms LST LAT, 
        MSPEC_LOOKUP MST Msig = Some ((ms,, LST),, LAT) ->
        findMethod P Msig = Some M ->
      forall MB , METHOD.body M = Some MB ->
      forall b A B I l,
        SP_Judgement P MST M LST LAT b l A B I -> 
      forall s0 s S,
        A s0 (s,S) -> valid M LAT B I s0 l s S.
  Proof.
   intros M Msig ms LST LAT Hlookup Hfind MB HMB b A B0 I0 l HSP s0 s S HA;repeat split;
    intros.
   eapply SOUND_POST;eauto.
   eapply SOUND_ANNO;eauto.
   eapply SOUND_STRONG;eauto.
  Qed.

 End VALID. 

End Validity.
