Add LoadPath "../Bicolano".

Require Export BaseLogic2.

Open Scope type_scope. 

Module Validity (Rules:RULES).

 Import Rules.

 Section SEM.
  Variable (P: Program).
   
  Lemma exec_call_isCallInstr : forall M l i s ret' m' s' r', 
    instructionAt M l = Some i ->
    exec_call P M (l, s) ret' m' s' r' -> kind_of_instr i = CallInstr.
  Proof.
    intros M l i s ret' m' s' r' Hi Hcall;inversion_mine Hcall;
    match goal with 
    | [H1: ElementaryCallStep _ _ _ _ |- _] =>
      inversion_mine H1;simpl_eq;trivial
    | _ => idtac
    end.
  Qed.

  Lemma elemExcep_isBasicInstr : forall M l i s h' loc,
     instructionAt M l = Some i ->
     ElementaryIntraExceptionStep P M (l, s) (h', loc) ->
     kind_of_instr i = BasicInstr \/ 
     (exists mid, i = Invokevirtual mid) \/
     (exists mid, i =(Invokestatic mid) /\ 
        exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true).
  Proof.
    intros M l i s h' loc Hi Hel;inversion_mine Hel; simpl_eq;simpl;auto.
    right;right;eauto.
    inversion_mine H3;simpl_eq;auto;eauto.
   Qed.

  Lemma elemIntra_kind : forall M l i s1 s3, 
     instructionAt M l = Some i ->
     ElementaryIntraNormalStep P M (l, s1) s3 ->
      match kind_of_instr i with
      | BasicInstr => 
         exists ll, next M l = Some ll /\ exists s2, s3 = (ll, s2)
      | CallInstr => 
         exists mid, i =(Invokestatic mid) /\ 
           exists M0, findMethod P mid = Some M0 /\ METHOD.isNative M0 = true
      | ReturnInstr => False
      | IfInstr o => 
            (exists ll, next M l = Some ll /\ exists s2, s3 = (ll,s2))
         \/ (exists s2, s3 = (OFFSET.jump l o, s2))
      | GotoInstr o => exists s2, s3 = (OFFSET.jump l o, s2)
      | SwitchInstr => True
      end.
  Proof.
   intros M l i s1 s3 H Hi;inversion_mine Hi;simpl_eq;simpl;eauto.
  Qed.

  Lemma elemReturn_isBasicReturn : forall M l i s r,
      instructionAt M l = Some i ->
      ElementaryReturnStep P M (l,s) r ->
      kind_of_instr i = ReturnInstr.
  Proof. 
   intros M l i s r Hi Hr;inversion_mine Hr;simpl_eq;simpl;trivial.
  Qed.

 End SEM.

 Section VALID.

  Variables (P: Program) 
            (G: Assum_TYPE) 
            (Anno: AnnoType)
            (MST: MSPEC_TYPE).

  Definition startState MB (s0:LocalVar.t) (h0:Heap.t) :=
     (BYTECODEMETHOD.firstAddress MB, (h0, OperandStack.empty,s0)).
 


  Hypothesis HGP: GP G Anno MST P.


  Variable subclass : ClassName -> ClassName -> Prop.

  Hypothesis BST: BehavSubTyping MST subclass.

  Hypothesis Bic1: forall (h:Heap) e loc h1,
      Heap.new h P e = Some (loc, h1) -> Heap.typeof h1 loc = Some e.

  Hypothesis Bic2: forall c C Msig MNAME mname m D,
                                            CLASS.method c Msig = Some m ->
                                            CLASS.name c = C ->
                                            METHODSIGNATURE.name Msig = MNAME ->
                                            MNAME = (C,mname) ->
                                            subclass D C ->
                                            (exists E, exists e, exists Msig1, exists m', 
                                                     subclass D E /\ subclass E C /\
                                                     CLASS.name e = E /\
                                                     METHODSIGNATURE.name Msig1 = (E,mname) /\
                                                     CLASS.method e Msig1 = Some m').

  Lemma SOUND_POST:  forall (M: Method) S r, V5.BigStep P M S r ->
      forall T MI , MSPEC_LOOKUP MST M = SomeT (T,,  MI) ->
      forall MB , METHOD.body M = Some MB ->
      forall C b A B I l s0 h0 s1,
        S = (l,s1) ->
        SP_Judgement G Anno MST P C M b l A B I -> 
        A (s0,h0) s1 ->
        B (s0,h0) s1 r.
  Proof.
  induction 1 using V5.BigStep_ind;intros T MI Hlookup MB HMB C b A B0 I0 
   l s0 h0 s1 Heq Hsp HA;subst;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  (* exec_return *)  
  inversion_mine H.
   (* return *)
    rewrite (elemReturn_isBasicReturn _ _ _ _ _ _ HinstrAt H0) in HSP;
     inversion_mine HSP;eauto.
   (* uncaught exception *)
   destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H2) as 
    [Heq| [(mid,Heq) | (mid,(Heq,(m0,(HH1,HH2))))]];simpl_eq;
    rewrite Heq in HSP.
    (* basic instruction *)
    inversion_mine HSP;eauto.
    (* A virtual call absurd up to now *)
    inversion_mine HSP. destruct H. destruct H. simpl_eq. destruct H.  simpl_eq.
       inversion_mine H0. clear H11; clear H10; clear H9; clear H8.
       eapply (H14 (s0, h0) (h1, s2, l1) (h2, Exception loc2)).
       inversion_mine H2. simpl_eq. simpl_eq.
       inversion_mine H9. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. 
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq.
      assert (AA := H13 (s0,h0) (h1, args ++ Null :: s, l1) HA).
      assert (X := H12 (s0,h0) args s l1 h1 AA h2 loc2 H17 H18). destruct X.
      eauto.
    (* A native call, absurd up to now *)
    destruct HSP. inversion_mine H. destruct H0;  simpl_eq.
       inversion_mine H0. clear H11; clear H10; clear H9. simpl_eq.
       inversion_mine H0. simpl_eq. 
  (* exec_intra *)
  inversion_mine H.
   (* No exception *)
   assert (HH := elemIntra_kind _ _ _ _ _ _ HinstrAt H1);
    destruct (kind_of_instr instr);simpl in HH, HSP;
    match goal with [H:False |- _ ] => elim H | _ => idtac end.
    (* isBasicInstr *)
    destruct HH as (ll, (Heq1, (s2, Heq2))); inversion_mine HSP;simpl_eq;eauto.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    assert (UU := IHBigStep _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
                  (H3 _ Heq1) Hpre);unfold SP_post in UU;eauto.
    (*InvStatic native (6 goals left)*)
       destruct HH. destruct H. destruct H2. destruct H2.
       inversion_mine HSP. destruct H4. destruct H. simpl_eq. 
       inversion_mine H4. simpl_eq.
       destruct H. simpl_eq.
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    assert (UU := IHBigStep _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
                  H5 Hpre);unfold SP_post in UU;eauto.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2). 
      econstructor;eauto.
    assert (UU := IHBigStep _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
                  H4 Hpre);unfold SP_post in UU;eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (s2,Heq);simpl_eq;subst.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2).
      econstructor;eauto.
    assert (UU := IHBigStep _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
                  H3 Hpre);unfold SP_post in UU;eauto.
   (* Exception Caught *)
    destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H3) as 
       [Heq|[(mid,Heq)|(mid,(Heq,(m0,(HH1,HH2))))]]; rewrite Heq in HSP.
    (* basic instruction *)
    inversion_mine HSP. inversion_mine H5. simpl_eq.
    assert (Hpre:SP_preEC P m l e A0 (s0, h0) 
             (h2, Ref loc2 :: OperandStack.empty, l1)).
      econstructor;eauto.
    assert (UU := IHBigStep _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
               (H4 _ _ _ HMB H17) Hpre).
    inversion_mine UU. injection H10;intros;subst;eauto.
    (*InvVirt*)
        inversion_mine HSP. destruct H. destruct H. simpl_eq. 
        destruct H. simpl_eq.
        inversion_mine H3. simpl_eq. simpl_eq.
        inversion_mine H7. simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. 
            Focus 2. simpl_eq. Focus 2. simpl_eq.  Focus 2. simpl_eq. Focus 2. simpl_eq.
            Focus 2. simpl_eq. Focus 2. simpl_eq.
        inversion_mine H1. clear H9. clear H10. clear H8. clear H13.
        assert (H15_1 := H15 (s0, h0) (h1, args ++ Null :: s, l1) HA). clear H15.
        assert (H14_1 := H14 (s0,h0) args s l1 h1 H15_1 h2 loc2 H11 H12). clear H14. destruct H14_1.
        assert (H1_1 := H1 pc' MB HMB). clear H1.
        inversion_mine H5. simpl_eq.
        cut ( Heap.typeof h2 loc2 = Some (Heap.LocationObject (javaLang, NullPointerException))).
          Focus 2. eapply (Bic1 h1). assumption.
        intros. simpl_eq.
        assert (H1_2 := H1_1 H18). clear H1_1.
        assert (IHBS := IHBigStep T MI Hlookup bm HMB C true 
                           (VINV_preECNull P A0 (length (METHODSIGNATURE.parameters mid0)))
                           (VINV_postECNull P B1 (length (METHODSIGNATURE.parameters mid0)))
                           (VINV_invECNull P I1 (length (METHODSIGNATURE.parameters mid0)))
                           pc' s0 h0 (h2, Ref loc2 :: OperandStack.empty, l1) (refl_equal _) H1_2).
               clear IHBigStep.
        cut (VINV_preECNull P A0 (length (METHODSIGNATURE.parameters mid0))
           (s0, h0) (h2, Ref loc2 :: OperandStack.empty, l1)).
            Focus 2. econstructor. eauto. eauto. eauto. apply (refl_equal).
        intros.
        assert (IHBS2 := IHBS H1). clear IHBS.
        inversion_mine IHBS2.
        assert (H5_1 := H5 s l1 h2 h1 args (h1, args ++ Null :: s, l1) loc2 H11 H12 (refl_equal _) (refl_equal _)).
        eauto.
    (*InvStat*) (*Native*)
        inversion_mine HSP. destruct H. Focus 2. destruct H. simpl_eq. 
        destruct H. simpl_eq.
        inversion_mine H1. simpl_eq.
  (* exec_call, uncaught exception *)
  destruct HGP as (HGP1,HGP2);clear HGP1.
  inversion_mine H. inversion_mine H4;simpl_eq. 
    (*InvStatic*)
     destruct HSP. destruct H. Focus 2. destruct H. simpl_eq.  destruct H. simpl_eq. 
     destruct H1. clear H7. clear H6. clear H10.
     assert (H15_1 := H15 _ _ HA). clear H15.
     assert (H11_1 := H11 _ _ _ _ _ H15_1 h2 loc). clear H11. simpl_eq. 
     assert (H11_2 := H11_1 H14). clear H11_1.
     cut (T0
              (stack2localvar (args ++ os)
                 (length (METHODSIGNATURE.parameters x)), h1)
              (h2, Exception loc)).
     eauto.
    destruct (HGP2 _ _ _ _ _ H8 H2 (refl_equal _)) as (C', HJ).
    assert (HH := IHBigStep _ _ H2 _ H8 C' _ _ _ _ _ 
        (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
        (refl_equal _) (refl_equal _)).
    rewrite H14 in HH;eauto.
    (*InvVirtual*)
     simpl_eq.
     destruct HSP. destruct H. destruct H. simpl_eq.  destruct H. simpl_eq. 
     destruct H1. clear H7. clear H6. clear H11. clear H5.
     assert (H15_1 := H15 _ _ HA). clear H15.
     assert (H10_1 := H10 _ _ _ _ _ _ H15_1 h2 loc). clear H10. simpl_eq.
       cut (length args = N). Focus 2. transitivity (length (METHODSIGNATURE.parameters x)). assumption. 
         symmetry. assumption. intros. 
     assert (H10_2 := H10_1 H5). clear H10_1.
     cut (T0 (stack2localvar (args ++ Ref loc0 :: os) (1 + N), h1)
            (h2, Exception loc)).
     eauto.
     simpl_eq. 
     (*now exploit BST to show that T0 is respected*)

        (*The situation is as follows: 
           m: current method (ie caller), with specification (T,MI)
           C: current class (ie caller)
 
           x: MethodSignature of the invoked method, 
           m0: Method to which x refers to statically
           (T0, MI0): specification of m0

           cn: dynamic class name of the object on which the method is invoked

           cl: name of the class that is dynamically identified as being the closest
                 superclass of cn that contains a definition of x

           m': the dynamically invoked Method, i.e. the implementation found in cl *)

         (*In order to apply Bic2, we explicitly identify the components of signatures.*)
         cut (exists cnn, exists mnn, METHODSIGNATURE.name x = (cnn,mnn)). 
           Focus 2.  destruct (METHODSIGNATURE.name x). exists c. exists s. reflexivity. intro.         
         elim H4. clear H4. intro StaticCN. intro. elim H4. clear H4. intro StaticMN. intro.
         (*I.e. StaticCN and StaticMN are the class and method names of m0*)

         cut (exists StaticClass, PROG.class P StaticCN = Some StaticClass). intros. 
                elim H5. intro StaticClass. intros. clear H5.
         (*I.e. StaticClass is the class whose name is by StaticCN*)
         cut (CLASS.name StaticClass = StaticCN). Focus 2. symmetry. 
           destruct (PROG.name_class_invariant1 P StaticCN StaticClass H7). reflexivity. intros.

         cut (CLASS.method StaticClass x = Some m0). intros. (*Proof delayed*)
         cut (subclass cn StaticCN). (* Proof delayed. this should follow for bc-verified 
                                       progs for Bicolano*)
         intros.
         assert (Bic2_1 := Bic2 StaticClass StaticCN x _ StaticMN _ cn H10 H5 (refl_equal _) H4 H11). clear Bic2.

         elim Bic2_1. clear Bic2_1. intro EN; intro XX.
         elim XX. clear XX. intro EClass. intro XX.
         elim XX. clear XX. intro ESig. intro XX.
         elim XX. clear XX. intro EMethod. intros. destruct H15. destruct H18.  destruct H19.  destruct H20.
         (*EClass is now the class on which the method is actually invoked, 
           EN the name of EClass, ESig the signature of x in E, and EMEthod the dynamically invoked Method*)


         inversion_mine BST.
         cut (subclass (CLASS.name EClass) (CLASS.name StaticClass) = True). (*Prop vs bool*) intro.
         cut (METHOD.signature m0 = x). (*Proof: Another Bicoalno property.*) intro.
         cut (METHOD.signature m' = ESig). (*Proof: Bicolano (looukup-property)*) intro.
         assert (H22_1 := H22 (CLASS.name EClass) (CLASS.name StaticClass) m0 m' T0 MI0 x ESig StaticMN H5 H1 H19 H4 H20 H23). clear H22.
         elim H22_1. clear H22_1. intro TE. intro XX.
         elim XX. clear XX. intro InvE. intro. destruct H22. destruct H24.
         assert (TE (stack2localvar (args ++ Ref loc0 :: os)
                    (1 + length (METHODSIGNATURE.parameters x)), h1) 
                    (h2, Exception loc)). eauto.

         cut (m' = EMethod). (*Bicolano*) intros.

         simpl_eq.
         assert (HGP2_1 := HGP2 m' bM TE InvE (BYTECODEMETHOD.firstAddress bM) H8 H22 (refl_equal _)).
         elim HGP2_1. clear HGP2_1. intros.
         intros.
         cut (subclass (CLASS.name EClass) (CLASS.name StaticClass) = True). Focus 2. eauto. simpl_eq. 
  Bic2 : forall (c : Class) (C : ClassName) (Msig : MethodSignature)
           (MNAME : MethodName) (mname : ShortMethodName) 
           (m : Method) (D : ClassName),
         CLASS.method c Msig = Some m ->
         CLASS.name c = C ->
         METHODSIGNATURE.name Msig = MNAME ->
         MNAME = (C, mname) ->
         subclass D C ->
         exists E : ClassName,
           (exists e : Class,
              (exists Msig1 : MethodSignature,
                 (exists m' : Method,
                    subclass D E /\
                    subclass E C /\
                    CLASS.name e = E /\
                    METHODSIGNATURE.name Msig1 = (E, mname) /\
                    CLASS.method e Msig1 = Some m')))


         elim Bic2_1. clear Bic2_1. intros. destruct H15. destruct H18.
       inversion_mine BST. 
       cut (subclass cn cl = True).
         Focus 2. eauto. intros. 
       cut (exists mname , METHODSIGNATURE.name x = (cn,mname)).
       intros. elim H7. intros. clear H7.
       cut (METHOD.signature m0 = x). intros.
       assert (H4_1 := H4 cn cl m0 m' T0 MI0 x). clear H4.
       assert( (METHOD.signature m') x0 H5 H1 H7 H10). clear H4.
       cut
     inversion_mine H.findMethod P x = Some m0 -> 
    destruct (HGP2 _ _ _ _ _ H8 H1 (refl_equal _)) as (C', HJ).
    assert (HH := IHBigStep _ _ H2 _ H8 C' _ _ _ _ _ 
        (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
        (refl_equal _) (refl_equal _)).
    rewrite H14 in HH;eauto.

  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP.
   destruct HSP. destruct H1.
   (*InvStatic*)
      destruct H1. simpl_eq.
      destruct HGP. 
      inversion_mine H. inversion_mine H5;simpl_eq. 
      inversion_mine H2. simpl_eq.
      clear IHBigStep. clear H8. clear H7.
      assert (AA := H16 (s0, h0) (h1, args ++ os, l1) HA). clear HA.
      assert (H12_1 := H12 (s0,h0) args os l1 h1 AA h2 loc H15). clear H12.
      cut (T0
            (stack2localvar (args ++ os)
               (length (METHODSIGNATURE.parameters mid)), h1)
            (h2, Exception loc)). intros.
      assert (H12_2 := H12_1 H H10). clear H12_1. eauto.
     instructionAt m l = Some (Invokestatic mid) -> BigStep P m0
         (BYTECODEMETHOD.firstAddress bM,
         (h1, OperandStack.empty, stack2localvar (args ++ os) (length args)))
         (h2, Exception loc) T0
     (stack2localvar (args ++ os) (length (METHODSIGNATURE.parameters mid)),
     h1) (h2, Exception loc).

     MSPEC_LOOKUP MST m0 = SomeT (T0,, MI0) 
      assert (HH := IHBigStep T MI Hlookup MB HMB C  _ _ H3 _ H20 C' _ _ _ _ _ 
      (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
      (refl_equal _) (refl_equal _)).
  rewrite H26 in HH;eauto.

   destruct HSP as (msig, (Heq1,HSP));inversion_mine HSP.
  destruct HGP as (HGP1,HGP2);clear HGP1.
  inversion_mine H. inversion_mine H16;simpl_eq. 
  destruct (HGP2 _ _ _ _ _ H20 H3 (refl_equal _)) as (C', HJ).
  assert (HH := IHBigStep _ _ H3 _ H20 C' _ _ _ _ _ 
      (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
      (refl_equal _) (refl_equal _)).
  rewrite H26 in HH;eauto.
  (* exec_call *)
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP;
   destruct HSP as (msig, (Heq1,HSP));inversion_mine HSP.
  destruct HGP as (HGP1,HGP2);clear HGP1.
  inversion_mine H. 
   (* no exception *)
   inversion_mine H17;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H21 H4 (refl_equal _)) as (C', HJ).
   assert (HH := IHBigStep1 _ _ H4 _ H21 C' _ _ _ _ _ 
      (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
      (refl_equal _) (refl_equal _)).
   rewrite H27 in HH.
   assert (Hpre : SINV_pre T0 A0 (length (METHODSIGNATURE.parameters mid)) 
           (s0,h0) (h2, cons_option ov os, l1)).  econstructor;eauto.
   assert (UU := IHBigStep2 _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
               (H9 _ H22) Hpre).
   inversion_mine UU;eauto.
   (* exception caught *)
   inversion_mine H17;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H21 H4 (refl_equal _)) as (C', HJ).
   assert (HH := IHBigStep1 _ _ H4 _ H21 C' _ _ _ _ _ 
      (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ 
      (refl_equal _) (refl_equal _)).
   rewrite H27 in HH.
   inversion_mine H22.
   assert (Hpre : SINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid)) 
            l e (s0,h0) (h2, Ref loc :: nil, l1)). econstructor;eauto.
   assert (UU := IHBigStep2 _ _ Hlookup _ HMB _ _ _ _ _ _ _ _ _ (refl_equal _) 
               (H10 _ _ _ H16 H19) Hpre).
   inversion_mine UU;eauto.
  Qed.


 Lemma SOUND_ANNO:  forall (M: Method) S S', V5.IntraStepStar_intra P M S S' ->
      forall T MI , MSPEC_LOOKUP MST M = SomeT (T,,  MI) ->
      forall MB , METHOD.body M = Some MB ->
      forall C b A B I s0 h0 l s l' s',
        S = (l,s) ->
        S' = (l',s') ->
        SP_Judgement G Anno MST P C M b l A B I -> 
        A (s0,h0) s ->
        forall (Q:LocalAssn), Anno_LOOKUP Anno (C,M,l') = SomeT Q -> 
        Q (s0,h0) s'.
 Proof.
  induction 1 using V5.IntraStepStar_intra_ind;
   intros T MI Hlookup MB HMB C b A B0 I0 s0 h0 l 
          S l' S' Heq1 Heq2 Hsp HA Q HQ;subst;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  (* Base case : S = S' *)
  injection Heq2;clear Heq2;intros;subst.
  destruct (kind_of_instr instr);inversion_mine HSP;eauto.
   eapply H;eauto.
   destruct H as (H1,H2);inversion_mine H2;eauto. eapply H4;eauto.
   eapply H;eauto.
   eapply H0;eauto.
   eapply H;eauto.
  (* exec_intra *)
  inversion_mine H.
   (* No exception *)
   assert (HH := elemIntra_kind _ _ _ _ _ _ HinstrAt H1);
    destruct (kind_of_instr instr);simpl in HH, HSP;
    match goal with [H:False |- _ ] => elim H | _ => idtac end.
    (* isBasicInstr *)
    destruct HH as (ll, (Heq1, (s2, Heq2))); inversion_mine HSP;simpl_eq;eauto.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    eapply IHIntraStepStar_intra;eauto.
    (* A native call, absurd up to now *)
    destruct HSP as (msig, (Heq1, Hsp));
     destruct HH as (mid, (Heq2,(m0,(Heq3,Heq4))));
      inversion_mine Hsp;simpl_eq.
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2).
      econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (s2,Heq);simpl_eq;subst.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2). 
      econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
   (* Exception Caught *)
   destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H3) as 
     [Heq|[(mid,Heq)|(mid,(Heq,(m0,(HH1,HH2))))]]; rewrite Heq in HSP.
    (* basic instruction *)
    inversion_mine HSP. inversion_mine H5. simpl_eq.
    assert (Hpre:SP_preEC P m l e A0 (s0, h0) 
             (h2, Ref loc2 :: OperandStack.empty, l1)). econstructor;eauto.
    eapply  IHIntraStepStar_intra;eauto.
    (* A virtual call absurd up to now *)
    destruct HSP as (msig, (HH1, HH2));subst;simpl_eq.
    (* A native call, absurd up to now *)
    destruct HSP as (msig, (Heq1, Hsp));inversion_mine Hsp;simpl_eq.
   (* exec_call *) 
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP;
   destruct HSP as (msig, (Heq1,HSP));inversion_mine HSP.
  destruct HGP as (HGP1,HGP2);clear HGP1.
  inversion_mine H. 
   (* no exception *)
   inversion_mine H17;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H21 H4 (refl_equal _)) as (C', HJ).
   assert (HH := SOUND_POST _ _ _ H0 _ _ H4 _ H21 C' _ _ _ _ _ 
       (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ
       (refl_equal _) (refl_equal _)).
   rewrite H27 in HH.
   assert (Hpre : SINV_pre T0 A0 (length (METHODSIGNATURE.parameters mid)) 
            (s0,h0) (h2, cons_option ov os, l1)).  econstructor;eauto.
   eapply  IHIntraStepStar_intra;eauto.
   (* exception caught *)
   inversion_mine H17;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H21 H4 (refl_equal _)) as (C', HJ).
   assert (HH := SOUND_POST _ _ _ H0 _ _ H4 _ H21 C' _ _ _ _ _ 
       (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ
       (refl_equal _) (refl_equal _)).
   rewrite H27 in HH.
   inversion_mine H22.
   assert (Hpre : SINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid)) 
            l e (s0,h0) (h2, Ref loc :: nil, l1)). econstructor;eauto.
   eapply  IHIntraStepStar_intra;eauto.
 Qed.


 Lemma SOUND_STRONG:  forall M l s l' s', V5.Reachable P M (l,s) (l',s') ->
      forall T MI , MSPEC_LOOKUP MST M = SomeT (T,,  MI) ->
      forall MB , METHOD.body M = Some MB ->
      forall C b A B I s0 h0,
        SP_Judgement G Anno MST P C M b l A B I -> 
        A (s0,h0) s ->
        I s s'.
 Proof.
  intros M l s l' s' (M', HR).
  refine (V5.ReachableStar_ind P 
    (fun (ms ms': Method * IntraNormalState) =>
     let (M,ls) := ms in 
     let (l,s) := ls in 
     let (M',ls') := ms' in 
     let (l',s') := ls' in 
     forall T MI , MSPEC_LOOKUP MST M = SomeT (T,,  MI) ->
     forall MB , METHOD.body M = Some MB ->
     forall C b A B I s0 h0,
     SP_Judgement G Anno MST P C M b l A B I -> 
     A (s0,h0) s ->
     I s s') _ _ _ _ _ _ HR);clear M l s l' s' M' HR;intros.
  (* Base case refl *)
  destruct s as (l,s);intros T MI Hlookup MB HMB C b A B0 I0 s0 h0 Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  destruct (kind_of_instr instr);inversion_mine HSP;eauto.
  destruct H as (H1,H2);inversion_mine H2;eauto. 
  (* exec_intra *)
  destruct s as (l,s);destruct s' as (l',s');destruct s'' as (l'',s'');
  intros T MI Hlookup MB HMB C b A B0 I0 s0 h0 Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  inversion_mine H.
   (* No exception *)
   assert (HH := elemIntra_kind _ _ _ _ _ _ HinstrAt H2);
    destruct (kind_of_instr instr);simpl in HH, HSP;
    match goal with [H:False |- _ ] => elim H | _ => idtac end.
    (* isBasicInstr *)
    destruct HH as (ll, (Heq1, (s2, Heq2))); inversion_mine HSP;simpl_eq;eauto.
    inversion_mine Heq2.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    assert (UU := H1 _ _ Hlookup _ HMB _ _ _ _ _ _ _ (H4 _ Heq1) Hpre _ H2);
    eauto.
    (* A native call, absurd up to now *)
    destruct HSP as (msig, (Heq1, Hsp));
     destruct HH as (mid, (Heq2,(m0,(Heq3,Heq4))));
      inversion_mine Hsp;simpl_eq.
    (* isIfInstr *)
    inversion_mine HSP.
    destruct HH as [(ll', (Heq1, (s2,Heq2)))|(s2,Heq2)];simpl_eq;subst.
    inversion_mine Heq2.
    assert (Hpre:SP_pre P m l ll A0 (s0, h0) s2). econstructor;eauto.
    assert (UU := H1 _ _ Hlookup _ HMB _ _ _ _ _ _ _ H6 Hpre _ H2);eauto.
    inversion_mine Heq2.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2).
       econstructor;eauto.
    assert (UU := H1 _ _ Hlookup _ HMB _ _ _ _ _ _ _ H5 Hpre _ H2);eauto.
    (* isGotoInstr *)
    inversion_mine HSP;destruct HH as (s2,Heq);
      simpl_eq;subst;inversion_mine Heq.
    assert (Hpre:SP_pre P m l (OFFSET.jump l o) A0 (s0, h0) s2). 
      econstructor;eauto.
    assert (UU := H1 _ _ Hlookup _ HMB _ _ _ _ _ _ _ H4 Hpre _ H2);eauto.
   (* Exception Caught *)
   destruct (elemExcep_isBasicInstr _ _ _ _ _ _ _ HinstrAt H5) 
    as [Heq|[(mid,Heq)|(mid,(Heq,(m0,(HH1,HH2))))]]; rewrite Heq in HSP.
    (* basic instruction *)
    inversion_mine HSP. inversion_mine H7. simpl_eq.
    assert (Hpre:SP_preEC P m l e A0 
               (s0, h0) (h2, Ref loc2 :: OperandStack.empty, l1)).
      econstructor;eauto.
    assert (UU := H1 _ _ Hlookup _ HMB _ _ _ _ _ _ _ 
             (H4 _ _ _ HMB H18) Hpre);inversion_mine UU.
    injection H7;clear H7;intros;subst;simpl_eq;eauto.
    (* A virtual call absurd up to now *)
    destruct HSP as (msig, (HH1, HH2));subst;simpl_eq.
    (* A native call, absurd up to now *)
    destruct HSP as (msig, (Heq1, Hsp));inversion_mine Hsp;simpl_eq.
  (* exec_call *) 
  destruct s as (l,s);destruct s2 as (l2,s2);destruct s3 as (l3,s3).
  intros T MI Hlookup MB HMB C b A B0 I0 s0 h0 Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  rewrite (exec_call_isCallInstr _ _ _ _ _ _ _ _ _ HinstrAt H) in HSP;
   destruct HSP as (msig, (Heq1,HSP));inversion_mine HSP.
  destruct HGP as (HGP1,HGP2);clear HGP1.
  simpl_eq;inversion_mine H. 
   (* no exception *)
   inversion_mine H22;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H23 H5 (refl_equal _)) as (C', HJ).
   assert (HH := SOUND_POST _ _ _ H0 _ _ H5 _ H23 C' _ _ _ _ _ 
       (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ
       (refl_equal _) (refl_equal _)).
   rewrite H28 in HH.
   assert (Hpre : SINV_pre T0 A0 (length (METHODSIGNATURE.parameters mid)) 
           (s0,h0) (h2, cons_option ov os, l1)).  econstructor;eauto.
   assert (UU:= H2 _ _ Hlookup _ HMB _ _ _ _ _ _ _ 
             (H10 _ H24) Hpre);inversion_mine UU;eauto.
   (* exception caught *)
   inversion_mine H22;simpl_eq. 
   destruct (HGP2 _ _ _ _ _ H23 H5 (refl_equal _)) as (C', HJ).
   assert (HH := SOUND_POST _ _ _ H0 _ _ H5 _ H23 C' _ _ _ _ _ 
       (stack2localvar (args ++ os) (length args)) h1 _ (refl_equal _) HJ
       (refl_equal _) (refl_equal _)).
   rewrite H28 in HH.
   inversion_mine H24.
   assert (Hpre : SINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid)) l 
            e (s0,h0) (h2, Ref loc :: nil, l1)). econstructor;eauto.
   assert (UU:= H2 _ _ Hlookup _ HMB _ _ _ _ _ _ _ 
             (H11 _ _ _ H17 H20) Hpre);inversion_mine UU;eauto.
  (* Call step *)
  destruct s'' as (l'',s'').
  intros T MI Hlookup MB HMB C b A B0 I0 s0 h0 Hsp HA;
   destruct (GP_SP_Judgement_inv _ _ _ _ HGP _ _ _ _ _ _ _ Hsp)
   as (instr, (HinstrAt, HSP));clear Hsp.
  inversion_mine H;simpl_eq;destruct HSP as (msig, (Heq1,HSP));simpl_eq.
  inversion_mine HSP.
  destruct HGP as (HGP1,HGP2);clear HGP1. simpl_eq.
  destruct (HGP2 _ _ _ _ _ H0 H4 (refl_equal _)) as (C', HJ).
  assert (HH:= H2 _ _ H4 _ H0 _ _ _ _ _
          (stack2localvar (args ++ os') (length args)) h HJ (refl_equal _)
          _ _ (refl_equal _)).
  clear HJ H2 HGP2 H1.
  rewrite H14 in HH;eauto.
 Qed.

  Definition valid C M (B:LocalPost) (I:StrongInv) s0 h0 l s := 
   (forall r, V5.BigStep P M (l,s) r -> B (s0,h0) s r) /\
   (forall l' s', 
      V5.IntraStepStar_intra P M (l,s) (l',s') ->
      forall (Q:LocalAssn), 
        Anno_LOOKUP Anno (C,M,l') = SomeT Q ->  Q (s0,h0) s') /\
   (forall l' s', V5.Reachable P M (l,s) (l',s') -> I s s').

  Lemma SOUND:  
      forall M T MI , MSPEC_LOOKUP MST M = SomeT (T,,  MI) ->
      forall MB , METHOD.body M = Some MB ->
      forall C b A B I l,
      SP_Judgement G Anno MST P C M b l A B I -> 
      forall s0 h0 s,
        A (s0,h0) s -> valid C M B I s0 h0 l s.
  Proof.
   intros M T MI Hlookup MB HMB C b A B0 I0 l HSP s0 h0 s HA;repeat split;
    intros.
   eapply SOUND_POST;eauto.
   eapply SOUND_ANNO;eauto.
   eapply SOUND_STRONG;eauto.
  Qed.

 End VALID. 

End Validity.

        inversion_mine HSP. destruct H. Focus 2. destruct H. simpl_eq. 
        destruct H. simpl_eq.
        inversion_mine H1. simpl_eq.

        inversion_mine H3. simpl_eq. 
        simpl_eq.
        inversion H5. simpl_eq.
        inversion_mine H1. clear H9. clear H11. clear H18.
        assert (H13_1 := H13 pc' e bm HMB H10). clear H13.
        assert (IHBS := IHBigStep T MI Hlookup bm HMB C true 
            (SINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid0)) l e)
            (SINV_postEC T0 B1 (length (METHODSIGNATURE.parameters mid0)) l e)
            (SINV_invEC T0 I1 (length (METHODSIGNATURE.parameters mid0)) l e)
                           pc' s0 h0  (h2, Ref loc2 :: OperandStack.empty, l1) (refl_equal _) H13_1).
               clear IHBigStep.
        cut (SINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid0)) l e
           (s0, h0) (h2, Ref loc2 :: OperandStack.empty, l1)).
            Focus 2. econstructor. eauto. eauto. eauto. apply (refl_equal).
        intros.
        assert (IHBS2 := IHBS H1). clear IHBS.
        inversion_mine IHBS2.
        assert (H5_1 := H5 s l1 h2 h1 args (h1, args ++ Null :: s, l1) loc2 H11 H12 (refl_equal _) (refl_equal _)).
          inversion_mine H7. simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. simpl_eq. simpl_eq.  simpl_eq. simpl_eq.
           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
            Focus 2. simpl_eq. Focus 2. simpl_eq.  Focus 2. simpl_eq. Focus 2. simpl_eq.
            Focus 2. simpl_eq. Focus 2. simpl_eq.
        inversion_mine H1. clear H9. clear H10. clear H8. clear H13.
        assert (H15_1 := H15 (s0, h0) (h1, args ++ Null :: s, l1) HA). clear H15.
        assert (H14_1 := H14 (s0,h0) args s l1 h1 H15_1 h2 loc2 H11 H12). clear H14. destruct H14_1.
        assert (H1_1 := H1 pc' MB HMB). clear H1.
        inversion_mine H5. simpl_eq.
        cut ( Heap.typeof h2 loc2 = Some (Heap.LocationObject (javaLang, NullPointerException))).
          Focus 2. eapply (Bic1 h1). assumption.
        intros. simpl_eq.
        assert (H1_2 := H1_1 H18). clear H1_1.
        assert (IHBS := IHBigStep T MI Hlookup bm HMB C true 
                           (VINV_preECNull P A0 (length (METHODSIGNATURE.parameters mid0)))
                           (VINV_postECNull P B1 (length (METHODSIGNATURE.parameters mid0)))
                           (VINV_invECNull P I1 (length (METHODSIGNATURE.parameters mid0)))
                           pc' s0 h0 (h2, Ref loc2 :: OperandStack.empty, l1) (refl_equal _) H1_2).
               clear IHBigStep.
        cut (VINV_preECNull P A0 (length (METHODSIGNATURE.parameters mid0))
           (s0, h0) (h2, Ref loc2 :: OperandStack.empty, l1)).
            Focus 2. econstructor. eauto. eauto. eauto. apply (refl_equal).
        intros.
        assert (IHBS2 := IHBS H1). clear IHBS.
        inversion_mine IHBS2.
        assert (H5_1 := H5 s l1 h2 h1 args (h1, args ++ Null :: s, l1) loc2 H11 H12 (refl_equal _) (refl_equal _)).
        eauto.

        inversion_mine H3. simpl_eq. simpl_eq.
        inversion_mine H18. simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq. 
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
                           simpl_eq. simpl_eq. simpl_eq. simpl_eq. simpl_eq.
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. 
          Focus 2. simpl_eq. clear H12. clear H9; clear H10.
          inversion_mine H5.
          assert (H11I := H11 pc' e bm H10 H18). clear H11.
      assert (AA := H14 (s0,h0) (h1, args ++ Null :: s, l1) HA).
      assert (H13_1 := H13 (s0,h0) args s l1 h1 AA h2 loc2 H22 H23). clear H13. destruct H13_1.
      assert (H1_1 := H1 pc' bm H10). clear H1.
      cut (lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm) l
           (javaLang, NullPointerException) pc').
      intros.
      assert (H1_2 := H1_1 H1). clear H1_1. inversion_mine H1_2.
        Focus 2. eapply (Bic1).

      assert (IHBS := IHBigStep T MI Hlookup MB HMB). clear IHBigStep.
      assert (IHBS2 := IHBS C true  (VINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid0)) l e) (VINV_postEC T0 B1 (length (METHODSIGNATURE.parameters mid0)) l e) (VINV_invEC T0 I1 (length (METHODSIGNATURE.parameters mid0)) l e) pc' s0 h0 (h2, Ref loc2 :: OperandStack.empty, l1) (refl_equal _) H11I). clear IHBS.

      cut (VINV_preEC T0 A0 (length (METHODSIGNATURE.parameters mid0)) l e
            (s0, h0) (h2, Ref loc2 :: OperandStack.empty, l1)).
        intros.  assert (IHBS3 := IHBS2 H1).
                 inversion_mine IHBS3.
                 apply (H33 := H3 )
 eauto.
      auto.
      
      assert (X := H12 (s0,h0) args s l1 h1 AA h2 loc2 H17 H18). destruct X.
      eauto.

    (* A virtual call absurd up to now *)
    destruct HSP as (msig, (HH1, HH2));subst;simpl_eq.
    (* A native call, absurd up to now *)
    destruct HSP as (msig, (Heq1, Hsp));inversion_mine Hsp;simpl_eq.